/*  This code handles I/O with the

    Acromag 9412 Digital I/O Board
    ------------------------------

    This board has 32 outputs accessible from the front panel from
    connectors P3 and P4. The outputs can be configured as 4 8-bit
    ports or 2 16-bit ports.

    Installation
    ------------

    J6 - IN

    J10 - IN

    J17 - Set Base Address for short (A16) memory map.
          We set it to 0x0400 i.e pins 1-2 IN, all others OUT.

    J18 - Set Base Address for standard (A24) memory map.
          Not applicable for our use.

    J19 - VME Bus Address Modifier codes.
          We use short (A16) supervisory/non-privileged access (AM codes
          0x29 and 0x29). So pins 1-2 are IN, 3-4 are OUT.

    Note: Board asserts SYSFAIL upon power-up. This is due to jumper
    J20 which is factory-installed.
    

    Memory Map
    ----------

    On the MVME16X boards, the VME A16 address space is accessible locally
    at 0xffff0000.

    With the Acromag 9412 base address set at 0x0400, the memory map i.t.o
    an offset from 0xffff0400 is

    Offset
    ------
 
    0x01 - 0x3F (odd addresses)     Module ID PROM (Read-only)

    0x81                            Status/Control Register (Read-write)

                             Bit 4: Writing 1 causes a software reset.
                                    Turns all ports off, configures them
                                    for inputs. In our case, asserts
                                    SYSFAIL. Writing 0, or reading has
                                    no effect.

                             Bit 1: 1 turns GREEN LED on, 0 turns it off.
                                    Read reflects status.
                            
                             Bit 0: 1 turns RED LED (SYSFAIL) OFF, 0 turns
                                    it on. Read reflects status.

                             Write 0x10 -> Software Reset
                                   0x0  -> Green OFF, Red ON
                                   0x1  -> Green OFF, Red OFF
                                   0x2  -> Green ON, Red ON,
                                   0x3  -> Green ON, Red OFF

    0x100                           Output Port 0 } write
    0x101                                       1 } only
    0x102                                       2 }
    0x103                                       3 }                                   
 */


#include "vxWorks.h"

#define ACROBASE       0xFFFF0400    /* start of VME A16 address space for 
                                         the 9412 board */
#define ACROPROM_STA   0x1           /* Offset to Start of Model ID PROM */
#define ACROPROM_END   0x27          /* Offset to End of Model ID PROM */
#define ACROSCR        0x81          /* Offset to Status & Control Register */
#define ACROIOP0       0x100         /* Offset to IO Port 0 */
#define ACROIOP1       0x101         /* Offset to IO Port 1 */
#define ACROIOP2       0x102         /* Offset to IO Port 2 */
#define ACROIOP3       0x103         /* Offset to IO Port 3 */
#define ACROIOP4       0x104         /* Offset to IO Port 4 */
#define ACROIOP5       0x105         /* Offset to IO Port 5 */
#define ACROIOP6       0x106         /* Offset to IO Port 6 */
#define ACROIOP7       0x107         /* Offset to IO Port 7 */

/* Print PROM contents
   In ASCII:  V  M  E  I  D  A  C  R  9  4  1  2  x  x  x  1  x  1  0  x
   In Hex:   56 4D 45 49 44 41 43 52 39 34 31 32 20 20 20 31 20 31 30 20
  
 */

 void pprom() {

    u_char i, val;
    u_char vrd8();

    printf("ACROMAG 9412 PROM Contents:\n");
    for (i = ACROPROM_STA; i <=  ACROPROM_END; i += 2) {
	 val = vrd8(ACROBASE + i);
	 printf("%c ",val);
    }
    printf("\n");

 }

/* Initialize the board. This turns off the red SYS_FAIL LED that is asserted
   on power up */

 void acro_init() {

    wacro8(ACROSCR, 0x1);
    printf("ACROMAG 9412 Initialized\n");

 }

/* Causes a software reset. Red SYS_FAIL LED is turned on. If the board is
   IO, all ports are configured as inputs */

 void acro_reset() {

    wacro8(ACROSCR, 0x10);
    printf("ACROMAG 9412 Reset\n");

 }

/* Reads status of the board from the status/control register */

 void acro_status() {

      u_char value;
      value = racro8(ACROSCR);
      if (value == 0x0) {
         printf ("Board uninitialized or failed\n");
      }
      if (value == 0x1) {
         printf ("Board inactive\n");
      }
      if (value == 0x2) {
         printf ("Board undergoing diagnostic checkout\n");
      }
      if (value == 0x3) {
         printf ("Board Uninitialized/failed\n");
      }

 }

/* Only for an IO board. Reads from specified port */

 void acro_read(u_short port) {

      u_char value;
      value = racro8(ACROIOP0 + port);
      printf(" %d read from Port %d\n", value, port);

 }

/* For IO or Output-only boards. Writes to specified port */

 void acro_write(u_short port, u_char value) {
   
      wacro8(ACROIOP0 + port, value);
      printf(" %d written to Port %d\n", value, port);

 }

 /* Miscellaneous Utility Read/Write Routines */

 void racro8(u_short offset) {

      u_char  data8;
      u_char  vrd8();

      data8 = vrd8(ACROBASE + offset);

      printf("data8 = 0x%x\n", data8);

 }

 void wacro8(u_short offset, u_char value) {

      void vwr8();
      vwr8(ACROBASE + offset,value);

 }

 void racro16(u_short offset) {

      u_short data16;    
      u_short vrd16();
 
      data16 = vrd16(ACROBASE + offset);
  
      printf("data16 = 0x%x\n", data16);

 }

 void wacro16(u_short offset, u_short value) {

      void vwr16();
      vwr16(ACROBASE + offset,value);

}

 u_short vrd16(u_short *addr)
      { 
        return(*addr);
      }


 u_char vrd8(u_char* addr)
   {
      return(*addr);
   }

 void vwr8(u_char* addr, u_char value)
   {
      *addr = value;
   }

 void vwr16(u_short* addr, u_short value)
   {
      *addr = value;
   }

