/*  This code handles I/O with the

    LeCroy 1151E Board
    ------------------

    The LeCroy 1151E is a 6U VME Scaler with 16 channels, 80 MHz counting,
    ECL inputs, and 32 bit dynamic range. 

    Installation
    ------------

    On-board switches allow setting the base address to access the board
    over VME. Currently set at 3000 => the board is to be accessed from
    0xf0300000 using D16 or D32 data transfers.

    Memory Map
    ----------

    This board requires A24/D16/D32 addressing.

    Notation: CRA = Counting Registers A
              CRB = Counting Registers B
              PCR = Preset Counting Register

    With Base Address set at 0xf0300000

    BASE + 0x0        Master Reset      (D16 Write)
                      Sets CRs to 0 and sets all bits in PCR to Preset Count 
                      mode regardless of data.

    BASE + 0x10       Bus Interrupter Registers (R/W)
           0x1F       odd addresses -> D8 access
                      even addresses -> lower 8 bits from D16 access
                      Control Register at 0x11
                      Control Vector   at 0x19

    BASE + 0x40       (CRA) D32 Read causes automatic reset to 0
           0x7C       and sets corresponding bit in PCR to Count mode.
                      Note D16 read of LSB (0x42) has same effect but
                      D16 read of MSB (0x40) has no effect. So read MSB
                      first.

                      Use these registers to preset counts with one D32 or two
                      D16 writes. Corresponding bit in the PCR is then set
                      to the Preset Count mode.

    BASE + 0x80       (CRB) Read without any effect (D32 or two D16 reads)
           0xBC      

    BASE + 0xFA       Should read 0xFAF5 (D16)
   
    BASE + 0xFC       Bits 0-9 -> Module Type = 7
                      Bits 10-15 -> Mfg. # = 1
                      Should Read 1024*1 + 7 = 1031 = 0x407 (D16)

    BASE + 0xFE       Bits 0-10 -> Serial # of Module = 2
                      Bits 12-15 -> Rev # = 1
                      Should Read 4096*1 + 2 = 4098 = 0x1002 (D16)


 */


#include "vxWorks.h"

#define LCSC_BASE       0xF0300000    /* start of VME A24 address space for 
                                         the 9412 board. Also address of
                                         the Master Reset Register */

/* Offsets */

#define LCSC_ID1        0xFA          
#define LCSC_ID2        0xFC          
#define LCSC_ID3        0xFE          

#define LCSC_CR         0x11         /* Control Register */
#define LCSC_CV         0x19         /* Control Vector */

#define LCSC_MRR        0x0          /* Master Reset Register */
#define LCSC_CRA_BASE   0x40         /* Start of CRA */
#define LCSC_CRB_BASE   0x80         /* Start of CRB */


/* Functions:
   Module Identify    void lcsc_id()
   Initialize -       void lcsc_init()
   Reset -            void lcsc_reset()
   Preset Counts -    void lcsc_preset(u_char chan, int value)
   Read Counts -      int  lcsc_read(u_char chan)
   Read and Clear -   int  lcsc_read_clr(u_char chan)
   Interrupt Set -    void lcsc_intset(u_char level)
   
*/

 void lcsc_id() {

    u_short id1, id2, id3;
    u_short rd16();

    id1 = rd16(LCSC_ID1);
    id2 = rd16(LCSC_ID2);
    id3 = rd16(LCSC_ID3);
 
    if ((id1 == 64245) && (id2 == 1031) && (id3 == 4098)) {      
       printf("This is LeCroy 1151E 16 chan VME scaler\n");
    }
    else {
       printf("Unknown Module\n");
    }
 }

 void lcsc_reset() {

       void wr16();
       wr16(LCSC_MRR, 0x0);
       printf("LeCroy 1151E reset\n");
 
 }

 
 void lcsc_preset(u_char chan, int value) {

       void wr32();
       wr32(LCSC_CRA_BASE+4*chan, value);
       printf("Channel %d preset to %d\n", chan, value);

 }

 void lcsc_read(u_char chan) {

       int value;
       int rd32();
       
       value = rd32(LCSC_CRB_BASE+4*chan);
       printf("Channel %d reads %d\n", chan, value);

 }

 void lcsc_read_loop(u_char chan, int times) {

       int i, value;
       int rd32();
       
       for (i = 0; i <= times; i++) {
           value = rd32(LCSC_CRB_BASE+4*chan);
       }

       printf("Channel %d reads %d\n", chan, value);

 }

 void lcsc_read_clr(u_char chan) {

       int value;
       int rd32();
       
       value = rd32(LCSC_CRA_BASE+4*chan);
       printf("Channel %d reads %d\n", chan, value);
       printf("Channel %d cleared\n", chan);
 }

 void lcsc_intset(u_char level) {

       void wr8();
       wr8(LCSC_CV, level);
       printf("Interrupt level set to %d\n", level);

 }

 void lcsc_intenable() {

       u_char rd8();
       void wr8();
       u_char value;

       value = rd8(LCSC_CV);
       value &= 0x10;
       wr8(LCSC_CV, value);
       printf("Interrupts enabled\n");
 }

 
 /* Miscellaneous Utility Read/Write Routines */

 u_char rd8(u_short offset) {

      u_char  data8;
      u_char  vrd8();

      data8 = vrd8(LCSC_BASE + offset);

      /*      printf("data8 = 0x%x\n", data8); */

      return data8;
 }

 void wr8(u_short offset, u_char value) {

      void vwr8();
      vwr8(LCSC_BASE + offset,value);

 }

 u_short rd16(u_short offset) {

      u_short data16;    
      u_short vrd16();
 
      data16 = vrd16(LCSC_BASE + offset);
  
      /*      printf("data16 = %d (0x%x)\n", data16, data16); */

      return data16;
 }

 void wr16(u_short offset, u_short value) {

      void vwr16();
      vwr16(LCSC_BASE + offset,value);

 }

 int rd32(u_short offset) {

      int data32;    
      int vrd32();
 
      data32 = vrd32(LCSC_BASE + offset);
  
      /*      printf("data32 = 0x%x\n", data32); */

      return data32;
 }

 void wr32(u_short offset, u_short value) {

      void vwr32();
      vwr32(LCSC_BASE + offset,value);

 }

 u_short vrd16(u_short *addr)
      { 
        return(*addr);
      }

 u_char vrd8(u_char* addr)
   {
      return(*addr);
   }

 void vwr8(u_char* addr, u_char value)
   {
      *addr = value;
   }

 void vwr16(u_short* addr, u_short value)
   {
      *addr = value;
   }

 int vrd32(int* addr)
   {
      return(*addr);
   }

 void vwr32(int* addr, int value)
   {
      *addr = value;
   }


