/************************************************************************/
/* File: tod_mk48t08.h							*/
/*									*/
/* This is a Time-Of-Day device driver for the MK48T02 and MK48T08.	*/
/* It is implemented for the MVME147, MVME167.				*/
/*									*/
/* Generic MACRO Identifiers:						*/
/*									*/
/*	TOD_DEVICE_ID		Device String Identifier		*/
/*	TOD_DEVICE_ON		Enable Time-Of-Day Clk MK48T0X		*/
/*	TOD_DEVICE_OFF		Disable Time-Of-Day Clk (save Battery)	*/
/*	TOD_READ_DEVICE		Read the Time-Of-Day MK48T0X		*/
/*	TOD_WRITE_DEVICE	Write the Time-Of-Day MK48T0X		*/
/*	TOD_DEVICE_NVRAM	Size of NV RAM				*/
/*									*/

/*
SCCS release:
@(#) tod_mk48t08.h 1.1 Delta: 93/04/07 17:21:42 Extraction: 93/04/12 16:26:29 @(#)
*/

#ifndef INCtod_mk48t08
#define INCtod_mk48t08

/************************************************************************/
/*									*/
/* MACROS BCD Conversions						*/
/*						1			*/
/*			 8421	       BCD	2631			*/
/*			 0000 8421 	vs	84268421		*/
/*			|nnnn|nnnn|   Binary   |nnnnnnnn|		*/
/*									*/
/*	The conversion macros operate on byte values only where the	*/
/*	Binary to BCD conversion is available for bytes values under	*/
/*	100.								*/

#define TOD_BCDtoBIN(n) ((((n) & 0xf0) >> 4) * 10 + ((n) & 0x0f))
#define TOD_BintoBCD(n) ((((n) / 10) << 4) | ((n) % 10))

/************************************************************************/
/*									*/
/* MACRO Device Driver							*/
/*									*/
/* Device:	MK48T08 Time Of Day Clock				*/
/*									*/
/* Clock Setup:		24:00 Hour Format				*/
/*									*/
/* MK48T08 Registers: 		(See SGS-Thompson Specification)	*/
/*									*/
/*	Address		Bits	|7|6|5|4|3|2|1|0|			*/
/*		0		|W|R|S| offset  |	Command Reg	*/
/*		1		|ST   Seconds   |	Seconds		*/
/*		2		|0|   Minutes   |	Minutes		*/
/*		3		|KS0| Hours     |	Hours		*/
/*		4	     	|0|FT Day	|	Day of week	*/
/*		5		|0|0| Day       |	Day of Month	*/
/*		6		|0|0|0| Month   |	Month		*/
/*		7		|     Year      |	Year 2 Digit	*/
/*    _______________________________________________________________	*/
/*   | Bit 7 | Bit 6 | Bit 5 | Bit 4 | Bit 3 | Bit 2 | Bit 1 | Bit 0 |	*/
/*									*/
/*   Control Register 0							*/
/*    _______________________________________________________________	*/
/*   | Write | Read  | Sign  |          O F F S E T +/-              |	*/
/*   |   1   |   0   |   X   | Write    Offset Clk   not used        |	*/
/*   |   0   |   1   |   X   | Read     Offset Clk   not used        |	*/
/*									*/
/*   Seconds Register 1							*/
/*    _______________________________________________________________	*/
/*   | Stop  |                   S E C O N D S                       |	*/
/*   |   0   |                   TOD CLK Running                     |	*/
/*									*/
/*   Hours Register 3 (MK48T02 only KickStart Bit)			*/
/*    _______________________________________________________________	*/
/*   | KickST|   0   |             H O U R S                         |	*/
/*   |   X   |   0   |                                               |	*/
/*									*/
/*									*/

#ifdef MK48T02
#define TOD_DEVICE_ID "MK48T02\0"	/* Give it a string name */
#define TOD_DEVICE_NVRAM 2048		/* Size of NVRAM */
#else
#define TOD_DEVICE_ID "MK48T08\0"	/* Give it a string name */
#define TOD_DEVICE_NVRAM 8192		/* Size of NVRAM */
#endif

#ifndef TOD_CLOCK		/* MVME147, MVME167 definition for MK48T08 */
#ifdef CLK_BASE
#define TOD_CLOCK CLK_BASE	/* if CLK_BASE, it's an FSCC */
#endif
#endif

#define TOD_WRUNLOCK		/* Write Lock/Unlock used for NVRAM (FRC) */
#define TOD_WRLOCK

#ifndef TOD_REGOFF
#define TOD_REGOFF 1		/* displacement between register bytes */
#endif

#define TOD_BASEOFFSET(n) 	*( (char *) TOD_CLOCK+(n))

#define TOD_MK48T08_CNTRL	TOD_BASEOFFSET(0x00*TOD_REGOFF)	/* Control register */
#define TOD_MK48T08_SEC		TOD_BASEOFFSET(0x01*TOD_REGOFF)	/* BCD Seconds */
#define TOD_MK48T08_MIN		TOD_BASEOFFSET(0x02*TOD_REGOFF)	/* BCD Minutes */
#define TOD_MK48T08_HRS		TOD_BASEOFFSET(0x03*TOD_REGOFF)	/* BCD Hours */
#define TOD_MK48T08_DAY		TOD_BASEOFFSET(0x04*TOD_REGOFF)	/* BCD Day of Week */
#define TOD_MK48T08_DAYDATE	TOD_BASEOFFSET(0x05*TOD_REGOFF)	/* BCD Date Day */
#define TOD_MK48T08_MODATE	TOD_BASEOFFSET(0x06*TOD_REGOFF)	/* BCD Date Month */
#define TOD_MK48T08_YRDATE	TOD_BASEOFFSET(0x07*TOD_REGOFF)	/* BCD Date Year */

/* MK48T08 Command register bits */

#define TOD_MK48T08_WR_BIT	0x80	/* "Write" bit in control byte */
#define TOD_MK48T08_RD_BIT	0x40	/* "Read" bit in control byte */
#define TOD_MK48T08_S_BIT	0x20	/* "Sign" bit in control byte */

#define TOD_MK48T08_STOP_BIT	0x80	/* "Stop Clock" bit in seconds field */
#define TOD_MK48T08_FT_BIT	0x40	/* "Frequency test" bit in day field */

/* MK48T02 Command register bits */
#define TOD_MK48T02_KS_BIT	0x80	/* Kick-start bit in hour field	*/

	/****************************************************************/
	/*								*/
	/* Device ON/OFF MACROs						*/
	/*								*/

#define TOD_MK48T02_KICKSTART { \
 	if ( TOD_MK48T08_SEC & TOD_MK48T08_STOP_BIT ) /* Clock stopped? */ \
	{	/* What to do from MK48T02 doc */ \
		TOD_WRUNLOCK \
		TOD_MK48T08_CNTRL |= TOD_MK48T08_WR_BIT; \
		TOD_MK48T08_SEC &= ~TOD_MK48T08_STOP_BIT; \
		TOD_MK48T08_HRS |= TOD_MK48T02_KS_BIT; \
		TOD_MK48T08_CNTRL &= ~(TOD_MK48T08_WR_BIT | TOD_MK48T08_RD_BIT); \
		taskDelay( 2 * sysClkRateGet() ); \
		TOD_MK48T08_CNTRL |= TOD_MK48T08_WR_BIT; \
		TOD_MK48T08_HRS &= ~TOD_MK48T02_KS_BIT; \
		TOD_MK48T08_CNTRL &= ~(TOD_MK48T08_WR_BIT | TOD_MK48T08_RD_BIT); \
		TOD_WRLOCK \
	} \
}

#define TOD_MK48T08_START { \
 	if ( TOD_MK48T08_SEC & TOD_MK48T08_STOP_BIT ) /* Clock stopped? */ \
	{ \
		TOD_WRUNLOCK \
		TOD_MK48T08_CNTRL |= TOD_MK48T08_WR_BIT; \
		TOD_MK48T08_SEC &= ~TOD_MK48T08_STOP_BIT; \
		TOD_MK48T08_CNTRL &= ~(TOD_MK48T08_WR_BIT | TOD_MK48T08_RD_BIT); \
		TOD_WRLOCK \
	} \
}

#ifdef MK48T02
#define TOD_MK48T08_ON TOD_MK48T02_KICKSTART
#else
#define TOD_MK48T08_ON TOD_MK48T08_START
#endif

#define TOD_MK48T08_OFF {	/* Is the clock on ? If so, turn it off */ \
 	if ( !(TOD_MK48T08_SEC & TOD_MK48T08_STOP_BIT) ) /* Clock on? */ \
	{ \
		TOD_WRUNLOCK \
		TOD_MK48T08_CNTRL |= TOD_MK48T08_WR_BIT; \
		TOD_MK48T08_SEC |= TOD_MK48T08_STOP_BIT; \
		TOD_MK48T08_CNTRL &= ~(TOD_MK48T08_WR_BIT | TOD_MK48T08_RD_BIT); \
		TOD_WRLOCK \
	} \
}

	/****************************************************************/
	/*								*/
	/* Device Start/Stop External Register Update MACROs		*/
	/*								*/
	/* Stop the update prior to reading/writing the Time Registers	*/
	/* This has no effect on the internal Time registers		*/
	/*								*/

#define TOD_MK48T08_RDSTOPUPDATE { /* Stop Updating to Read */ \
	TOD_WRUNLOCK \
	TOD_MK48T08_CNTRL |= TOD_MK48T08_RD_BIT; /* Stop updating clock */ \
	TOD_WRLOCK \
}

#define TOD_MK48T08_WRSTOPUPDATE { /* Stop Updating to Write */ \
	TOD_WRUNLOCK \
	TOD_MK48T08_CNTRL |= TOD_MK48T08_WR_BIT; /* Stop updating clock */ \
	TOD_WRLOCK \
}

#define TOD_MK48T08_STARTUPDATE { /* Start Updating */ \
	TOD_WRUNLOCK \
	TOD_MK48T08_CNTRL &= ~(TOD_MK48T08_WR_BIT | TOD_MK48T08_RD_BIT); \
	TOD_WRLOCK \
}

	/****************************************************************/
	/*								*/
	/* Device Read/Write MACROs					*/
	/* Assuming the External registers are stopped for Rd/Wr;	*/
	/* otherwise, updated continuously.				*/
	/*								*/
	/* Offsets:				Arrays [0..6], [0..11]	*/
	/*								*/
	/* The Time-Of-Day devices see the DayOfWeek as 1-7 (Sun-Sat)	*/
	/* The program wants DayOfWeek as 0-6 (Sun-Sat)			*/
	/*								*/
	/* The Time-Of-Day devices see the Months as 1-12 (Jan-Dec)	*/
	/* The program wants Months as 0-11 (Jan-Dec)			*/
	/*								*/

#define TOD_MK48T08_RDTIME(data) { \
	TOD_MK48T08_ON /* Make sure RTC is running */ \
	TOD_MK48T08_RDSTOPUPDATE  \
	(data)->tm_sec = TOD_BCDtoBIN(TOD_MK48T08_SEC & 0x7f); \
	(data)->tm_min = TOD_BCDtoBIN(TOD_MK48T08_MIN & 0x7f); \
	(data)->tm_hour = TOD_BCDtoBIN(TOD_MK48T08_HRS & 0x3f); \
	(data)->tm_wday = (TOD_BCDtoBIN(TOD_MK48T08_DAY & 0x07) - 1) & 0x07; \
	(data)->tm_mday = TOD_BCDtoBIN(TOD_MK48T08_DAYDATE & 0x3f); \
	(data)->tm_mon = (TOD_BCDtoBIN(TOD_MK48T08_MODATE & 0x1f) - 1) & 0x1f; \
	(data)->tm_year = TOD_BCDtoBIN(TOD_MK48T08_YRDATE); \
	TOD_MK48T08_STARTUPDATE  /* Restart clock updating */ \
}

#define TOD_MK48T08_WRTIME(data) { \
	TOD_MK48T08_ON  /* Make sure RTC is running */ \
	TOD_MK48T08_WRSTOPUPDATE \
	TOD_WRUNLOCK \
	TOD_MK48T08_SEC = TOD_BintoBCD((data)->tm_sec) & 0x7f; \
	TOD_MK48T08_MIN = TOD_BintoBCD((data)->tm_min) & 0x7f; \
	TOD_MK48T08_HRS = TOD_BintoBCD((data)->tm_hour) & 0x3f; \
	TOD_MK48T08_DAY = TOD_BintoBCD((data)->tm_wday + 1 ) & 0x07; \
	TOD_MK48T08_DAYDATE = TOD_BintoBCD((data)->tm_mday) & 0x3f; \
	TOD_MK48T08_MODATE = TOD_BintoBCD((data)->tm_mon + 1) & 0x1f; \
	TOD_MK48T08_YRDATE = TOD_BintoBCD((data)->tm_year % 100); \
	TOD_WRLOCK \
	TOD_MK48T08_STARTUPDATE  /* Restart clock updating */ \
}

/************************************************************************/
/*									*/
/* Generic Macro Names							*/
/*									*/

#define TOD_DEVICE_ON		TOD_MK48T08_ON
#define TOD_DEVICE_OFF		TOD_MK48T08_OFF
#define TOD_READ_DEVICE		TOD_MK48T08_RDTIME
#define TOD_WRITE_DEVICE	TOD_MK48T08_WRTIME

#endif /* INCtod_mk48t08 */

/* file: tod_mk48t08.h */
