/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
/*****************************************************************************
 * phatdaq - Phobos Data Acquisition program
 *
 * DESCRIPTION
 *
The phatdaq is the socket-based data taking program, it comprises:

Data server:    for taking data from Read Out Controllers (ROC) using single 
                UDP socket. 
Command server: for receiving commands from an interactive Phat (using TCP 
                socket) and distributing them to the ROCs via the UDP 
                socket. 

The UDP communication is secured by simple handshaking, when recipient confirms 
each message. Using UDP protocol make it possible to significantly reduce 
overhead (3 times in comparison with TCP), avoid data splitting between packets 
and to organize well-defined handshake.
 *
 * EXAMPLE:
 *
 *     To run this phatdaq task from phobosx.chm.bnl.gov do the following: 
 *     % phatdaq 7000 256000 256000
 *
 *     Start the simulated client:
 *     % /export/home/phobos/vxworks/networking>uxudpc phobosx.chm.bnl.gov 7000 0 51200 51200
 *
 *     To control phatdaq:
 *     - start phat or root
 *     - .x phatonl_init.C
 *     - .x phatonl_command.C(2)   (start the run)
 *     - .x phatonl_command.C(16)  (get status)
 *     - .x phatonl_command.C(0)   (stop the run)
 *     - .x phatonl_exit.C         (exit phatdaq and all ROCs)
 *     
 */
/*
modification history
--------------------
02a,24mar99,as   UDP data server, TCP command server, handshaking, 
                 selfconfiguration and many more.
01a,04mar99,as   created, based on uxBlastee.c by Wind River Systems
                 modifications: interval timer instead of alarm
                 sigaction (more standard) instead of signal
                 It was problems using signal, it worked as one shot.
*/
 
// System stuff
#include <stdlib.h>
#include <unistd.h>
//#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
//#include <netdb.h>
#include <arpa/inet.h> //for inet_ntoa
//#include <sys/times.h>
//#include <signal.h>
 
// ROOT stuff
#include "TROOT.h"
#include "TObjArray.h"
#include "TList.h"
#include "TFile.h"
#include "TMessage.h"
#include "TTree.h"

/*
#include "TRandom.h"
#include "TTree.h"
#include "TBranch.h"
#include "TClonesArray.h"
#include "TStopwatch.h"
*/
// Phat stuff
#include "TPhEvent.h"
#include "TPhHashContainer.h"
#include "TPhRawHitFECs0.h"

// phatdaq stuff
#include "daqmessages.h"
#include "daq.h"
#include "TPhRawHitTest.h"

#define TRUE 1
#define FALSE 0

/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
// Globals

int	  sockdat= -1, socksrv= -1, sockcmd= -1, sockdst= -1;
struct DAQMSG    *status_msg;
struct DAQSTATUS daqstatus;
#define   DS_PO (daqstatus.control & CONTROL_PRINTOUT)
TClientArray *clients;
TList     *freeq;
int       maxclients = 1;  // Change it to maximum number of ROCs to serve
//map//TMapFile* mfile;
TFile     *tfile = 0;
TPhEvent  *pevent;
TPhEvent  *sorevent = 0;
TPhEvent  *eorevent = 0;
//TDatime *datime = 0;
TTree     *tree = 0;
TMessage  smess(kMESS_OBJECT);
UInt_t    request_for_event = 0;
struct    fd_set readFds;
struct    fd_set writeFds;
struct    fd_set saveFds;
struct    fd_set saveWFds;
int       widthfds = 0;
char*     roc_type[TYPE_MAX+1] = {"Test",
				  "Trigger",
				  "Silicon",
				  "TOF"};

/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
void dumpbuf(char* buf, int bytes)
{
  cout.flush();
  int i;
  for(i=0; i < bytes; i++)
    {
      if((i % 32) == 0)
	printf("\n%4.4x:",i);
      printf("%2.2x",(UChar_t)(*buf++));
      if((i % 4) == 3)
        printf(" ");
    }
  printf("\n");
}
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
int sendmess(int sock, TMessage *pmess)
{
  // Send TMessage using system socket
  // Do as in TSocket::Send(const TMessage &mess)
  //
  char *buf = pmess->Buffer();
  char *savebuf = buf;
  Int_t ilen = pmess->Length();
  //cout<<"ilen="<<ilen<<" tobuf="<<(UInt_t)(ilen - sizeof(UInt_t))<<endl;
  tobuf(buf, (UInt_t)(ilen - sizeof(UInt_t))); // it advances the buf pointer
  ilen = send(sock,savebuf,ilen,0);
  if(ilen == -1 )
    perror("\nError during sending an event");
  //cout<<"message["<<ilen<<"] sent:";cout.flush();
  return ilen;
}
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
void initfdset()
{
  FD_ZERO (&saveFds);    /* initialize the set - all bits off */
  FD_ZERO (&saveWFds);   
  widthfds = sockdat;
  if (sockdat != -1)
    FD_SET (sockdat, &saveFds); /* Turn on bit for sockdat */
  if (sockcmd != -1)
    FD_SET (sockcmd, &saveFds); /* Turn on bit for sockcmd */
  widthfds = (sockcmd > widthfds) ? sockcmd : widthfds;
  if (socksrv != -1)
    FD_SET (socksrv, &saveFds); /* Turn on bit for socksrv */
  widthfds = (socksrv > widthfds) ? socksrv : widthfds;
  widthfds++;
  if(DS_PO) {cout<<"sockdat="<<sockdat
                <<" sockcmd="<<sockcmd
                <<" socksrv="<<socksrv
		 <<" with="   <<widthfds<<endl; cout.flush();}
}
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
int preveventn = 0;
int checkeventformat(struct DAQEVENT* ev, int nwords)
{
  int *iptr, ecode, sample, *evend, headerlength;

  headerlength = ev->headersize/sizeof(int) ;
  iptr = (int*)ev + headerlength;
  evend = (int*)ev + nwords;
  //dumpbuf((char*)ev, 16*4);
  if(daqstatus.control & CONTROL_EVENT_FORMAT_CHECK)
    {
      if(ev->testword !=  0x12345678)  /*major format error*/
	{
	  printf("wrong event[%d]: %8.8x, should be 12345678\n",
                  &(ev->testword) - (unsigned long*)ev, ev->testword);
	  return 1;
	}
      /* 
	 if (nwords != ev->bytes*sizeof(int))
	 {
	 printf("wrong size %x, should be %x\n",
	 numRead,ev->bytes*sizeof(int));
	 return 1;
	 }*/
      if(preveventn)
        {
	  if ((ecode = (preveventn+1) - ev->seqn))
            {
	      printf("wrong evn: %8.8x %8.8x\n",
		     ev->seqn,preveventn+1);
	      preveventn = ev->seqn;
	      return ecode;
	    }
	}
      preveventn = ev->seqn;
      if( (headerlength < 0) || headerlength > 10)
	{printf("Wrong header length, %8.8x\n",headerlength);         return 3;}
      if (iptr > evend)
	{printf("Wrong header length, %8.8x\n",headerlength);         return 3;}
    }
  if(daqstatus.control & CONTROL_EVENT_SPECIAL_CHECK)
    {
      sample = ev->seqn;
      while(iptr < evend - 1)
        {
	  if( *iptr++ != sample)
	    {
	      printf("wrong data %8.8x: %8.8x should be %8.8x\n",
		     iptr - 1 - (int*)ev, iptr[-1],sample);
              dumpbuf((char*)ev, 16*4);
	      return 2;
	    }
	  if( *iptr++ != ~sample)
	    {
	      printf("wrong data %8.8x: %8.8x should be %8.8x\n",
		     iptr - 1 - (int*)ev, iptr[-1],sample);
              dumpbuf((char*)ev, 16*4);
	      return 2;
	    }
	  sample++;
	}
    } 
  return 0;
}
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
int handle_command_message()
{
  // Handle the message in the command socket
  // 
  struct DAQMSG cmd;
  struct DAQMSG response;
  int len;
  int goon = 1;
  char filename[80];
  TBranch* branch = 0;

  len = recv(sockcmd, (char*)&cmd, sizeof(struct DAQMSG), 0);
  if(len==0)
    {
      perror("command pipe is broken\n");
      // this should not happen                  
      close (sockcmd); sockcmd = -1;
      exit(1);
      //goto again1;
    }
  if(DS_PO) cout<<"Command received, size "<<len<<endl;
  if(DS_PO) dumpbuf((char*)&cmd,(len < 8*4) ? len : 8*4);
  response.data[0] = 1; //Unknown command
  response.command = cmd.command | RESPONSE;
  char *ptr = (char*)&response;
  int sendsize = sizeof(struct DAQMSG);
  //
  // Execute the command
  //
  switch (cmd.command)
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
    {
    case DAQMSG_START:

      if(daqstatus.clients < maxclients)
	{
	  cout<<"Error: Only "<<daqstatus.clients<<" clients out of "
	      <<maxclients<<" are connected\n";
	  response.data[0] = 2; //Error
          break;
	}
      if(daqstatus.state != STATE_STOPPED)
	{
	  cout<<"run is already in progress\n"; 
	  response.data[0] = 2; 
	  break;
	}
      daqstatus.runnum++;
      daqstatus.bytes     = 0;
      daqstatus.recv_bytes     = 0;
      daqstatus.transfers = 0;
      daqstatus.messages  = 0;
      daqstatus.events    = 0;
      daqstatus.errors    = 0;
      daqstatus.lost_events = 0;
      daqstatus.control = cmd.data[0];
      daqstatus.analized_events = 0;
      cmd.command = DAQMSG_START;
      cmd.seqn = 0;
      cmd.data[0] = daqstatus.runnum;
      cout<<"Starting run "<<daqstatus.runnum<<endl; 
      if(daqstatus.control & CONTROL_EVENT_WRITING)
	{
	  sprintf(filename,"daq%d.root",daqstatus.runnum);
	  tfile = new TFile(filename, "RECREATE","PHOBOS DAQ file 990428");
	  //tfile = new TFile(filename, "RECREATE");
	  if (tfile)
	    {
	      //tfile->cd(); not necessary
	      cout<<"File opened:"<<tfile->GetPath()<<endl;
	      tfile->SetCompressionLevel(0);
	      // Create the tree
	      tree = new TTree("phatdaq_tree","PHOBOS DAQ tree 990428");
	      // Keeps as many buffers in memory until
	      // the sum of all buffers is greater than following number
	      tree->SetMaxVirtualSize(64000000); // it seems it does not work
	      tree->SetAutoSave(1000000000);  // autosave when 1 Gbyte written
	      // Create one branch. If splitlevel is set, event is a superbranch
	      // creating a sub branch for each data member of the Event object.
	      // The basket size should be close to maximum event size. The
	      // events will be collected in the basket until it is full, then 
	      // the basket will be written to the disk. Strange, but with big 
	      // basket the writing speed is dropping a little while the 
	      // activity lamp on the disk flushes more seldom 
	      branch = tree->Branch("Event", TPhEvent::Class()->GetName(),
		&pevent,1024*64, 0);
              cout<<"Tree "<<tree->GetName()
		  <<" created, maxvirtualsize "<<tree->GetMaxVirtualSize()
		  <<", basketsize "<<branch->GetBasketSize()
		  <<", compression level "<<branch->GetCompressionLevel()
		  <<endl;
   	    }
	  else
	    {
	      cout<<"Error. could not open file "<<filename<<endl;
	      response.data[0] = 2; 
	      break;
	    }
	}
      else if(tfile)
	{
	  cout<<"File closed:"<<tfile->GetPath()<<endl;
	  tfile->Close();
	  tfile = 0;
	}
      response.data[0] = clients->Send_to_all(&cmd, sizeof(struct DAQMSG));
      if(sockdst != -1)
	{
	  cout<<"sending SOR to dst...";cout.flush();
	  smess.Reset();
	  smess.WriteObject(sorevent);
	  len = sendmess(sockdst,&smess);
	  if(len == -1 )
	    {cout<<"\nerror in sending SOR to dst\n";cout.flush();
	    goon = 0; break;}
	  smess.Reset();
	  cout<<" Sent "<<len<<" Done.\n";cout.flush();
	}
      //datime->Set();
      //daqstatus.started = datime->Get();
      daqstatus.started = (UInt_t)time(0);
      daqstatus.state = STATE_RESUMED;
      break;
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////

    case DAQMSG_STOP:

      cout<<"Stopping...\n"; 
      cmd.command = DAQMSG_STOP;
      if(tfile)
	{
	  if (tree)
	    tree->Print();
	  cout<<"File closed:"<<tfile->GetPath()<<endl;
	  // Save all objects in this file	  
	  tfile->Write();
	  tfile->Close();
          delete tfile;
          tfile = 0;
	}
      response.data[0] = clients->Send_to_all(&cmd, sizeof(struct DAQMSG));
      if(sockdst != -1)
	{
	  cout<<"sending EOR to dst...";cout.flush();
	  smess.Reset();
	  smess.WriteObject(eorevent);
	  len = sendmess(sockdst,&smess);
	  smess.Reset();
	  cout<<" sent "<<len<<". Done.\n";cout.flush();
	}
      //datime->Set();
      //daqstatus.stopped = datime->Get();
      daqstatus.stopped = (UInt_t)time(0);
      daqstatus.state = STATE_STOPPED;
      break;
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////

    case DAQMSG_SUSPEND:

      if(daqstatus.state < STATE_RUNNING)
	{
	  cout<<"Run is not running\n";
          response.data[0] = 2;
	}
      else
	{
	  cout<<"Run suspended...\n";
	  daqstatus.state = STATE_SUSPENDED;
          response.data[0] = 0;
	}
      break;

    case DAQMSG_RESUME:

      if(daqstatus.state < STATE_RUNNING)
	{
	  cout<<"Run is not running\n";
          response.data[0] = 2;
	}
      else
	{
	  cout<<"Run resuming...\n";; 
	  daqstatus.state = STATE_RESUMED;      
          response.data[0] = 0;
          clients->ResumeAll(); // does not work
	  cout<<"Run resumed\n";; 
	}
      break;
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////

    case DAQMSG_STATUS:

      memcpy((char*)(status_msg->data + 1), (char*)&daqstatus, 
              sizeof (struct DAQSTATUS));
      ptr = (char*)status_msg;
      status_msg->data[0] = 0;
      sendsize = sizeof(struct DAQMSG) + sizeof(struct DAQSTATUS);
      cout<<"Status:";
      dumpbuf((char*)ptr,sendsize);
      break;

    case DAQMSG_EXIT:

      goon = 0;
      cout<<"Exitting...\n"; 
      cmd.command = DAQMSG_EXIT;
      response.data[0] = clients->Send_to_all(&cmd, sizeof(struct DAQMSG));
      break;
    }
  len = send(sockcmd, ptr, sendsize, 0);
  //continue;
  return goon;
}
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
int handle_dst_message()
{
  static char cmd[256];
  int ilen = recv(sockdst, cmd, sizeof(cmd), 0);
  if(ilen==0) return -1;
  //cout<<"Message["<<ilen<<"]\n";
  request_for_event++;
  // Test only. Return message back
  //len = send(sockdst, cmd, len, 0);
  //if(ilen==0) return -1;
  return 1;
}
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
int add_subevent_test(struct DAQEVENT* ev)
{
  TPhRawHitTest *pTPhRawHitTest; 
  //
  // The following section is for 'simple' event, without hash table
  /*
  pTPhRawHitTest = new TPhRawHitTest((Int_t*)&(ev->data), 
                       (Int_t)ev->bytes/sizeof(long));
  pevent->AddObject(pTPhRawHitTest);
  */
  // The following section is for 'standard' event - with 8-entry hashtable
  /**/ 
  TPhHashContainer *hash = NULL;
  hash = new TPhHashContainer("ZSS");
  Int_t nobj = 64;
  pevent->AddObject(hash);
  for(int ii=0; ii < nobj; ii++)
    {
      pTPhRawHitTest = 
	new TPhRawHitTest((Int_t*)&(ev->data),(Int_t)ev->bytes/sizeof(long)/nobj);
      hash->Add((TPhDataObject*)pTPhRawHitTest);
    }
  /**/
  return 0;
}
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
int add_subevent_silicon(struct DAQEVENT* ev)
{
  UChar_t *Bytes, *bufend;
  UChar_t *ZSSArr;
  Int_t n=0, i=0;

  //TFile f("zss.root", "RECREATE");
  //printf("tphevent:\n");
  //dumpbuf((char*)buf,8*4);
  TPhHashContainer *hash = NULL;
  TPhRawHitFECs0 *rawhits = NULL;  
  Int_t event_num = 0, old_event_num = 0;
  //printf("constructing hash\n");
  hash = new TPhHashContainer("ZSS");
  pevent->AddObject(hash);
  UChar_t *buf = (UChar_t *)ev->data;
  bufend = buf + ev->bytes - sizeof(ev);
  ZSSArr = buf;

  for(Bytes = buf; Bytes < bufend; Bytes +=3)
  {
    if((Bytes[0] & 0xc0) == 0xc0) {
      if(rawhits) { 
        //dumpbuf((char*)ZSSArr,8*4);
	//old//rawhits->UnPack(ZSSArr, 3*n);
        rawhits->SetZSSArray(ZSSArr, 3*n);
	//printf("This was event # %d, FEC# %d\n", rawhits->GetInfoWord(),
	//       rawhits->GetFECNum());
      }
      //event_num = (Bytes[1] & 0xff << 8) + Bytes[2] & 0xff;
      rawhits = new TPhRawHitFECs0(Bytes[0] & 0x3f);
      hash->Add((TPhDataObject*)rawhits);
      n = 0;
      ZSSArr = Bytes;
     }
    n++;
  }
  return 0;
}
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
Int_t TDAQClient::fsomething_suspended = 0;
Int_t   TDAQClient::NextEvent(struct DAQEVENT** ev, Int_t* size, Int_t* next)
{
  //getEventAgain:
  if (fevend)
      fevstart  = (struct DAQEVENT*)fevend;
  else
    {
      if (frecvbuf)
	freeq->AddLast((TDAQRecvBuffer*)frecvbuf);
      frecvbuf = (TDAQRecvBuffer*)fqueue->Remove(fqueue->FirstLink());
      if( frecvbuf == 0 )
	{ // should not happen
	  cout<<"Error. Client queue is empty for client"<<(int)ftype;
	  return -1;
	}
      fbufstart = (long*)frecvbuf->GetArray();
      fbufend   = fbufstart + fbufstart[0]/sizeof(long);
      fevstart  = (struct DAQEVENT*)(fbufstart + sizeof(struct DAQMSG)/sizeof(long) - 1);
    }
  /* This shouldn't happen as we always check for nextv
    if((fevstart >= fbufend) || (*fevstart == 0))
    {
      // no more events
      ob = fqueue->Remove(fqueue->FirstLink());
      freeq->AddLast(ob);
      fevend = 0;
      goto getEventAgain;
    }
    */
  //dumpbuf((char*)fbufstart, 32);
  fevend    = (long*)fevstart + fevstart->bytes/sizeof(long);
  if(fevend > fbufend)
    fevend = fbufend;
  *size = (fevend - (long*)fevstart)*sizeof(long);
  *ev  = fevstart;
  //printf("fbufstart...%8.8x %8.8x %8.8x %8.8x\n",fbufstart,fbufend,fevstart,fevend);
  //printf("fevstart...%8.8x %8.8x\n",fevstart,fevstart->bytes);
  if(fevend < fbufend)
    *next = *fevend;
  else
    *next = 0;
  //printf("size %8.8x next %8.8x queue size %8.8x\n",*size,*next,fqueue->GetSize());
  if(*next == 0)
    {
      fevend = 0;
      *next = -(fqueue->GetSize());
      //There is no more events in the current buffer but there 
      //are more buffers
    }
  return fevstart->seqn;
}
Int_t TDAQClient::Handshake(struct DAQMSG *msg)
{
  int hssent;
  struct DAQMSG response;
  
  if(msg->seqn)
  if(msg->seqn != fmsg_received + 1)
    printf("Lost message for client%d, expected %8.8x, got %8.8x\n",
	   ftype, fmsg_received + 1, msg->seqn);
  fmsg_received     = msg->seqn;
  response.bytes    = sizeof (struct DAQMSG);
  // 
  // By now we will respond with current event number.
  // In future, when event can span several transfer it should be changed.
  //response.seqn     = fmsg_confirmed++;
  response.seqn     = daqstatus.events;
  response.command  = msg->command | RESPONSE;
  response.source   = SRC_SERVER;
  response.source_type   = 0;
  response.data[0]  = 0;
  if ((hssent = sendto (sockdat, (char*) &response, 
			response.bytes, 0,
			(struct sockaddr *) &fSockaddr, 
			sizeof(struct sockaddr_in))
       ) < 0)
    {
      perror ("phatdaq handshake sending error");
      return -1;
    }
  return 0;
}
Int_t TClientArray::Send_to_all(struct DAQMSG *msg, int len)
{
  int bytes = 0;
  TDAQClient *pclient;

  msg->source = SRC_SERVER;
  for(int ii = 0; ii < daqstatus.clients; ii++)
    {
      if(DS_PO) cout<<"Send to "<<ii<<endl;
      if(DS_PO) dumpbuf((char*)msg,(len < 8*4) ? len : 8*4);

      pclient = (TDAQClient*)At(ii);
      if ((bytes = sendto (sockdat, (char*)msg, len, 0, 
			   (struct sockaddr *) pclient->Sockaddr(), 
			   sizeof(struct sockaddr_in))) < 0)
	{
	  perror ("phatdaq send error");
	  return 3;
	}
      //TODO wait/check for response
    }
  return 0;
}
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
int process_subevents()
{
  //TPhHashContainer *hash = NULL;
  TDAQClient* daq_client;
  struct DAQEVENT* ev = 0;
  Int_t eventn=0;
  Int_t nextevent=1;
  Int_t size=0;
  int ecode = 0;
  int ii, len=0;

  if(DS_PO) cout<<"Constructing event\n";
  pevent = new TPhEvent(0, daqstatus.events);
  while(ecode == 0)
    {
      for(ii=0; ii < daqstatus.clients ; ii++)
	{
	  daq_client = (TDAQClient*)clients->At(ii);
	  eventn = daq_client->NextEvent(&ev,&size,&nextevent);
	  if(DS_PO) cout<<"GetEvent "<<eventn<<", next "<<nextevent<<endl;
	  if(eventn != daqstatus.events)
	    {
	      cout<<"Event mismatch. Client"<<daq_client->Type()
                  <<", event="<<eventn<<", expected="<<daqstatus.events<<endl;
	      daqstatus.events = eventn+1;
	      ecode= -1;
	      break;
	    }
	  // As we are here then we are sure that we will have all subevents 
	  // available. Othervise we would get nextevent = 0 before and returned.
	  //
	  // If TPhEvent construction is enabled then add subevent to event
	  // else write the subeven to disk
	  //
	  if((daqstatus.control & CONTROL_NOTPHEVENT) == 0)
	    {
	      switch (daq_client->Type())
		{
		case TYPE_TEST:
		  if(DS_PO) cout<<"add_subevent_test\n";cout.flush();
		  ecode = add_subevent_test(ev);
		  break;
		case TYPE_TRIGGER:
		  if(DS_PO) cout<<"decoding Trigger\n";
		  ecode = 0;
		  break;
		case TYPE_SILICON:
		  if(DS_PO) cout<<"add_subevent_silicon\n";
		  ecode = add_subevent_silicon(ev);
		  break;
		case TYPE_TOF:
		  if(DS_PO) cout<<"decoding TOF\n";
		  ecode = 0;
		  break;
		default:
		  cout<<"Error: Unrecognized client type\n";
		  ecode = -1;
		}
	    }
	  else
	    {
	      if(tfile)
		{
		  tfile->WriteBuffer((char*)ev->data,(int)ev->bytes);
		}
	    }
	  if(nextevent == 0)
	    {
	      // all events from that client had been processed, 
              // move them all to the free queue
	      // and continue
	      daq_client->MoveAllBuffersTo(freeq);
	      ecode++;
	    }
	}
      daqstatus.events = eventn+1;
    }
  if(ecode < 0)
    {
      // Severe error, release all client buffers
      clients->MoveAllBuffersTo(freeq);
      // The message was printed. So far it is enough.
    }
  else if (ecode > 0)
    {
      // At least one client is empty
      // Suspend all non-empty clients.
      clients->SuspendAllNonEmpty();
    }
  //
  // Event have been constructed, write and distribute it if necessary.
  //
  if(ecode >= 0)
  if((daqstatus.control & CONTROL_NOTPHEVENT) == 0)
    {
      if(DS_PO) 
	pevent->ls("d");
      if(tfile)
	{
	  if(DS_PO) cout<<"Writing event "<<daqstatus.events<<", size "
              <<", total "<<daqstatus.bytes<<endl;
	  //tfile->cd(); //not necessary
          if (daqstatus.control & CONTROL_WRITING_TREE)
	    {
	      //cout<<"tree->Fill()=";cout.flush();
	      len = tree->Fill();
	      //cout<<len<<endl;
	    }
	  else
	    pevent->Write();
	}
         
      if(request_for_event)
      if(sockdst != -1)
	{
	  if(DS_PO) 
	    {cout<<"Distributing "<<pevent->GetName()<<" #"<<daqstatus.events
              <<", total "<<daqstatus.bytes
	      <<", requested: "<<request_for_event;
	    cout.flush();}
	  smess.Reset();
	  smess.WriteObject(pevent);
          len = sendmess(sockdst,&smess);
	  if(DS_PO) cout<<", sent "<<len<<" bytes\n";
	  daqstatus.analized_events++;
	  if(request_for_event)	  
	    request_for_event--;
	}
    }
  if(tfile)
    daqstatus.bytes = (ULong_t)tfile->GetBytesWritten();
  //cout<<"Deleting event...\n";cout.flush();
  delete pevent;
  return ecode;
}
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
int subevents_filled()
{
  TDAQClient* daq_client;
  int n = 0;
  for(int ii=0; ii < daqstatus.clients; ii++)
    {
      daq_client = (TDAQClient*)clients->At(ii);
      if (daq_client->NBuffers())
	n++;
    }
  return n;
}
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////

int main(int argc, char **argv)
{
  struct sockaddr_in  serverAddr; /* server's address */
  struct sockaddr_in  clientAddr; /* client's address */
  struct sockaddr_in  dstclientAddr; /* dst client's address */
  struct sockaddr_in  peer;
  int                 peerlen = sizeof(peer);
  int    sockaddrsize = sizeof(struct sockaddr_in);
  u_long saddr=0;
  char   *buffer = 0;
  int	 size = 65536;
  int	 blen = 65536;
  int    port = 7000;
  //struct itimerval oval;
  int    getval;
  int    lenval = sizeof(int);
  Int_t  len;

  /* Our Socket options*/
  int                     opt_linger=1;
  int                     opt_tcp=0;
  int                     opt_reuse=1;
  int                     optenable=1;

  struct timeval timeout;
  int numFds;
  int ii;

  TROOT simple("simple","Phobos Event Builder");

  cout.setf(ios::hex);
  cout.width(8);

  freeq   = new TList();

  //TDAQClientKey client_key = TDAQClientKey(0);
  TDAQClient *daq_client;

  if (argc < 2)
    {
      cout<<"usage: "<<argv [0]<<" #clients [port] [size] [bufLen]\n";
      exit (1);
    }

  maxclients = atoi(argv[1]);
  clients = new TClientArray(maxclients);
  if(argc > 2) port = atoi (argv [2]);
  if(argc > 3) {size = atoi (argv [3]); blen = size;}
  if(argc > 4) blen = atoi (argv [4]);
    
  cout<<"#clients "       <<maxclients
      <<", data port "    <<port
      <<", command port " <<port+1<<endl;

  //datime = new TDatime();

  memset (&daqstatus, 0, sizeof (struct DAQSTATUS));
  daqstatus.control = CONTROL_EVENT_FORMAT_CHECK | 
                      CONTROL_EVENT_DISTRIBUTION;
  //
  // Allocate free queue
  //
  //buffer = (char *) malloc (size);
  TDAQRecvBuffer* daqrecvbuffer;
  //for(ii = 0; ii < (maxclients*2 + 1); ii++)
  for(ii = 0; ii < 10; ii++)
    {
      daqrecvbuffer = new TDAQRecvBuffer(size);
      if (!daqrecvbuffer)
	{perror ("cannot create TDAQRecvBuffer"); exit (1);}
      freeq->AddLast(daqrecvbuffer);
    }
  cout<<"Number of buffers "<<freeq->GetSize()<<endl;
  status_msg = (struct DAQMSG*) malloc( sizeof(struct DAQMSG) + 
                sizeof(struct DAQSTATUS));
  if (status_msg == NULL)
    {
      perror ("cannot allocate status_msg");
      exit (1);
    }
  status_msg->bytes   = sizeof(struct DAQMSG) + sizeof(struct DAQSTATUS);
  status_msg->command = DAQMSG_STATUS | RESPONSE;
  status_msg->source  = SRC_SERVER;
 
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
// Open the data socket. Use datagram sockets.
 
  // Zero out the sock_addr structures. 
  // This MUST be done before the socket calls.
  memset (&serverAddr, 0, sizeof (serverAddr));
  while(1)
    {
      sockdat = socket (AF_INET, SOCK_DGRAM, 0);
      if (sockdat == -1)
	{perror ("cannot open data socket"); exit (1);}
      // Set up our internet address, and bind it so the client can connect.  
      serverAddr.sin_family	= AF_INET;
      serverAddr.sin_port	= htons (port);
      serverAddr.sin_addr.s_addr = htonl (INADDR_ANY);
      if (bind (sockdat,(struct sockaddr*) &serverAddr,sizeof(serverAddr)) >= 0)
	break;
      perror ("bind error UDP");
      //exit (1);
      close(sockdat); sockdat = -1;
      sleep(1);
    }
  /* maximum size of socket-level receive buffer */
  if (setsockopt(sockdat,SOL_SOCKET,SO_RCVBUF,(char*)&blen,sizeof(blen)) < 0)
    {
      perror ("setsockopt SO_RCVBUF failed");
      free (buffer);
      exit (1);
    }
  setsockopt(sockdat,SOL_SOCKET,SO_LINGER,
	     (char*)&opt_linger,sizeof(opt_linger));
  setsockopt(sockdat,SOL_SOCKET,SO_REUSEADDR,
	     (char*)&opt_reuse,sizeof(opt_reuse));
  //setsockopt(sockdat,IPPROTO_TCP,TCP_NODELAY,(char*)&opt_tcp,sizeof(opt_tcp));
  //setsockopt(sockdat,SOL_SOCKET,SO_KEEPALIVE, &opt_enable,1) < 0);
  getsockopt (sockdat, SOL_SOCKET, SO_RCVBUF, (char *)&getval,&lenval);
  cout<<"data socket size "<<getval<<endl;

/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
// Open the command socket. Use stream sockets.

  memset (&serverAddr, 0, sizeof(serverAddr));
  serverAddr.sin_family	= AF_INET;
  serverAddr.sin_port	= htons (port + 1);
  while(1)
    {
      socksrv = socket (AF_INET, SOCK_STREAM, 0);
      if (socksrv == -1)
	{perror ("cannot open command socket"); exit (1);}
      // Important! One must set SO_REUSEADDR for every server stream socket to 
      // prevent bind error: "Adress already in use".  
      setsockopt(socksrv,SOL_SOCKET,SO_REUSEADDR,
		 (char*)&opt_reuse,sizeof(opt_reuse));
      if (bind(socksrv,(struct sockaddr*) &serverAddr, sizeof(serverAddr)) >= 0)
	break;
      printf("Error port %d=0x%x\n",serverAddr.sin_port,serverAddr.sin_port); 
      perror ("");
      close(socksrv); socksrv = -1;
      sleep(1);
    }
  /* Listen, for the client to connect to us. */
  cout<<"Listening command port "<<serverAddr.sin_port<<"...";cout.flush();
  if (listen (socksrv, 1) < 0)
    {perror ("listen failed"); exit (1);}
  cout<<" Done. Accepting...\n";cout.flush();
  len = sizeof (clientAddr);
  memset (&clientAddr, 0, len);
  sockcmd = accept (socksrv, (struct sockaddr *) &clientAddr, &len);
  if (sockcmd == -1)
    {cout<<"accept failed"<<endl; cout.flush(); close (sockcmd);exit(1);}
  if (getpeername(sockcmd, (struct sockaddr *) &peer, &peerlen) < 0) 
    perror("Accepted connection from Unidentified Client");
  else
    cout<<"Accepted command connection to "<<inet_ntoa(peer.sin_addr)<<endl;
  setsockopt(sockcmd,SOL_SOCKET,SO_LINGER,
	     (char*) &opt_linger,sizeof(opt_linger));
  setsockopt(sockcmd,SOL_SOCKET,SO_REUSEADDR,
	     (char*)&opt_reuse,sizeof(opt_reuse));
  close (socksrv); socksrv = -1; // don't need it any longer

/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
// Open the distribution socket. Use stream sockets.

  memset (&serverAddr, 0, sizeof(serverAddr));
  serverAddr.sin_family	= AF_INET;
  serverAddr.sin_port	= htons (port + 2);
  while(1)
    {
      socksrv = socket (AF_INET, SOCK_STREAM, 0);
      if (socksrv == -1)
	{perror ("cannot open distribution's server socket "); exit (1);}
      setsockopt(socksrv,SOL_SOCKET,SO_REUSEADDR,
		 (char*)&opt_reuse,sizeof(opt_reuse));
      if (bind(socksrv,(struct sockaddr*) &serverAddr, sizeof(serverAddr)) >= 0)
	break;
      printf("Error, port %d=0x%x\n",serverAddr.sin_port,serverAddr.sin_port); 
      perror ("");
      /*exit (1);*/
      close(socksrv); socksrv = -1;
      sleep(1);
    }
  /* Listen, for the client to connect to us. */
  cout<<"Listening dst port "<<serverAddr.sin_port<<"...";cout.flush();
  if (listen (socksrv, 1) < 0)
    {perror ("listen failed"); exit (1);}
  cout<<" Done.\n";cout.flush();

/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
// simulate SOR and EOR events -- don't delete this event until next run start!

  sorevent = new TPhEvent(999,999);
  sorevent->SetName("SOR");
  eorevent = new TPhEvent(999,999);
  eorevent->SetName("EOR");

/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////

  daqstatus.bytes = 0;
  daqstatus.recv_bytes = 0;
  //daqstatus.control |= CONTROL_PRINTOUT; // for debugging
  timeout.tv_sec = 1;
  timeout.tv_usec = 0;
  int goon = 1;

  initfdset();
  // Create a new memory mapped file. The memory mapped file can be
  // opened in an other process on the same machine and the objects
  // stored in it can be accessed.
  //map//mfile = TMapFile::Create("data.map","RECREATE", 100000,
  //map//                          "memory mapped file with data");
  //map//pevent = new TPhEvent(0, 0);
  //map//mfile->Add(pevent);
  // Print status of mapped file
  //map//mfile->Print();
  
/////////1/////////2/////////3/////////4/////////5/////////6/////////7//////////
// Everything is set. Main loop

  while(goon)
    {
      int numRead=0;
      char *partbuf = buffer;
      int partsize = size;
      int partread;
      //
      // Message handling loop
      //
      while((partsize > 0) && goon )
	{
	  readFds = saveFds;
	  //writeFds = saveWFds;
          //
          // use select to check for pending messages
          // So far we do not use writing request.
	  //if ((numFds = select(width,&readFds,&writeFds,NULL,&timeout)) == -1)
	  if ((numFds = select(widthfds,&readFds,NULL,NULL,&timeout)) == -1)
	    {
	      perror (" ERROR in select");   
	      return (-1);
	    };
	  if (numFds == 0)
	    {
	      /* Timeout. Send waitingfor message to server */
	      //printf("timeout\n");
	      continue;
	    }
	  //cout<<"Number of pending hot sockets "<<numFds
          //    <<" out of "<<widthfds<<endl;
	  if (FD_ISSET (sockcmd, &readFds))
	    {
              goon = handle_command_message();
              continue;           
	    }
	  if (FD_ISSET (socksrv, &readFds))
	    {
	      cout<<"Request on socksrv. Accepting new connection...\n";
	      cout.flush();
              if(sockdst != -1)
		{
		  cout<<"Something wrong. Connection is already open.\n";
		  cout<<"As we can handle only one connection, we should "
                      <<"close old one\n";
		  close (sockdst); sockdst = -1;
		} 
	      len = sizeof (dstclientAddr);
	      memset (&dstclientAddr, 0, len);
	      sockdst = accept (socksrv,(struct sockaddr*)&dstclientAddr, &len);
	      if (sockdst == -1)
		{cout<<"accept failed"<<endl; cout.flush(); 
		close (sockcmd); sockcmd = -1;
		goon = 0; break;}
	      if (getpeername(sockdst,(struct sockaddr *) &peer, &peerlen) < 0) 
		perror("Accepted connection from Unidentified Client");
	      else
		cout<<"Accepted connection to "<<inet_ntoa(peer.sin_addr)<<endl;
	      setsockopt(sockdst,SOL_SOCKET,SO_LINGER,
			 (char*) &opt_linger,sizeof(opt_linger));
	      setsockopt(sockdst,SOL_SOCKET,SO_REUSEADDR,
			 (char*)&opt_reuse,sizeof(opt_reuse));
	      setsockopt(sockdst,SOL_SOCKET,SO_SNDBUF,
                         (char*)&blen,sizeof(blen));
	      getsockopt(sockdst,SOL_SOCKET,SO_SNDBUF,(char*)&getval,&lenval);
	      cout<<"dst socket size "<<getval<<endl;
              // Connection accepted, add new socket to select list.
	      //FD_CLR (socksrv, &saveFds);
	      FD_SET (sockdst, &saveFds);
	      //FD_SET (sockdst, &saveWFds);
	      widthfds = (sockdst+1 > widthfds) ? sockdst+1 : widthfds;
              cout<<"sockdst="<<sockdst<<" width="<<widthfds<<endl;
              continue;           
	    }
          if(sockdst != -1)
	    {
	      if (FD_ISSET (sockdst, &readFds))
		{
		  //cout<<"Request on sockdst. Handling message...\n";
		  //cout.flush();
		  goon = handle_dst_message();
		  if (goon == -1 )
		    {
		      cout<<"dst pipe is broken. You must restart dst client\n";
		      cout.flush();
		      // Remove socket from the select list
		      FD_CLR (sockdst, &saveFds);
		      FD_CLR (sockdst, &saveWFds);
		      close (sockdst);
		      sockdst = -1;
		      goon = 1;
		    }
		  //cout<<"...Done\n";cout.flush();
		  continue;           
		}
	      /*if (FD_ISSET (sockdst, &writeFds))
		{
		  static char testmess[20] = "Test\n";
		  len = send(sockdst,testmess,strlen(testmess),0);
		  cout<<"Writing request on sockdst, written "<<len<<endl;
		  //cout<<"Press enter"; cout.flush();
                  //static char tststr[10];
                  //gets(tststr);
		  continue;           
		}*/
	    }
          if (FD_ISSET (sockdat, &readFds))
	    {
	      // Handle the message in the data socket
	      //
              daqrecvbuffer = (TDAQRecvBuffer*)freeq->Remove(freeq->LastLink());
              if(!daqrecvbuffer)
		{ 
		  cout<<"Error, no free buffer left\n"; 
                  clients->PrintNBuffers();
                  clients->MoveAllBuffersTo(freeq);
		  //goon = 0; 
		  break;
		}
              partbuf  = (char*)daqrecvbuffer->GetArray();
              long * ipartbuf = (long*)partbuf;
              partsize = daqrecvbuffer->GetSize()*sizeof(Int_t);
              if(DS_PO) {cout<<"partbuf "<<(int*)partbuf
			     <<" partsize "<<partsize<<endl; cout.flush();}
	      if ((partread = recvfrom (sockdat, partbuf, partsize, 0, 
					(struct sockaddr *) &clientAddr, 
                                        &sockaddrsize)) < 0)
		{
		  perror ("phatdaq read error");
                  freeq->Add((TObject*)daqrecvbuffer);
		  break;
		}
	      if (partread == 0)
		{
		  cout<<"Error: pipe is broken\n";
		  //close (sockdat);
		  //goto again;
		}
	      daqstatus.transfers++;
	      saddr = clientAddr.sin_addr.s_addr;
	      struct DAQMSG *msg = (struct DAQMSG*)partbuf;

	      // check for new client
	      if( (msg->source >= maxclients) ||
		   (msg->source_type > TYPE_MAX))
		{
		  cout<<"Wrong client"<<(int)msg->source_type
                      <<" #"<<(int)msg->source
                      <<", from "<<inet_ntoa(clientAddr.sin_addr)
		      <<" rejected\n";
		  dumpbuf((char*)partbuf,(partread < 8*4)?partread:8*4);
                  freeq->Add((TObject*)daqrecvbuffer);
		  continue;
		} 
	      daq_client = (TDAQClient*)clients->At(msg->source);
	      if(msg->command != DAQMSG_WAITING_FOR)
		{
	      if(DS_PO) cout<<"freeq size "     <<freeq->GetSize()
                  <<", daqrecvbuffer "<<daqrecvbuffer<<endl;
		  if(DS_PO) cout<<"Data received from "<<inet_ntoa(clientAddr.sin_addr)
		      <<", "<<partread<<" bytes\n";
		  if(DS_PO) dumpbuf((char*)partbuf,(partread < 16*4)?partread:16*4);
		  if(DS_PO) cout<<"daq_client "<<daq_client<<endl;
		}
	      if( daq_client == 0 )
		{
		  if(DS_PO) cout<<"constructing the client\n";
		  daq_client = new TDAQClient(&clientAddr, msg->source_type);
		  //printf("adding the client %8.8x, source %d, type %d\n",
                  //        daq_client,msg->source,msg->source_type);
                  //daq_client->Dump();
		  clients->AddAt(daq_client,msg->source);
		  daqstatus.clients = clients->Items();
		  cout<<"Next ROC#"<<(int)msg->source
		      <<" out of "<<daqstatus.clients
                      <<" registered, '"<<roc_type[(int)daq_client->Type()]
		      <<"' at "<<daq_client->TextAddr();
		  if(daqstatus.clients < maxclients)
		    cout<<". Waiting for other "<<maxclients-daqstatus.clients;
		  else
		    cout<<". Ready to start";
		  cout<<endl;
		}

	      /* Handshake.*/


              if( ((msg->command == DAQMSG_1ST_EV_PACKET) || 
		   (msg->command == DAQMSG_EV_PACKET)))
		{
                  if(daqstatus.state < STATE_RESUMED)
		    {
		      // Run is suspended, release the buffer and don't respond
                      cout<<"state="<<daqstatus.state<<endl;
		      daq_client->Suspend();
		      daq_client->StoreMessage(msg);
		      continue;
		    }
		  if(daq_client->Suspended())
		    daq_client->StoreMessage(msg);
		  else if(daq_client->Handshake(msg))
		    {
		      cout<<"Wrong handshake\n";
		      freeq->Add((TObject*)daqrecvbuffer);
		      break;
		    }
		}
	      /*else
		{
		  daq_client->Handshake(msg);
		  freeq->Add((TObject*)daqrecvbuffer);
		  continue;
		}*/
	      numRead += partread;
	      daqstatus.recv_bytes += partread;
	      //if(!daqstatus.control)  // Nothing to do with the message
	      //break;
	      if(ipartbuf[7] !=  0x12345678)  /*major format error*/
		{
		  daqstatus.errors++;
                  freeq->Add((TObject*)daqrecvbuffer);
		  break;
		}
	      //partbuf += partread;
	      //partsize = ipartbuf[0] - numRead;
	      if(partsize<0)
		cout<<"crossed boundary\n";
              daq_client->AddBuffer((TObject*)daqrecvbuffer);
              int filled = subevents_filled();
              if(DS_PO) cout<<"subevents available "<<filled<<endl;
              if( filled == daqstatus.clients)
		{
                  clients->ResumeAll(); //EbyE//
                  if(DS_PO) cout<<"process_subevents...\n";
		  process_subevents();
		}
              else
		daq_client->Suspend();  //EbyE//
              if(freeq->GetSize() == 0)
		{
                  cout<<"Bad! lost all buffers\n";
                  clients->PrintNBuffers();
                  clients->MoveAllBuffersTo(freeq);
		}
	    } 
          else
	    {
	      cout<<"Something unexpected in the select\n"; cout.flush();
	      initfdset();
	    }
	}
    }
  clients->Delete();
  close (sockdat);
  close (sockcmd);
  close (socksrv);
  close (sockdst);
  smess.Close();
  delete sorevent;
  //tfile->Close();
  delete tfile;
  delete freeq;
    
  free (buffer);
  cout<<"phatdaq end.\n";
}

