/***************************** Copyright Notice ********************************
 *                                                                             *
 * Copyright (c) 1992 Universities Research Association, Inc.                  *
 *               All Rights Reserved.                                          *
 *                                                                             *
 ******************************************************************************/

/*++
 * FACILITY:	Utilities
 *
 * ABSTRACT:	Provide access to 680x0 control registers from C language.
 *
 * ENVIRONMENT:	GNU Assembler for 68020 or higher.
 *		VxWorks.
 *
 * AUTHOR:	David M. Berg, Creation date: 22-Jan-1992
 *
 * MODIFIED BY:
 *
 *	V01-000	DMB		David M. Berg			 22-Jan-1992
 *		Initial entry.
 *--
 ******************************************************************************/

/*++
 * FUNCTIONAL DESCRIPTION:
 *
 *	Get or Set the value of a control register, which is specified
 *	using the code as required by the 680x0 MOVEC instruction:
 *
 *	Hex	Control Register		      68020   68030   68040
 *	---	------- --------		      -----   -----   -----
 *	000	SFC   Source Function Code		X	X	X
 *	001	DFC   Destination Function Code		X	X	X
 *	800	USP   User Stack Pointer		X	X	X
 *	801	VBR   Vector Base Register		X	X	X
 *	002	CACR  Cache Control Register		X	X	X
 *	802	CAAR  Cache Address Register		X	X	-
 *	803	MSP   Master Stack Pointer		X	X	X
 *	804	ISP   Interrupt Stack Pointer		X	X	X
 *	003	TC    Translation Control Register	-	-	X
 *	004	ITT0  Instr. Transparent Trans. Reg. 0	-	-	X
 *	005	ITT1  Instr. Transparent Trans. Reg. 1	-	-	X
 *	006	DTT0  Data Transparent Trans. Reg. 0	-	-	X
 *	007	DTT1  Data Transparent Trans. Reg. 1	-	-	X
 *	805	MMUSR MMU Status Register		-	-	X
 *	806	URP   User Root Pointer			-	-	X
 *	807	SRP   Supervisor Root Pointer		-	-	X
 *	
 *
 * FORMAL PARAMETERS:	status = cpuRegsGet (cr, &crv)
 *
 *   status		Optional.  Output by value.
 *
 *   cr			Required.  Input  by value. Unsigned longword.
 *			Control register.  One of the codes indicated above.
 *
 *   crv		Required.  Output by reference. Unsigned longword.
 *			The value of the specified control register.
 *
 * FORMAL PARAMETERS:	status = cpuRegsSet (cr, crv)
 *
 *   status		Optional.  Output by value.
 *
 *   cr			Required.  Input  by value. Unsigned longword.
 *			Control register.  One of the codes indicated above.
 *
 *   crv		Required.  Input  by value. Unsigned longword.
 *			The value of the specified control register.
 *
 * IMPLICIT INPUTS:	None
 *
 * IMPLICIT OUTPUTS:	None
 *
 * REGISTER USAGE:
 *
 *   d0			Return status.
 *   d1			Control register.
 *   d2			Temporaries.
 *
 * SIDE EFFECTS:	None
 *
 *--
 ******************************************************************************/
/*
 *
 * Routine arguments.  These are offsets from A6 after a LINK A6 statement.
 */

arg_cr	= 0x08
arg_crv	= 0x0c

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

	.text

	.globl	_cpuRegsGet

_cpuRegsGet:

	link	a6, #0
	moveml	d1-d2, sp@-		/* save registers */

	movel	a6@(arg_cr), d1		/* control register */
	moveq	#7, d2			/* max valid code */
	btst	#11, d1
	bnes	get_cr_set2		/* 800 numbers */

get_cr_set1:
	cmpl	d2, d1
	jhi	get_fail
	movew	pc@(get_tbl1-.-2:b,d1:w:2), d2
	jmp	pc@(2,d2:w)

get_tbl1:
	.word	get_cr_sfc-get_tbl1
	.word	get_cr_dfc-get_tbl1
	.word	get_cr_cacr-get_tbl1
	.word	get_cr_tc-get_tbl1
	.word	get_cr_itt0-get_tbl1
	.word	get_cr_itt1-get_tbl1
	.word	get_cr_dtt0-get_tbl1
	.word	get_cr_dtt1-get_tbl1
	
get_cr_sfc:
	movec	sfc, d0		/* source function code */
	jra	get_done

get_cr_dfc:
	movec	dfc, d0		/* destination function code */
	jra	get_done

get_cr_cacr:
	movec	cacr, d0	/* cache control register */
	jra	get_done

get_cr_tc:
|	movec	tc, d0		/* translation control register */
	.word	0x4e7a, 0x0003
	jra	get_done

get_cr_itt0:
|	movec	itt0, d0	/* instr. transparent translation reg. 0 */
	.word	0x4e7a, 0x0004
	jra	get_done

get_cr_itt1:
|	movec	itt1, d0	/* instr. transparent translation reg. 1 */
	.word	0x4e7a, 0x0005
	jra	get_done

get_cr_dtt0:
|	movec	dtt0, d0	/* data transparent translation reg. 0 */
	.word	0x4e7a, 0x0006
	jra	get_done

get_cr_dtt1:
|	movec	dtt1, d0	/* data transparent translation reg. 1 */
	.word	0x4e7a, 0x0007
	jra	get_done

get_cr_set2:
	bclr	#11, d1
	cmpl	d2, d1
	bhis	get_fail
	movew	pc@(get_tbl2-.-2:b,d1:w:2), d2
	jmp	pc@(2,d2:w)

get_tbl2:
	.word	get_cr_usp-get_tbl2
	.word	get_cr_vbr-get_tbl2
	.word	get_cr_caar-get_tbl2
	.word	get_cr_msp-get_tbl2
	.word	get_cr_isp-get_tbl2
	.word	get_cr_mmusr-get_tbl2
	.word	get_cr_urp-get_tbl2
	.word	get_cr_srp-get_tbl2
	
get_cr_usp:
	movec	usp, d0		/* user stack pointer */
	jra	get_done

get_cr_vbr:
	movec	vbr, d0		/* vector base register */
	jra	get_done

get_cr_caar:
	movec	caar, d0	/* cache address register */
	jra	get_done

get_cr_msp:
	movec	msp, d0		/* master stack pointer */
	jra	get_done

get_cr_isp:
	movec	isp, d0		/* interrupt stack pointer */
	jra	get_done

get_cr_mmusr:
|	movec	mmusr, d0	/* MMU status register */
	.word	0x4e7a, 0x0805
	jra	get_done

get_cr_urp:
|	movec	urp, d0		/* user root pointer */
	.word	0x4e7a, 0x0806
	jra	get_done

get_cr_srp:
|	movec	srp, d0		/* supervisor root pointer */
	.word	0x4e7a, 0x0807
	jra	get_done

get_fail:
	moveq	#-1, d0			/* return get_failure status */
	jra	get_exit

get_done:
	movel	d0, a6@(arg_crv)@(0)
	clrl	d0			/* return success status */

get_exit:
	moveml	sp@+, d1-d2		/* restore registers */
	unlk	a6
	rts

	.globl	_cpuRegsSet

_cpuRegsSet:

	link	a6, #0
	moveml	d1-d2, sp@-		/* save registers */

	movel	a6@(arg_crv), d0	/* control register value */
	movel	a6@(arg_cr), d1		/* control register */
	moveq	#7, d2			/* max valid code */
	btst	#11, d1
	bnes	set_cr_set2		/* 800 numbers */

set_cr_set1:
	cmpl	d2, d1
	jhi	set_fail
	movew	pc@(set_tbl1-.-2:b,d1:w:2), d2
	jmp	pc@(2,d2:w)

set_tbl1:
	.word	set_cr_sfc-set_tbl1
	.word	set_cr_dfc-set_tbl1
	.word	set_cr_cacr-set_tbl1
	.word	set_cr_tc-set_tbl1
	.word	set_cr_itt0-set_tbl1
	.word	set_cr_itt1-set_tbl1
	.word	set_cr_dtt0-set_tbl1
	.word	set_cr_dtt1-set_tbl1
	
set_cr_sfc:
	movec	d0, sfc		/* source function code */
	jra	set_done

set_cr_dfc:
	movec	d0, dfc		/* destination function code */
	jra	set_done

set_cr_cacr:
	movec	d0, cacr	/* cache control register */
	jra	set_done

set_cr_tc:
|	movec	d0, tc		/* translation control register */
	.word	0x4e7b, 0x0003
	jra	set_done

set_cr_itt0:
|	movec	d0, itt0	/* instr. transparent translation reg. 0 */
	.word	0x4e7b, 0x0004
	jra	set_done

set_cr_itt1:
|	movec	d0, itt1	/* instr. transparent translation reg. 1 */
	.word	0x4e7b, 0x0005
	jra	set_done

set_cr_dtt0:
|	movec	d0, dtt0	/* data transparent translation reg. 0 */
	.word	0x4e7b, 0x0006
	jra	set_done

set_cr_dtt1:
|	movec	d0, dtt1	/* data transparent translation reg. 1 */
	.word	0x4e7b, 0x0007
	jra	set_done

set_cr_set2:
	bclr	#11, d1
	cmpl	d2, d1
	bhis	set_fail
	movew	pc@(set_tbl2-.-2:b,d1:w:2), d2
	jmp	pc@(2,d2:w)

set_tbl2:
	.word	set_cr_usp-set_tbl2
	.word	set_cr_vbr-set_tbl2
	.word	set_cr_caar-set_tbl2
	.word	set_cr_msp-set_tbl2
	.word	set_cr_isp-set_tbl2
	.word	set_cr_mmusr-set_tbl2
	.word	set_cr_urp-set_tbl2
	.word	set_cr_srp-set_tbl2
	
set_cr_usp:
	movec	d0, usp		/* user stack pointer */
	jra	set_done

set_cr_vbr:
	movec	d0, vbr		/* vector base register */
	jra	set_done

set_cr_caar:
	movec	d0, caar	/* cache address register */
	jra	set_done

set_cr_msp:
	movec	d0, msp		/* master stack pointer */
	jra	set_done

set_cr_isp:
	movec	d0, isp		/* interrupt stack pointer */
	jra	set_done

set_cr_mmusr:
|	movec	d0, mmusr	/* MMU status register */
	.word	0x4e7b, 0x0805
	jra	set_done

set_cr_urp:
|	movec	d0, urp		/* user root pointer */
	.word	0x4e7b, 0x0806
	jra	set_done

set_cr_srp:
|	movec	d0, srp		/* supervisor root pointer */
	.word	0x4e7b, 0x0807
	jra	set_done

set_fail:
	moveq	#-1, d0			/* return set_failure status */
	jra	set_exit

set_done:
	clrl	d0			/* return success status */

set_exit:
	moveml	sp@+, d1-d2		/* restore registers */
	unlk	a6
	rts
