/* #ifndef VXWORKS_5_0_2b */
/*
 * =========================================================================
 * FILE:
 *       dvx_var.c     DVX : Dart Vxworks eXtensions Task Variable facility.
 *		 	     This includes the exit handler facility as  a
 *			     sublayer plus sockets handling.
 *			     This package is meant to be used only
 *			     under vxworks versions >= 5.1.1
 * -------------------------------------------------------------------------
 * -------------------------------------------------------------------------
 * AUTHORS:
 
 *      Margherita Vittone
 *	Dave Berg
 *	
 * -------------------------------------------------------------------------
 * REVISION HISTORY

 *      Apr 15, 1994    Creation date.
 *		
 *	May  3, 1994    Adding the exit handler as a sublayer
 *
 *	May  5, 1994	Adding the socket handling, closing at task deletion.
 *      
 *      Jun 14, 1994    Adding mods after talking to Dave B.
 *  
 *      Apr 13, 1995	Moving the call to dvx_init_verify internally
 *			so that the user does not need to know about it.
 *
 *	May 17, 1995	Adding dvx_verify_taskvar and dvx_delete_taskvar;
 *			Also, added dvx_create_internal_pptr and
 *			dvx_get_internal_ptr to simply the user interface
 *			calls with an internal check on the existance of 
 *			the task variable.
 * -------------------------------------------------------------------------

 * ENTRY POINTS            SCOPE   	DESCRIPTION
 *
 * dvx_setup		   public  This function is called only once at 
 *				   startup time: it is called by the system.
 *				   It initializes the VxWorks task variable and 
 *				   hook facilities; it registers hooks for task
 *				   creation and deletion.
 *				   It creates and initializes
 *				   the tag array of global indexes that will 
 *				   be used to register the number of dvx task
 *				   variables  in the whole system.
 *
 * dvx_init_verify	   static  This function is called internally
 *				   to check that the  DVX task
 *				   variable package has been installed.
 *
 * dvx_task_create_hook    static  This function will be called at each task 
 *				   creation. 
 *				   It allocates the space for exit handlers
 *				   using either code based on DVX task variables
 *				   or code based on VxWorks task variables.
 *				   The LOCAL array to store the
 *				   task specific private dvx task variables. 
 *				   is created and space is malloced for it.
 *				   The LOCAL array pointer will be the only 
 *				   vxworks task variable , it will be shared 
 *				   by all tasks.
 *				   The local array is an array of structures.
 *				
 *
 * dvx_task_delete_hook    static  This function will be called at each task 
 *				   deletion. It calls the exit handlers registered
 *				   by the user with atexit .
 *				   It will free all malloced users' space
 *
 *
 *
 * dvx_init_xhandler       static  This function is called at task creation.
 *   				   It allocate space for exit handler data 
 *				   structure.
 *
 * dvx_create_internal_ptr static  This function will be called by a task
 *				   to create a dvx private variable when is
 *			 	   needed internally, typically in a hook function.
 *
 * dvx_get_internal_ptr    static  This function will be called by a task
 *				   to retrieve an internal dvx private variable.
 *
 * dvx_validate_handle     static  It is internally used to check if a 
 *				   handle is valid or not.
 *
 * dvx_verify_taskvar      static  It is called by the user to verify the
 *                                 existance of a task variable for that
 *                                 handle.
 *
 * dvx_create_ptr          public  This function will be called by a task
 *				   to create a dvx private variable when is a 
 *				   pointer to a malloced space.
 *
 * dvx_get_ptr             public  This function will be called by a task
 *				   to retrieve a dvx private variable.
 *
 * dvx_create_scalar       public  This function will be called by a task
 *				   to add a dvx private scalar variable.
 *				   It returns an available handle for it
 *				   to be used to reference the variable.
 *
 * dvx_set_scalar          public  This function will be called by a task
 *			 	   to set a dvx private variable's value.
 *
 * dvx_get_scalar          public  This function will be called by a task
 *				   to retrieve  a dvx private variable.
 *
 * dvx_var_info            public  This function will be called 
 *				   to access a task variables.
 *
 *
 * atexit		   public  It is called by the user to register exit
 *				   handlers.
 *
 * dvx_atexit		   public  It is called by the user to register exit
 *				   handlers that pass an argument.
 *
 *
 * dvx_delete_taskvar      public  It deletes a task variable and , in case of 
 *				   a pointer to malloced space, it frees it.
 * -------------------------------------------------------------------------
 *
 * -------------------------------------------------------------------------
 * ENVIRONMENT:
 *      VxWorks		Warning: this implementation is reliable only under
 *				 a VxWorks version >= 5_1_1.
 * =========================================================================
 */

/* Include files		*/

#include "dvx.h"			 /* DVX prototypes and constants definitions */

/* --------------------------------------------------------------------------------  */

typedef struct prvData DVX_PDP_STR;

typedef struct tsksock DVX_SOCKET_STR;

static DVX_PDP_STR	*dvxPdp;		    /* Pointer to task private data LOCAL array */


static void	dvx_task_create_hook (WIND_TCB *taskPtr);
static void	dvx_task_delete_hook (WIND_TCB *taskPtr);
static STATUS	dvx_init_xhandler (int);
static STATUS	dvx_create_internal_ptr (int tId, int *dvxHandle, int malsiz, DVX_PTR **varAdd, int varType );
static STATUS	dvx_get_internal_ptr (int tId, int dvxHandle, DVX_PTR **rdptr, int varType );
static STATUS 	dvx_validate_handle (int,int,int);


STATUS          dvx_var_info_get (int tId, DVX_PDP_STR *varList);

typedef struct tskfun_atexit 		DVX_FUN_STR;	/* Used by atexit     */
typedef struct tskfun_dvx_atexit 	DVX_FUN_STRX;	/* Used by dvx_atexit */

DVX_FUN_STR 	*FunPtr;	     /* Pointer to task exit handler space.
					Each task has its own copy of the 
					data. This is used by atexit. */
DVX_FUN_STRX    *FunPtrx;	     /* Same as above, but used by 
					dvx_atexit. */



int	dvxTagCnt;			    /* Index in tag array */
int	*dvxTags [DVX_MAX_VAR];
int	dvxFlag;			    /* This flag is set by dvx_setup
					               and used by dvx_init_verify to check is dvx
						       facility has been initialized */

char    *dvxVarType[] = {"SCALAR","POINTER"};


#ifdef _DVX_VAR                                     /* Using DVX task variables */

static int 	ExitHandle;			    /* To identify special handle for exit 
					               handlers pointers registered with atexit */

static int 	ExitHandlex;			    /* To identify special handle for exit 
					               handlers pointers registered with
					               dvx_atexit */
#endif
/*============================================================================= */

/*============================================================================= */

static void dvx_task_create_hook (WIND_TCB *taskPtr)
{

   int	i;
   DVX_PDP_STR	*pdp;	/* Pointer to task private data LOCAL array */
/* -------------------------------------------------------------------------- */

/* Initializing private task variable array				      */


   taskLock();
/* Storing pointer as Task variable					      */

   if ( taskVarAdd ( (int)taskPtr, (int *)&dvxPdp) != OK)
   {
      logMsg ("\n DVX dvx_task_create_hook: ERROR in adding private data pointer \n",0,0,0,0,0,0);
      taskUnlock();
      return;
   }
   if(( pdp = (DVX_PDP_STR *)malloc (sizeof (DVX_PDP_STR) * DVX_MAX_VAR)) == NULL)
   {
      logMsg ("\n DVX dvx_task_create_hook: ERROR:can't malloc private data space: errno 0x%x \n",
 		  (int)errno,0,0,0,0,0);
      taskUnlock();
      return;
   }
   for (i=0; i < DVX_MAX_VAR; i++) 
   {
      pdp[i].pdpVal  = NULL;
      pdp[i].pdpType = NULL;
      pdp[i].pdpSize = NULL;
   }

   if (taskVarSet ( (int)taskPtr, (int *)&dvxPdp, (int)pdp) != OK)
   {
      logMsg ("\n DVX dvx_task_create_hook ERROR: can't set private data pointer \n",0,0,0,0,0,0);
      taskUnlock();
      return;	     
   }

/* Dealing with exit handler space  */
   dvx_init_xhandler ( (int)taskPtr);

   taskUnlock();
 }

/*============================================================================= */

/*============================================================================= */
/*
   This function will be called at each task  deletion. 
   It calls exit handlers, it frees malloced spaces and closes sockets.
   
   As far as exit handlers, we  allow to use exit handler bases on DVX 
   task variable package as well as exit handler that uses VxWorks task variables. */

static void dvx_task_delete_hook (WIND_TCB *taskPtr)

{

  DVX_FUN_STR 		*ExitFunPtr;
  DVX_FUN_STRX 		*ExitFunPtrx;
/*  DVX_SOCKET_STR  	*SockPtrStr; */			 /* Not used yet */
  int 			j;
  int			xhand_dat;

#ifdef _DVX_VAR

  DVX_PDP_STR   	*pdp;
  STATUS		ret_status;
  STATUS		ret_statusx;
  int			i;

#endif

/* Calling first task exit handler: first those registered with dvx_atexit
   then those registered with atexit.						*/


#ifdef _DVX_VAR

/* Get pointer to private data array 		*/

   pdp = (DVX_PDP_STR *)taskVarGet ( (int)taskPtr, (int *)&dvxPdp);
   if (pdp == (DVX_PDP_STR *)ERROR )
	{
	if (errno != S_taskLib_TASK_VAR_NOT_FOUND)
      		logMsg ("\nDVX dvx_task_delete_hook: Failed getting task variable\n",0,0,0,0,0,0); 
	return;
	}

   ret_statusx = dvx_get_internal_ptr ( (int)taskPtr, ExitHandlex, (DVX_PTR **)&ExitFunPtrx, DVX_XHAND_PTR);
   if ( ret_statusx != OK) 
   {
      logMsg ("\n\n DVX dvx_task_delete_hook: Task delete hook error: pointer/handle to exit handler\n\t not setup correctly \n",0,0,0,0,0,0); 
      return;
   }
   ret_status  = dvx_get_internal_ptr ( (int)taskPtr, ExitHandle, (DVX_PTR **)&ExitFunPtr, DVX_XHAND_PTR);
   if ( ret_status != OK) 
   {
      logMsg ("\n\n DVX dvx_task_delete_hook: pointer/handle to exit handler\n\t not setup correctly \n",0,0,0,0,0,0); 
      return;
   }
#endif

#ifdef _VXW_VAR		/* VxWorks task variables */

   ExitFunPtr  = (DVX_FUN_STR *)taskVarGet ( (int)taskPtr,(int *)&FunPtr );
   ExitFunPtrx = (DVX_FUN_STRX *)taskVarGet ( (int)taskPtr,(int *)&FunPtrx );

#endif

   if (ExitFunPtrx->next !=0)
   {
          for (j= ExitFunPtrx->next-1;j>=0; j--) 
	  {
		xhand_dat = ExitFunPtrx->xhand_dat[j];
		(*ExitFunPtrx->funArrPtr [j])(xhand_dat);
	  }
   }

   if (ExitFunPtr->next !=0)
   {
          for (j= ExitFunPtr->next-1;j>=0; j--) (*ExitFunPtr->funArrPtr [j])();
   }


#ifdef _DVX_VAR



/* Cleaning up the rest of the task variables			*/
   
   for (i=0; i<DVX_MAX_VAR; i++)
   {

/* Closing sockets   (Not yet for now)		*/

/*
     if ( (pdp[i].pdpType & DVX_SOCKET_PTR ) == DVX_SOCKET_PTR )
     {
        SockPtrStr = (DVX_SOCKET_STR *)pdp[i].pdpVal;
 
        if (SockPtrStr->nextSock !=0)
        {
          for (j= SockPtrStr->nextSock-1;j>=0; j--) 
	  {
		close (*(int *)SockPtrStr->sockPtr[j]);
	  }
        }

     }

*/
/* Freeing malloced space for the user						*/

     if ( ( pdp[i].pdpType & DVX_MAL_PTR) == DVX_MAL_PTR) free ( (void *)pdp[i].pdpVal);
   }
  free ( (void *)pdp);

#endif		/* On DVX task variables	*/
	
}
/*============================================================================= */

/*============================================================================= */

STATUS dvx_init_verify()
{
   if (!dvxFlag)
   {   
      logMsg ("\n DVX - ERROR : DVX facility not installed on system. \n\t Contact your system manager or invoke dvx_setup from vxworks shell\n",0,0,0,0,0,0);
      logMsg ("\n DVX - ERROR : Exiting task \n",0,0,0,0,0,0);
      exit(ERROR);
   }
   return (OK);
}
/*============================================================================= */
STATUS dvx_setup()
{
	
   int		i;
   STATUS	ret_hook;

/* ---------------------------------------------------------------------------- */
/* Initialize tag array								*/

   for (i=0; i<DVX_MAX_VAR; i++)  dvxTags[i] = NULL;

/* Initialize global index							*/
   dvxTagCnt = 0;

/* Initializing hook facility and task variable facility			*/

   taskHookInit();  		  /* Warning, it exists only under v5.1 */
   taskVarInit();

/* Make sure only one of these hooks is installed				*/
	
/* Hook called at creation of a task						*/

   if ( (taskCreateHookDelete ((FUNCPTR) dvx_task_create_hook) == ERROR)
	   && (errno != S_taskLib_TASK_HOOK_NOT_FOUND) ) return(ERROR);


   if ( (taskDeleteHookDelete ((FUNCPTR) dvx_task_delete_hook) == ERROR)
	   && (errno != S_taskLib_TASK_HOOK_NOT_FOUND) ) return(ERROR);

/* ---------------------------------------------------------------------------- */

/* Register the functions */

   ret_hook = taskCreateHookAdd ( (FUNCPTR) dvx_task_create_hook);   
   if ( ret_hook != OK) logMsg ("\n DVX dvx_setup: Task create table is full \n",0,0,0,0,0,0); 

   ret_hook = taskDeleteHookAdd ( (FUNCPTR) dvx_task_delete_hook);   
   if ( ret_hook != OK) logMsg ("\n DVX dvx_setup: Task delete table is full \n",0,0,0,0,0,0); 

   dvxFlag = DVX_VAR_START;

   return(OK);
}

/*============================================================================= */

/*============================================================================= */
/*
   This function is called by the initialization hook at task creation.
   It allocate space for task variable and it initializes the data structures.
   The structure is defined in the file dvx.h			      	    
   We  allow to use exit handler bases on DVX task variable package and 
   exit handler that uses VxWorks task variables.           			*/

/* ---------------------------------------------------------------------------- */

static STATUS   dvx_init_xhandler ( int tId )

{

   DVX_FUN_STR *FunStrPtr;	/* Structure to hold function pointers for exit
				   handlers registerd with atexit	       */
   DVX_FUN_STRX *FunStrPtrx;	/* Structure to hold function pointers for exit
				   handlers registerd with dvx_atexit	       */
   int i;

#ifdef _DVX_VAR
   int ret_val;   
#endif
/* -------------------------------------------------------------------------- */


/* Initializing task variable structure for exit handlers based on DVX	      */

#ifdef _DVX_VAR

   ret_val = dvx_create_internal_ptr (tId, &ExitHandle, sizeof(DVX_FUN_STR), (DVX_PTR **)&FunStrPtr, DVX_XHAND_PTR);

   if (ret_val != OK)
   {
      logMsg ("\n DVX dvx_init_xhandler : ERROR in creating space for exit handlers (atexit)\n",0,0,0,0,0,0);
      return(ERROR);
   }

   ret_val = dvx_create_internal_ptr (tId, &ExitHandlex, sizeof(DVX_FUN_STRX), 
	(DVX_PTR **)&FunStrPtrx, DVX_XHAND_PTR);

   if (ret_val != OK)
   {
      logMsg ("\n DVX dvx_init_xhandler: ERROR in creating space for exit handlers (dvx_atexit)\n",0,0,0,0,0,0);
      return(ERROR);
   }

#endif

/* Here if using strictly VxWorks task variables				*/

#ifdef _VXW_VAR

/* Defining pointer to structure to be used as task variable, allocating space 
   and initializing it.							      */


/* Storing pointer as Task variable (for atexit functions)		      */

   if ( taskVarAdd ( tId, (int *)&FunPtr) != OK)
             logMsg ("\n DVX dvx_init_xhandler: ERROR in adding task variable \n",0,0,0,0,0,0);

   if( (FunStrPtr = (DVX_FUN_STR *) malloc (sizeof (DVX_FUN_STR)) ) == NULL)
	 logMsg ("DVX dvx_init_xhandler: ERROR : can't malloc task variable space: errno = %d\n",
	 (int)errno,0,0,0,0,0);

   taskVarSet ( tId, (int *)&FunPtr, (int)FunStrPtr);

/* Storing pointer as Task variable (for dvx_atexit functions)		      */

   if ( taskVarAdd ( tId, (int *)&FunPtrx) != OK)
             logMsg ("\n DVX dvx_init_xhandler: ERROR in adding task variable \n",0,0,0,0,0,0);

   if( (FunStrPtrx = (DVX_FUN_STR *) malloc (sizeof (DVX_FUN_STRX)) ) == NULL)
	 logMsg ("DVX dvx_init_xhandler: ERROR : can't malloc task variable space: errno = %d\n",
	 (int)errno,0,0,0,0,0);

   taskVarSet ( tId, (int *)&FunPtrx, (int)FunStrPtrx);

#endif   

   FunStrPtr->next = NULL ;
   FunStrPtrx->next = NULL ;
	
   for (i=0; i<DVX_MAX_XHAND; i++) 
   {
	FunStrPtr->funArrPtr[i] = NULL;

	FunStrPtrx->funArrPtr[i] = NULL;
	FunStrPtrx->xhand_dat[i] = NULL;
   }
   return (OK);
 }

/*============================================================================= */

/*============================================================================= */

static STATUS dvx_create_internal_ptr (int tId, int *dvxHandlePtr, int malsiz, DVX_PTR **malptr, int varType)
{

   DVX_PDP_STR *pdp;
   int i;
   int ret_status;
   int handle_flag;
/* ---------------------------------------------------------------------------- */

/* Check on dvx to be initialized */

   if ( dvx_init_verify() != OK)
   {
             logMsg ("\n DVX dvx_create_internal_ptr: ERROR dvx not initialized, exiting task \n",0,0,0,0,0,0);
	     exit(1);

   }


   handle_flag = DVX_NEW_HANDLE;

/* First validate handle	*/

   taskLock();

   ret_status = dvx_validate_handle ( tId, *dvxHandlePtr, DVX_CREATE_HANDLE );

   if (ret_status != OK) 
   {
	taskUnlock ();

#ifdef _DEBUG
   	logMsg ("\n DVX: dvx_create_ptr : invalid handle for tId = 0x%x \n",tId,0,0,0,0,0);
#endif
	return (ret_status);
   }			
/* Need to get available index in the tag array		*/
   
   i=0;


   while( (dvxTags [i] != 0) && (i<=dvxTagCnt) ) 
   {
      if ( dvxTags[i] == dvxHandlePtr ) 
      {
        handle_flag = DVX_OLD_HANDLE;
	break;				/* Found existing handle */
      }
      else
      {
        i++;
      }
   }

      *dvxHandlePtr = i;
      dvxTags [i] = dvxHandlePtr;

      if (handle_flag == DVX_NEW_HANDLE) dvxTagCnt = i;

      taskUnlock();   


/* Mallocing space	*/
/* ---------------------------------------------------------------------------- */

      if (malsiz != 0) 
      {
        if ( (*malptr = (DVX_PTR *)malloc( malsiz * sizeof (char)) ) == NULL)
        {
#ifdef _DEBUG
          logMsg ("DVX dvx_create_ptr: ERROR : can't malloc task variable space: errno = %d\n",
 		  (int)errno,0,0,0,0,0);
#endif
          return (errno);
        }	
      }

/* Storing private data information 	*/

/* Get pointer to private data array						*/

   pdp = (DVX_PDP_STR *)taskVarGet ( tId, (int *)&dvxPdp);

   pdp[i].pdpVal  = *(int *)malptr;
   pdp[i].pdpType = DVX_TASK_INTERNAL_VAR | varType;	
   pdp[i].pdpSize = malsiz;
   
   if (malsiz != 0) pdp[i].pdpType |= DVX_MAL_PTR ;	
   return (OK);
}
/*============================================================================= */

/*============================================================================= */

static STATUS dvx_get_internal_ptr (int tId, int dvxHandle, DVX_PTR **varPtr, int varType)

{

   DVX_PDP_STR *pdp;
   int i;
   int ret_status;

/* ---------------------------------------------------------------------------- */

/* Check first if dvx has been initialized					*/

   if ( dvx_init_verify() != OK) return (ERROR);


/* First validate handle	*/

   ret_status = dvx_validate_handle ( tId, dvxHandle, DVX_VERIFY_HANDLE );

   if (ret_status != OK) 
   {
#ifdef _DEBUG
   	logMsg ("\n DVX: dvx_get_ptr : invalid handle for tId = 0x%x \n",tId,0,0,0,0,0);
#endif

	return (ret_status);
   }

/* Verifying that the variable exists 						*/

   ret_status = dvx_verify_taskvar (tId, dvxHandle, varType);

   if (ret_status != OK)   return (ret_status);

   i = dvxHandle;


/* Get pointer to private data array						*/

   pdp = (DVX_PDP_STR *)taskVarGet ( tId, (int *)&dvxPdp);

   *varPtr = (DVX_PTR *)pdp[i].pdpVal;

   return(OK);

}
/*================================================================================== */

/*============================================================================= */

STATUS dvx_create_ptr (int tId, int *dvxHandlePtr, int malsiz, DVX_PTR **malptr)
{

   DVX_PDP_STR *pdp;
   int i;
   int ret_status;
   int handle_flag;
/* ---------------------------------------------------------------------------- */

/* Check on dvx to be initialized */

   if ( dvx_init_verify() != OK)
   {
             logMsg ("\n DVX dvx_create_ptr: ERROR dvx not initialized, exiting task \n",0,0,0,0,0,0);
	     exit(1);

   }

   handle_flag = DVX_NEW_HANDLE;

/* First validate handle	*/

   taskLock();

   ret_status = dvx_validate_handle ( tId, *dvxHandlePtr, DVX_CREATE_HANDLE );

   if (ret_status != OK) 
   {
	taskUnlock ();

#ifdef _DEBUG
   	logMsg ("\n DVX: dvx_create_ptr : invalid handle for tId = 0x%x \n",tId,0,0,0,0,0);
#endif
	return (ret_status);
   }			
/* Need to get available index in the tag array		*/
   
   i=0;


   while( (dvxTags [i] != 0) && (i<=dvxTagCnt) ) 
   {
      if ( dvxTags[i] == dvxHandlePtr ) 
      {
        handle_flag = DVX_OLD_HANDLE;
	break;				/* Found existing handle */
      }
      else
      {
        i++;
      }
   }

      *dvxHandlePtr = i;
      dvxTags [i] = dvxHandlePtr;

      if (handle_flag == DVX_NEW_HANDLE) dvxTagCnt = i;

      taskUnlock();   


/* Mallocing space	*/
/* ---------------------------------------------------------------------------- */


      if (malsiz != 0) 
      {
/*         if ( (*malptr = (DVX_PTR *)calloc( malsiz, sizeof (char)) ) == NULL) */

         if ( (*malptr = (DVX_PTR *)malloc( malsiz * sizeof (char)) ) == NULL)
        {
#ifdef _DEBUG
          logMsg ("DVX dvx_create_ptr: ERROR : can't malloc task variable space: errno = %d\n",
 		  (int)errno,0,0,0,0,0);
#endif
          return (errno);
        }	

  
     }

/* Storing private data information 	*/

/* Get pointer to private data array						*/

   pdp = (DVX_PDP_STR *)taskVarGet ( tId, (int *)&dvxPdp);

   pdp[i].pdpVal  = *(int *)malptr;
   pdp[i].pdpType = DVX_TASK_VAR | DVX_PTR_TYPE;	
   pdp[i].pdpSize = malsiz;
   if (malsiz != 0) pdp[i].pdpType |= DVX_MAL_PTR ;	

#ifdef _DEBUG
   if (malsiz == 0) printf ("\n DVX dvx_create_ptr , storing pointer at 0x%x \n",pdp[i].pdpVal);
   if (malsiz != 0) printf ("\n DVX dvx_create_ptr , returning pointer to malloced space at 0x%x, size = %d\n",
		    pdp[i].pdpVal,pdp[i].pdpSize);
#endif

   return (OK);
}
/*============================================================================= */

/*============================================================================= */

STATUS dvx_get_ptr (int tId, int dvxHandle, DVX_PTR **varPtr)

{

   DVX_PDP_STR *pdp;
   int i;
   int ret_status;

/* ---------------------------------------------------------------------------- */

/* Check first if dvx has been initialized					*/

   if ( dvx_init_verify() != OK) return (ERROR);


/* First validate handle	*/

   ret_status = dvx_validate_handle ( tId, dvxHandle, DVX_VERIFY_HANDLE );

   if (ret_status != OK) 
   {
#ifdef _DEBUG
   	logMsg ("\n DVX: dvx_get_ptr : invalid handle for tId = 0x%x \n",tId,0,0,0,0,0);
#endif

	return (ret_status);
   }

/* Verifying that the variable exists 						*/

   ret_status = dvx_verify_taskvar (tId, dvxHandle, DVX_PTR_TYPE);

   if (ret_status != OK)   return (ret_status);

   i = dvxHandle;


/* Get pointer to private data array						*/

   
   pdp = (DVX_PDP_STR *)taskVarGet ( tId, (int *)&dvxPdp);

   *varPtr = (DVX_PTR *)pdp[i].pdpVal;
/*   printf ("\n *** dvx_get_ptr , pointer is 0x%x\n",*varPtr); */

   return(OK);

}
/*============================================================================= */

/*============================================================================= */

STATUS dvx_create_scalar (int tId, int *dvxHandlePtr)
{

   DVX_PDP_STR *pdp;
   int i;
   int ret_status;
   int handle_flag;
/* ---------------------------------------------------------------------------- */

/* Check first if dvx has been initialized					*/

   if ( dvx_init_verify() != OK) return (ERROR);

   handle_flag = DVX_NEW_HANDLE;

/* First validate handle	*/

   taskLock();

   ret_status = dvx_validate_handle ( tId, *dvxHandlePtr, DVX_CREATE_HANDLE);

   if (ret_status != OK) 
   {
	taskUnlock();

#ifdef _DEBUG
   	logMsg ("\n DVX: dvx_create_scalar : invalid handle for tId = 0x%x \n",tId,0,0,0,0,0);
#endif
	return (ret_status);
   }

/* Need to get available index in the tag array		*/
   
   i=0;

   while( (dvxTags [i] != 0) && (i<=dvxTagCnt) ) 
   {
      if ( dvxTags[i] == dvxHandlePtr ) 
      {
   	handle_flag = DVX_OLD_HANDLE;
	break;
      }
      else
      {
        i++;
      }
   }

   *dvxHandlePtr = i;
   dvxTags [i] = dvxHandlePtr;

   if (handle_flag == DVX_NEW_HANDLE) dvxTagCnt = i;

   taskUnlock();   


/* Get pointer to private data array						*/
/* ---------------------------------------------------------------------------- */
   
   pdp = (DVX_PDP_STR *)taskVarGet ( tId, (int *)&dvxPdp);

   pdp[i].pdpType  = DVX_TASK_VAR | DVX_SCALAR_TYPE ;

   return (OK);
}
/*============================================================================= */

/*============================================================================= */

STATUS dvx_set_scalar (int tId, int dvxHandle, int varVal)
{

   DVX_PDP_STR *pdp;
   int i;
   int ret_status;
/* ---------------------------------------------------------------------------- */

/* Check first if dvx has been initialized					*/

   if ( dvx_init_verify() != OK) return (ERROR);

/* First validate handle	*/

   ret_status = dvx_validate_handle ( tId, dvxHandle, DVX_VERIFY_HANDLE );

   if (ret_status != OK) 
   {
#ifdef _DEBUG
   	logMsg ("\n DVX: dvx_set_scalar : invalid handle for tId = 0x%x \n",tId,0,0,0,0,0);
#endif

	return (ret_status);
   }

   ret_status = dvx_verify_taskvar (tId, dvxHandle, DVX_SCALAR_TYPE);

   if (ret_status != OK)  return (ret_status);

   i = dvxHandle;

/* Get pointer to private data array						*/
   
   pdp = (DVX_PDP_STR *)taskVarGet ( tId, (int *)&dvxPdp);

   pdp[i].pdpVal  = varVal;

   return (OK);
}
/*============================================================================= */

/*============================================================================= */

STATUS dvx_get_scalar (int tId, int dvxHandle, int *rdvar)
{

   DVX_PDP_STR *pdp;
   int i;
   int ret_status;
/* ---------------------------------------------------------------------------- */

/* Check first if dvx has been initialized					*/

   if ( dvx_init_verify() != OK) return (ERROR);

/* First validate handle	*/

   ret_status = dvx_validate_handle ( tId, dvxHandle, DVX_VERIFY_HANDLE);

   if (ret_status != OK) 
   {
#ifdef _DEBUG
   	logMsg ("\n DVX: dvx_get_scalar : invalid handle for tId = 0x%x \n",tId,0,0,0,0,0);
#endif

	return (ret_status);
   }

   ret_status = dvx_verify_taskvar (tId, dvxHandle, DVX_SCALAR_TYPE);

   if (ret_status != OK)   return (ret_status);

   i = dvxHandle;

/* Get pointer to private data array						*/

   pdp = (DVX_PDP_STR *)taskVarGet ( tId, (int *)&dvxPdp);
   *rdvar = pdp[i].pdpVal; 
   return (OK);

}
/*================================================================================== */

/*================================================================================== */
/* This function is called to print task variables				     */

STATUS dvx_var_info (int tId)
{
   int return_status;
   int numvar, i;
   DVX_PDP_STR *varList;

/* Verify the existance of the task first 	*/

   if ( (return_status = taskIdVerify (tId)) != OK) 
   {
   	logMsg ("\n DVX: dvx_var_info : NO task found with id = 0x%x \n",tId,0,0,0,0,0);
	return (OK);
   }
   return_status = dvx_var_info_get ( tId, varList);
   if (return_status == DVX_INFO_ERROR )
   {

	printf ("\n error from var_info = %d\n", return_status);
#ifdef _DEBUG
   	logMsg ("\n DVX: dvx_var_info : ERROR for tid = 0x%x , status= %d\n",tId,0,0,0,0,0);
#endif
	return (return_status);
   }
   else if (return_status == DVX_INFO_NOVAR)
   {
	printf ("\n no task var found, =%d\n",return_status);
#ifdef _DEBUG
   	logMsg ("\n DVX: dvx_var_info: NO task variables for tid = 0x%x\n",tId,0,0,0,0,0);
#endif
	return (return_status);
   }
   else
   {
/* Print here the task variables	*/
     numvar = return_status;
     clearerr (stdin);
     clearerr (stdout);
     printf ("\n    ===================================================== \n");
     printf ("\n\t  %d Variables found for task id =  0x%x \n",numvar,tId);
     for (i=0;i <numvar; i++)
     {
	if ( (varList[i].pdpType & DVX_SCALAR_TYPE ) == DVX_SCALAR_TYPE )
	{
	  printf ("\n\t var = 0x%08lx, type = %s \n", varList[i].pdpVal, dvxVarType[SCALAR_TYPE]);
	}
	else if ( (varList[i].pdpType & DVX_PTR_TYPE) == DVX_PTR_TYPE)
	{
	  printf ("\n\t var = 0x%08lx, type = %s , size = %d \n", 
		   varList[i].pdpVal, dvxVarType[PTR_TYPE], varList[i].pdpSize);
	}

     }   
     printf ("\n    ===================================================== \n");

   }
   return (OK);
}
/*================================================================================== */

/*============================================================================= */

/* This function is used to access a task task variables			*
   It returns the number of task varibles found or an error status		*/


/* ---------------------------------------------------------------------------- */

STATUS dvx_var_info_get (int tId, DVX_PDP_STR *varList)			
{

   DVX_PDP_STR *pdp;
   int i;
   int numvar;		/* Number of task variables found	*/
   int novarCount;	/* Increments when no var found		*/   

/* ---------------------------------------------------------------------------- */

/* Check first if dvx has been initialized					*/

   if ( dvx_init_verify() != OK) return (ERROR);


   numvar = 0;
   novarCount = 0;

   taskLock();
         
/* Get pointer to private data array						*/
   
   pdp = (DVX_PDP_STR *)taskVarGet ( tId, (int *)&dvxPdp);

   for (i= 0; i<=dvxTagCnt; i++)
   { 
     if ( (pdp[i].pdpType & DVX_TASK_VAR) == DVX_TASK_VAR)
     {
	
	varList[numvar].pdpVal  = pdp[i].pdpVal;
	varList[numvar].pdpType = pdp[i].pdpType;
        if ( (pdp[i].pdpType & DVX_MAL_PTR) == DVX_MAL_PTR)
        {	
             varList[numvar].pdpSize = pdp[i].pdpSize;
	}	
	numvar ++;
     }
     else
     {
	novarCount ++;
     }
   }	
   taskUnlock();

/*   Check if we did not find any task variable created by the user */

   if (numvar == 0 ) return (DVX_INFO_NOVAR); 

   return (numvar);
}

/*================================================================================== */

/*============================================================================= */
/*   atexit:
   Each task invokes atexit to register exit handlers that will be invoked upon
   task exit. A max of 32 handlers are allowed.
   It accesses the TCB block for the calling task to store in the TCB, in the 
   proper location, our task variable that is the pointer to the structure with 
   the function addresses. 
   Upon each call , it updates the index to the array of function pointers
   using the structure component "next".				        */

/* ---------------------------------------------------------------------------- */


int atexit (void (*regFunc)(void))
{


   DVX_FUN_STR *taskFunPtr;	/* Pointer to task variables private data */

   WIND_TCB *TaskTcbPtr;
   
#ifdef _DVX_VAR                 /* Using DVX task variables */
   STATUS ret_status;
#endif

   int  fun_index;
/*   int  tId; */

/* ----------------------------------------------------------- */

   TaskTcbPtr = taskTcb(0);
           
#ifdef _DVX_VAR			/* Using DVX task variables */


   dvx_init_verify();

/* Retrieving exit handler pointer using the handle 			*/

     ret_status = dvx_get_internal_ptr ( (int)TaskTcbPtr, ExitHandle, (DVX_PTR **)&taskFunPtr, DVX_XHAND_PTR); 
     if (ret_status != OK)
         logMsg ("DVX - dvx_get_ptr ERROR : unable to get pointer to exit handlers\n",
         0,0,0,0,0,0);

#endif


#ifdef _VXW_VAR			/* Using VxWorks task variables		*/

   taskFunPtr = (DVX_FUN_STR *)taskVarGet ( (int)TaskTcbPtr,(int *)&FunPtr);

#endif


   fun_index = (int) taskFunPtr->next;

   if (fun_index == DVX_MAX_XHAND) return (DVX_TABLE_FULL);
		   	   	   
   taskFunPtr->funArrPtr[fun_index] = regFunc;

   taskFunPtr->next++;

   return (OK);


}

/*================================================================================== */

/*============================================================================= */
/*   dvx_atexit:
   Each task invokes dvx_atexit to register exit handlers that will be invoked upon
   task exit. A max of 32 handlers are allowed.
   The difference with atexit is to allow passing an argument to the exit handler
   that is not allow according to the standards of atexit.

   It accesses the TCB block for the calling task to store in the TCB, in the 
   proper location, our task variable that is the pointer to the structure with 
   the function addresses. 
   Upon each call , it updates the index to the array of function pointers
   using the structure component "next".				        */

/* ---------------------------------------------------------------------------- */


int dvx_atexit (int (*regFunc)(int), int xhand_var)
{


   DVX_FUN_STRX *taskFunPtrx;	/* Pointer to task variables private data */

   WIND_TCB *TaskTcbPtr;
   
#ifdef _DVX_VAR                 /* Using DVX task variables */
   STATUS ret_status;
#endif

   int  fun_index;

/* ----------------------------------------------------------- */

   TaskTcbPtr = taskTcb(0);
           
   dvx_init_verify();

#ifdef _DVX_VAR			/* Using DVX task variables */

/* Retrieving exit handler pointer using the handle 			*/

     ret_status = dvx_get_internal_ptr ( (int)TaskTcbPtr, ExitHandlex, (DVX_PTR **)&taskFunPtrx, DVX_XHAND_PTR); 

#endif


#ifdef _VXW_VAR			/* Using VxWorks task variables		*/

   taskFunPtrx = (DVX_FUN_STRX *)taskVarGet ( (int)TaskTcbPtr,(int *)&FunPtr);

#endif


   fun_index = (int) taskFunPtrx->next;

   if (fun_index == DVX_MAX_XHAND) return (DVX_TABLE_FULL);
		   	   	   
   taskFunPtrx->funArrPtr[fun_index] = regFunc;
   taskFunPtrx->xhand_dat[fun_index] = xhand_var;

   taskFunPtrx->next++;

   return (OK);


}

/*================================================================================== */

/*============================================================================= */
/* =================================================================================== */
/* This function is called to verify that the handle has a valid value.		       */

static STATUS dvx_validate_handle (int tId, int handle, int handle_flag )
{

   switch (handle_flag)
   {
     case DVX_CREATE_HANDLE:

   	if (dvxTagCnt >= DVX_MAX_VAR - 1)
	{
#ifdef _DEBUG
	     logMsg ("\n DVX dvx_validate_handle: ERROR: MAX number of task variables reached , tId = 0x%x\n",
		     tId,0,0,0,0,0);
#endif
	     return (DVX_VAR_FULL);
   	}
	
	break;

     case DVX_VERIFY_HANDLE:

        if ( (handle <0) || (handle >dvxTagCnt) )
   	{
#ifdef _DEBUG
	     logMsg ("\n DVX dvx_validate_handle: ERROR: Invalid handle = 0x%x from tid= 0x%x\n",handle,tId,0,0,0,0);

	     if ( handle < 0) 		logMsg ("\n\t\t  handle <0 = 0x%x \n",handle,0,0,0,0,0);
	     if ( handle > dvxTagCnt )  logMsg ("\n\t\t  handle > max allowed = 0x%x  0x%x  \n",
					handle, dvxTagCnt,0,0,0,0);
#endif
	     return (DVX_VAR_INVALID);
	}
	else if ( dvxTags[handle] == 0) 
	{	

#ifdef _DEBUG
     	     logMsg ("\n DVX dvx_validate_handle: ERROR: Invalid handle address = 0x%x \n",(int)dvxTags[handle],0,0,0,0,0);
#endif
	     return (DVX_VAR_INVALID);
   	}

	break;
   }

   return(OK);
}
/*============================================================================= */

/* This function is called to verify that the handle has a valid address and that a 
   task variable exists associated to it for the current task		       */

STATUS dvx_verify_taskvar (int tId, int dvxHandle, int varType)
{

   DVX_PDP_STR *pdp;
   int return_status;
/* ---------------------------------------------------------------------------- */

/* Check on dvx to be initialized */

   if ( dvx_init_verify() != OK)
   {
             logMsg ("\n DVX - ERROR dvx not initialized, exiting task \n",0,0,0,0,0,0);
	     exit(1);

   }

/* Validate handle value	*/
   
   
   if (( return_status = dvx_validate_handle ( tId, dvxHandle, DVX_VERIFY_HANDLE )) != OK )
   {
#ifdef _DEBUG
	logMsg ("\n DVX dvx_verify_taskvar: invalid handle for tId =0x%x \n",tId,0,0,0,0,0);
#endif
   
	return (return_status) ;
   }

/* If we get here we have a valid handle; now check that the task variable really exists  */
   
   pdp = (DVX_PDP_STR *)taskVarGet ( tId, (int *)&dvxPdp);
        
   if ( ( pdp[dvxHandle].pdpType & varType) != varType )
   {
#ifdef _DEBUG

   	logMsg ("\n DVX dvx_verify_taskvar: ERROR: not existing task variable for tId= 0x%x \n",
		 tId,0,0,0,0,0);
#endif
	return(DVX_NO_TASKVAR);
   }

   return (OK);

}
/*============================================================================= */

/* This function is called to delete a task variable 				*/

STATUS dvx_delete_taskvar (int tId, int dvxHandle)
{

   DVX_PDP_STR *pdp;
   int i;
   int return_status;
/* ---------------------------------------------------------------------------- */

/* Check on dvx to be initialized */

   if ( dvx_init_verify() != OK)
   {
             logMsg ("\n DVX - ERROR dvx not initialized, exiting task \n",0,0,0,0,0,0);
	     exit(1);

   }

/* Validate handle value	*/
   
   if (( return_status = dvx_validate_handle ( tId, dvxHandle, DVX_VERIFY_HANDLE )) != OK )
   {
#ifdef _DEBUG
	logMsg ("\n DVX dvx_delete_taskvar: invalid handle for tId =0x%x \n",tId,0,0,0,0,0);
#endif
   return (return_status) ;

   }

   i = dvxHandle;   

/* Now check that the task variable really exists 					*/
   
   pdp = (DVX_PDP_STR *)taskVarGet ( tId, (int *)&dvxPdp);

   if ( (DVX_PTR *)pdp[dvxHandle].pdpType == NULL )
   {
#ifdef _DEBUG
   		logMsg ("\n DVX dvx_delete_taskvar: ERROR: not existing task variable for tId= 0x%x  \n",tId,0,0,0,0,0);
#endif
		return(DVX_NO_TASKVAR);
    }

/* Deleting task variable and free space in case it was a pointer to a malloced area. 	*/

    if ( ( pdp[i].pdpType & DVX_MAL_PTR) == DVX_MAL_PTR) free ( (void *)pdp[i].pdpVal);
    pdp[i].pdpVal  = NULL;
    pdp[i].pdpType = NULL;
    pdp[i].pdpSize = NULL;
    return (OK);

}
/*============================================================================= */

/* #endif	*/	/* On VxWorks version v5_0_2b */
