/****************************** Copyright Notice ******************************
 *                                                                            *
 * Copyright (c) 1992 Universities Research Association, Inc.                 *
 *               All Rights Reserved.                                         *
 *                                                                            *
 ******************************************************************************/

/* @+Public+@
 * PROJECT:	vx_tools
 *		timerTask.c
 *
 * ABSTRACT:	Provides timer alarm messages for other tasks.  This utility
 *              defines two possible message structures for the messages the
 *              task processes.
 *              Although, there may be a more effecient way to implement this
 *              utility by running it from the system clock interrupt service
 *              routine, a safe and simple implementation of this has not been
 *              developed; msgQReceive "must not be called by interrupt
 *              service routines." Ref. msgQReceive, tickAnnounce.
 *
 * ENVIRONMENT:	ANSI C.
 *
 * AUTHOR:      ron,  initial creation: 11-Apr-1995
 *
 * @-Public-@
 ******************************************************************************/
/*--------------------------------- INCLUDES ---------------------------------*/

#include "vxWorks.h"	/* NULL, inc. approp. "types.h" - STATUS, *FUNCPTR */
#include "msgQLib.h"	/* MSG_Q_ID	*/
#include "errno.h"	/* errno	*/
#include "usrLib.h"	/* printErrno	*/
#include "taskLib.h"	/* taskSpawn	*/
#include "stdio.h"	/* printf	*/
#include "logLib.h"	/* logMsg	*/

#include "tickLib.h"	/* for test - tickGet		*/
#include "sysLib.h"	/* for test - sysClkRateGet	*/

#include "timerTask.h"	/* enum and structure def	*/

#ifndef TIMER_DEBUG
# define TIMER_DEBUG 0
#endif

/*-------------------------- LOCAL DEFINES/DECLARES --------------------------*/

	MSG_Q_ID			 timer_msg_q;
	int				 timer_debug = TIMER_DEBUG;

#define TDBG( msg,p1,p2 )	if (timer_debug) logMsg(msg"\n",p1,p2,0,0,0,0)


enum e_typ
{     t_e_ptr
    , t_e_cpy
};

struct	s_que
{   struct s_que	 *next;
    enum e_typ		  snd_typ;
    struct s_tmr_msg_fmt *t_msg;
};

	struct s_que		*head, *tail;
	struct s_que		*empty_list;

/* debugging */	    
static	int			 qcnt=0, z;

static	MSG_Q_ID		 cpy_q;
static	MSG_Q_ID		 g_msg_q;	/* for timerSend */

/* for malloced areas */
static	struct s_que		*t_q;
static	struct s_tmr_msg_fmt	*msg_cpy;


/*--------------------------------- THE CODE ---------------------------------*/

/******************************************************************************
 * @+Public+@
 * ROUTINE: send_n_delete:  Added by ron on 11-Apr-1995
 *
 * DESCRIPTION
 *       support routine
 *
 * RETURN VALUES:       None.
 *
 * SIDE EFFECTS:        None
 *
 * SYSTEM CONCERNS:     None.
 *
 ******************************************************************************/  


static void send_n_delete(  struct s_que	*p
			  , struct s_que	*q )
{							/* @-Public-@ */
	STATUS	sts;
qcnt--;

    if (q->t_msg->u.tmr.return_q != 0)
    {
	if (q->snd_typ == t_e_ptr)
	{   sts = msgQSend(  q->t_msg->u.tmr.return_q, (char *)&(q->t_msg)
			   , sizeof(q->t_msg), NO_WAIT, MSG_PRI_NORMAL );
	    if (sts == ERROR)  printErrno(errno);
	    TDBG( "sent %d bytes", sizeof(q->t_msg), 0 );
	}
	else
	{   sts = msgQSend(  q->t_msg->u.tmr.return_q, (char *)(q->t_msg)
			   , sizeof(*q->t_msg), NO_WAIT, MSG_PRI_NORMAL );
	    if (sts == ERROR)  printErrno(errno);
	    TDBG( "sent %d bytes", sizeof(*q->t_msg), 0 );
	}
    }
    else
    {   logMsg( "warning: return message q == 0\n", 0,0,0,0,0,0 );
    }

    if (q->snd_typ != t_e_ptr)
    {   /* always return our cpy area to our q */
	msgQSend( cpy_q, (char *)&(q->t_msg), sizeof(q->t_msg)
		 , NO_WAIT, MSG_PRI_NORMAL );
    }

    if(q==head) head = q->next;
    else p->next = q->next;

    q->next = empty_list;
    empty_list = q;
}   /* send_n_delete */


/******************************************************************************
 * @+Public+@
 * ROUTINE: timerTask:  Added by ron on 11-Apr-1995
 *          task names > 9 chars mess up vxworks' spy output
 *
 * DESCRIPTION
 *       The actual task.  The task will handle two types of timer messages:
 *       a) a pointer to a struct which has a message type or b) an actual
 *       structure it self.  The determination is made by checking the number
 *       of bytes copied to the receive buffer.
 *
 * RETURN VALUES:       None.
 *
 * SIDE EFFECTS:        None
 *
 * SYSTEM CONCERNS:     None.
 *
 ******************************************************************************/


void timerTask(  MSG_Q_ID	mQ
	       , int		q_sz )
{   							/* @-Public-@ */
	struct s_tmr_msg_fmt	*r_msg_p;	/* receiving */
	struct s_tmr_msg_fmt	*msg_p;		/* working   */
	void			**v_p;
	int			i;
	register int		msz;		/* message size */
	register struct s_que	*q_p;
	register struct s_que	*prev_p;
	STATUS			sts;
	enum e_typ		typ;
	int			msg_found = 0;	/* flag used w/stat and abort*/


    if (q_sz == 0) q_sz = 50;
    if (mQ == 0)
    {   if (timer_msg_q == 0)
	{   timer_msg_q = msgQCreate( q_sz, sizeof(*msg_p), MSG_Q_FIFO );
	}
	else
	{   TDBG( "using q (0x%x) in global timer_msg_q", timer_msg_q, 0 );
	}
	mQ = timer_msg_q;
    }
    g_msg_q = mQ;	/* save in static copy - only for timerSend */

    /* INITIAL STATIC AREAS */
    if (msg_cpy == 0)
    {   msg_cpy
	   = (struct s_tmr_msg_fmt *)malloc(q_sz*sizeof(struct s_tmr_msg_fmt));
    }
    else
    {   logMsg( "warning: using existing malloc for msg_cpy\n" );
    }
    if (t_q == 0)
    {   t_q      =     (struct s_que *)malloc( q_sz*sizeof(struct s_que) );
    }
    else
    {   logMsg( "warning: using existing malloc for t_q\n" );
    }

    /* for when an entire msg is sent (as opposed to a ptr to a msg), we
       maintain a q of ptrs to places where we can copy the message */
    if (cpy_q == 0)
    {    cpy_q = msgQCreate( q_sz, sizeof(msg_p), MSG_Q_FIFO );
    }
    else
    {   logMsg(
           "warning: clearing existing static q - deleted q should be zeroed\n"
	       , 0,0,0,0,0,0);
	while (msgQReceive(cpy_q,(char *)&msg_p,sizeof(msg_p),NO_WAIT)!=ERROR)
	    ;
    }
    for (i=q_sz; i--; )
    {   msg_p = &(msg_cpy[i]);
	msgQSend( cpy_q, (char *)&msg_p, sizeof(msg_p)
		                                   , NO_WAIT, MSG_PRI_NORMAL );
    }

    for (i=0;i<q_sz-1;i++)	/* NOTE I really can only have q_sz-1 b/c I */
    {   t_q[i].next = &t_q[i+1];/* use tail to point to a que location that */
    }				/* ends up not being used (the last one off */
				/* the empty list */
    t_q[i].next = NULL;		/* end of the empty list */
    empty_list = &t_q[1];
    head = tail = &t_q[0];
    tail->next = NULL;

    /* initialize r_msg_p */
    sts = msgQReceive(cpy_q,(char *)&r_msg_p,sizeof(r_msg_p),NO_WAIT);
    if (sts == ERROR)
    {   logMsg( "fatal\n", 0,0,0,0,0,0 );
	taskSuspend( 0 );
    }
    TDBG( "msg_p=0x%x msg_cpy = 0x%x", msg_p, msg_cpy );

    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

    while(1)
    {   if (head->next == NULL)
	{   /* only when we have nothing to do can we WAIT_FOREVER */
	    msz = msgQReceive( mQ, (char *)r_msg_p, sizeof(*r_msg_p), WAIT_FOREVER );
	    taskDelay( 1 );	/* for consistant operation with when busy */
	    goto l1;
	}
	taskDelay( 1 );	/* cannot wait in msgQReceive as receipt of msg would
			   blow timing */
	while(   (msz=msgQReceive(mQ,(char *)r_msg_p,sizeof(*r_msg_p),NO_WAIT))
	      != ERROR )
l1:	{   if (msz == sizeof(struct s_tmr_msg_fmt))
	    {   typ = t_e_cpy;
		msg_p = r_msg_p; /* we'll get at the end if we need it */
		TDBG( "full msg received", 0,0 );
	    }
	    else
	    {   v_p = (void **)msg_p;
		msg_p   = (struct s_tmr_msg_fmt *)*v_p;
		typ = t_e_ptr;
		TDBG( "ptr msg received", 0,0 );
	    }
	    switch (msg_p->u.tmr.cmd)
	    {
	    case tmr_e_add_func:
	    case tmr_e_add:
		/* add to que and update lists */
		msg_p->u.tmr.sts = msg_p->u.tmr.cmd;	/* set to cmd as
							   initial status */
		if ( empty_list != NULL )
		{   tail->t_msg = msg_p;
		    tail->snd_typ = typ;

		    tail->next = empty_list;
		    tail = empty_list;
		    empty_list = empty_list->next;
		    tail->next = NULL;
qcnt++;
		}
		else
		{   logMsg("ERROR: timer que full\n");
		}
		break;
	    case tmr_e_mid_abort:
		msg_found = 0;
		/* abort all matching msgs */
		for (prev_p=q_p=head; q_p->next!=NULL; q_p=q_p->next)
		{   if (q_p->t_msg->u.tmr.mid == msg_p->u.tmr.mid)
		    {   if (   (q_p->t_msg->u.tmr.cmd==tmr_e_add_func)
				&& (*q_p->t_msg->u.tmr.func_p) )
			{   (*q_p->t_msg->u.tmr.func_p)(tmr_func_e_abort, q_p->t_msg->u.tmr.func_arg);
			}
			q_p->t_msg->u.tmr.sts = tmr_e_abort;
			send_n_delete( prev_p, q_p );
			msg_found = 1;		/* mark that we have found */
		    }
		    else
		    {   prev_p=q_p;	/* if current not deleted, it becomes prev for future send_n_delete */
		    }
		}
		if (! msg_found)
		    logMsg(  "warning: msg not found - cmd=%d\n"
			   , msg_p->u.tmr.cmd,0,0,0,0,0 );
		break;
	    case tmr_e_tid_abort:
		msg_found = 0;
		/* abort all matching msgs */
		for (prev_p=q_p=head; q_p->next!=NULL; q_p=q_p->next)
		{   if (q_p->t_msg->u.tmr.tid == msg_p->u.tmr.tid)
		    {   if (   (q_p->t_msg->u.tmr.cmd==tmr_e_add_func)
				&& (*q_p->t_msg->u.tmr.func_p) )
			{   (*q_p->t_msg->u.tmr.func_p)(tmr_func_e_abort, q_p->t_msg->u.tmr.func_arg);
			}
			q_p->t_msg->u.tmr.sts = tmr_e_abort;
			send_n_delete( prev_p, q_p );
			msg_found = 1;		/* mark that we have found */
		    }
		    else
		    {   prev_p=q_p;	/* if current not deleted, it becomes prev for future send_n_delete */
		    }
		}
		if (! msg_found)
		    logMsg(  "warning: msg not found - cmd=%d\n"
			   , msg_p->u.tmr.cmd,0,0,0,0,0 );
		break;
	    case tmr_e_status:
		/* status for the first msg it finds */
		for (prev_p=q_p=head; q_p->next!=NULL; q_p=q_p->next)
		{   if (  (q_p->t_msg->u.tmr.tid==msg_p->u.tmr.tid)
			&&(q_p->t_msg->u.tmr.mid==msg_p->u.tmr.mid) )
		    {   
			if (msg_p->u.tmr.return_q != 0)
			{   msg_p->u.tmr.sts = tmr_e_status;
			    msg_p->u.tmr.tick_cnt = q_p->t_msg->u.tmr.tick_cnt;
			    /* mid and tid must already be set, and they
			       completely identify the msg , so nothing
			       else need be set?? */
			    if (typ == t_e_ptr)
			    {   sts = msgQSend(  msg_p->u.tmr.return_q, (char *)&(msg_p)
					       , sizeof(msg_p), NO_WAIT, MSG_PRI_NORMAL );
				if (sts == ERROR)  printErrno(errno);
				TDBG( "sent %d bytes", sizeof(msg_p), 0 );
			    }
			    else
			    {   sts = msgQSend(  msg_p->u.tmr.return_q, (char *)(msg_p)
					       , sizeof(*msg_p), NO_WAIT, MSG_PRI_NORMAL );
				if (sts == ERROR)  printErrno(errno);
				msgQSend( cpy_q, (char *)&(msg_p), sizeof(msg_p)
					 , NO_WAIT, MSG_PRI_NORMAL );
				TDBG( "sent %d bytes", sizeof(*msg_p), 0 );
			    }
			}
		        else
			{   logMsg( "warning: return message q == 0\n", 0,0,0,0,0,0 );
			}
			break;			/* WE CAN QUIT NOW */
		    }
		    else
		    {   prev_p=q_p;	/* if current not deleted, it becomes prev for future send_n_delete */
		    }
		}
		if (q_p->next == NULL)
		    logMsg(  "warning: msg not found - cmd=%d\n"
			   , msg_p->u.tmr.cmd,0,0,0,0,0 );
		break;
	    default:
		/* assume abort */
		/* abort just/only aborts the first (one) msg it finds */
		for (prev_p=q_p=head; q_p->next!=NULL; q_p=q_p->next)
		{   if (  (q_p->t_msg->u.tmr.tid==msg_p->u.tmr.tid)
			&&(q_p->t_msg->u.tmr.mid==msg_p->u.tmr.mid) )
		    {   if (   (q_p->t_msg->u.tmr.cmd==tmr_e_add_func)
				&& (*q_p->t_msg->u.tmr.func_p) )
			{   (*q_p->t_msg->u.tmr.func_p)(tmr_func_e_abort, q_p->t_msg->u.tmr.func_arg);
			}
			q_p->t_msg->u.tmr.sts = tmr_e_abort;
			send_n_delete( prev_p, q_p );
			break;			/* WE CAN QUIT NOW */
		    }
		    else
		    {   prev_p=q_p;	/* if current not deleted, it becomes prev for future send_n_delete */
		    }
		}
		if (q_p->next == NULL)
		    logMsg(  "warning: msg not found - cmd=%d\n"
			   , msg_p->u.tmr.cmd,0,0,0,0,0 );
		/* if not found above, the task will have been sent the
		   message. It is assumed that in this case, the abort msg and
		   the expire msg have crossed paths and the sending task has
		   not yet checked his mailbox and that the normal expire msg
		   will be waiting (he can do a msgQReceive with WAIT_FOREVER
		   */ 
	    }
	    if ((typ==t_e_cpy) && (  (msg_p->u.tmr.cmd==tmr_e_add)
				   ||(msg_p->u.tmr.cmd==tmr_e_add_func)))
	    {   /* get another buffer (for the next msg) to copy an entire
		   tmr_msg into */
		sts = msgQReceive(  cpy_q, (char *)&r_msg_p, sizeof(r_msg_p)
				  , NO_WAIT);
		if (sts == ERROR)
		{   logMsg( "fatal\n", 0,0,0,0,0,0 );
		    taskSuspend( 0 );
		}
	    }
	}

z=0;
	for(prev_p=q_p=head; q_p->next!=NULL; )
	{
z++;
	    if (q_p->t_msg->u.tmr.tick_cnt == 0)
	    {   if (   (q_p->t_msg->u.tmr.cmd==tmr_e_add_func)
		    && (*q_p->t_msg->u.tmr.func_p) )
		{   (*q_p->t_msg->u.tmr.func_p)(tmr_func_e_expire, q_p->t_msg->u.tmr.func_arg);
		}
		q_p->t_msg->u.tmr.sts = q_p->t_msg->u.tmr.cmd; /* FINAL STAT */
		send_n_delete( prev_p, q_p );
		if (q_p==prev_p)			/* was at head */
		{   q_p = head;			/* NOW UPDATED */
		}
		else
		{   q_p = prev_p->next;		/* NOW UPDATED */
		}
	    }
	    else
	    {   if (q_p->t_msg->u.tmr.cmd == tmr_e_add_func)
		{   if (q_p->t_msg->u.tmr.sts != tmr_e_func_started)
		    {   if (*q_p->t_msg->u.tmr.func_p)
			    (*q_p->t_msg->u.tmr.func_p)(tmr_func_e_start, q_p->t_msg->u.tmr.func_arg);
			q_p->t_msg->u.tmr.sts = tmr_e_func_started;
		    }
		    else
		    {   if (*q_p->t_msg->u.tmr.func_p)
			    (*q_p->t_msg->u.tmr.func_p)(tmr_func_e_during, q_p->t_msg->u.tmr.func_arg);
		    }
		}
		q_p->t_msg->u.tmr.tick_cnt--;
		prev_p=q_p;	/* only save last ptr if it wasn't deleted */
		q_p=q_p->next;
	    }
	}
    }
}   /* timerTask */


/******************************************************************************
 * @+Public+@
 * ROUTINE: timerStart:  Added by ron on 12-Apr-1995
 *
 * DESCRIPTION
 *       simple (example) way to start timerTask.  Can be used from shell
 *
 * RETURN VALUES:       None.
 *
 * SIDE EFFECTS:        None
 *
 * SYSTEM CONCERNS:     The timerTask should not be CPU intentive, but to
 *                      insure fairly accurate timing, it's priority should
 *                      not be too low (high number).
 *
 ******************************************************************************/  


STATUS	timerStart(  MSG_Q_ID	mQ
		   , int	q_sz
		   , int	prio )
{							/* @-Public-@ */
	STATUS	s;

    if (prio == 0) prio = 5;	/* limit ? */
    s = taskSpawn(  "tTimerTask"	/* name  - should be <= 10 chars */
		  , prio		/* priority */
		  , 0			/* options - none, esp.not flting pnt*/
		  , 5000		/* stack - small to medium */
		  , (FUNCPTR)timerTask
		  , (int)mQ, q_sz	/* options - could be zero */
		  , 0,0,0,0,0,0 );	/* more options - all zero */
    return (s);
}   /* timerStart */


/******************************************************************************
 * @+Public+@
 * ROUTINE: timerSend:  Added by ron on 12-Apr-1995
 *
 * DESCRIPTION
 *       format and send a message to the timerTask.  This basically fill 5
 *       consecutive integer locations in memory (corresponding to the order
 *       of the parameters) and sends this (plus 2 extra integer locations) to
 *       the timerTask.
 *
 * RETURN VALUES:       status from msgQSend
 *
 ******************************************************************************/  


STATUS	timerSend(  int		msg_type
		  , int		cmd
		  , int		tick_cnt
		  , int		mid
		  , MSG_Q_ID	return_q )
{							/* @-Public-@ */
	struct s_tmr_msg_fmt	msg;

    msg.msg_type       = msg_type; /* the user will most likely key off this */
    msg.u.tmr.cmd      = cmd;
    msg.u.tmr.tick_cnt = tick_cnt;
    msg.u.tmr.mid      = mid;	   /* 2nd level user id */
    msg.u.tmr.return_q = return_q;
    msg.u.tmr.tid      = taskIdSelf();

    return (msgQSend(  g_msg_q, (char *)&msg, sizeof(msg), NO_WAIT
		     , MSG_PRI_NORMAL ));
}   /* timerSend */


/******************************************************************************
 * @+Public+@
 * ROUTINE: timerSendP:  Added by ron on 12-Apr-1995
 *
 * DESCRIPTION
 *       Format and send a message to the timerTask using the msg ptr provided.
 *       This routine sends just the pointer and not the copy of the message.
 *
 * RETURN VALUES:       status from msgQSend
 *
 ******************************************************************************/  


STATUS	timerSendP(  struct s_tmr_msg_fmt	*msg
		   , int			 cmd
		   , int			 tick_cnt
		   , int			 mid
		   , MSG_Q_ID			 return_q )
{							/* @-Public-@ */
    /* the user can fill in msg_type */
    msg->u.tmr.cmd      = cmd;
    msg->u.tmr.tick_cnt = tick_cnt;
    msg->u.tmr.tid      = taskIdSelf();
    msg->u.tmr.mid      = mid;
    msg->u.tmr.return_q = return_q;

    return (msgQSend(  g_msg_q, (char *)&msg, sizeof(msg), NO_WAIT
		     , MSG_PRI_NORMAL ));
}   /* timerSendP */


/******************************************************************************
 * @+Public+@
 * ROUTINE: timerSendF:  Added by ron on 12-Apr-1995
 *
 * DESCRIPTION
 *	format and send a Full message (or message with Function args) to the
 *	timerTask.  This basically fill 5 consecutive integer locations in
 *	memory (corresponding to the order of the parameters) and sends this
 *	(plus 2 extra integer locations) to the timerTask.
 *
 * RETURN VALUES:       status from msgQSend
 *
 ******************************************************************************/  


STATUS	timerSendF(  int	 msg_type
		   , int	 cmd
		   , int	 tick_cnt
		   , int	 mid
		   , MSG_Q_ID	 return_q
		   , TMRFUNC	*func_p
		   , int	 func_arg )
{							/* @-Public-@ */
	struct s_tmr_msg_fmt	msg;

    msg.msg_type	= msg_type;	/* the user will most ... */
    msg.u.tmr.cmd	= cmd;		/*...likely key off msg_type */
    msg.u.tmr.tick_cnt	= tick_cnt;
    msg.u.tmr.mid	= mid;		/* 2nd level user id */
    msg.u.tmr.return_q	= return_q;
    msg.u.tmr.tid	= taskIdSelf();
    msg.u.tmr.func_p	= func_p;
    msg.u.tmr.func_arg	= func_arg;

    return (msgQSend(  g_msg_q, (char *)&msg, sizeof(msg), NO_WAIT
		     , MSG_PRI_NORMAL ));
}   /* timerSend */


/******************************************************************************
 * @+Public+@
 * ROUTINE: timerSendFP:  Added by ron on 12-Apr-1995
 *
 * DESCRIPTION
 *	Format and send a Full message (or message with Function args) to the
 *	timerTask using the msg ptr provided. This routine sends just the
 *	pointer and not the copy of the message. 
 *
 * RETURN VALUES:       status from msgQSend
 *
 ******************************************************************************/  


STATUS	timerSendFP(  struct s_tmr_msg_fmt	*msg
		    , int			 cmd
		    , int			 tick_cnt
		    , int			 mid
		    , MSG_Q_ID			 return_q
		    , TMRFUNC			*func_p
		    , int			 func_arg )
{							/* @-Public-@ */
    /* the user can fill in msg_type */
    msg->u.tmr.cmd	     = cmd;
    msg->u.tmr.tick_cnt      = tick_cnt;
    msg->u.tmr.tid	     = taskIdSelf();
    msg->u.tmr.mid	     = mid;
    msg->u.tmr.return_q      = return_q;
    msg->u.tmr.func_p	= func_p;
    msg->u.tmr.func_arg	= func_arg;

    return (msgQSend(  g_msg_q, (char *)&msg, sizeof(msg), NO_WAIT
		     , MSG_PRI_NORMAL ));
}   /* timerSendP */


/******************************************************************************
 * @+Public+@
 * ROUTINE: timerShow:  Added by ron on 12-Apr-1995
 *
 * DESCRIPTION
 *       show timer messages qued
 *
 * SYSTEM CONCERNS:     This is only for debugging as there is no guard against
 *                      a message alarm occuring during printing.
 *
 ******************************************************************************/  


void timerShow()
{							/* @-Public-@ */
	struct s_que	*q_p;
	int		 entries=0;

    printf( "head=0x%x tail=0x%x\n", (int)head, (int)tail );
    for (q_p=head; q_p->next!=NULL; )
    {   entries++;
	printf(  "\
entry=%d q_p=0x%x q_p->next=0x%x q_p->snd_typ=%d q_p->t_msg=0x%x\n\
	msg_type=%d\n\
	 tick_cnt=%d\n\
	 tid=0x%x\n\
	 mid=%d\n\
	 return_q=0x%x\n", entries, (int)q_p, (int)q_p->next, q_p->snd_typ
	       , (int)q_p->t_msg
	       , q_p->t_msg->msg_type
	       , q_p->t_msg->u.tmr.tick_cnt
	       , q_p->t_msg->u.tmr.tid
	       , q_p->t_msg->u.tmr.mid
	       , (int)q_p->t_msg->u.tmr.return_q );
	q_p=q_p->next;
    }
}   /* timerShow */


/******************************************************************************
 * @+Public+@
 * ROUTINE: timerTest:  Added by ron on 19-Jan-1996
 *
 * DESCRIPTION
 *	Tests timer task.       
 *
 ******************************************************************************/  

/* static - not seen by vxworks shell, but seen by debuggers */
static	MSG_Q_ID	timerTest_msgQ = 0;
static	int		timerTest_current_duration = 0;
static	int		timerTest_target_duration  = 0;
static	int		timerTest_duration_loops   = 0;


static
void	timerTest_func(   enum e_tmr_func	time
			, int			test_num )
{
	struct s_tmr_msg_fmt	msg;

    switch (time)
    {
    case tmr_func_e_start:
	if (test_num == 0) break;
	logMsg( "hello from start  tickGet=%d curr/targ= %d/%d\n", tickGet()
	       , timerTest_current_duration, timerTest_target_duration, 0,0,0 );
	timerTest_current_duration++; /* just start and during should inc -
					 says "there will be another tick
					 duration" */
	break;
    case tmr_func_e_during:
	if (test_num == 0) break;
	timerTest_current_duration++;
	break;
    case tmr_func_e_expire:
	logMsg( "hello from expire tickGet=%d curr/targ= %d/%d\n", tickGet()
	       , timerTest_current_duration, timerTest_target_duration, 0,0,0 );
	if (timerTest_duration_loops--)
	{   timerSendF(  0				/* msg_type	*/
		       , tmr_e_add_func			/* cmd		*/
		       , timerTest_target_duration	/* tick_cnt	*/
		       , 0				/* mid		*/
		       , 0				/* return_q	*/
		       , timerTest_func			/* func_p	*/
		       , test_num );			/* func_arg	*/
	}
	else
	    msgQSend(  timerTest_msgQ, (char *)&msg, sizeof(msg), NO_WAIT
		     , MSG_PRI_NORMAL );
	break;
    default:	/* assume case tmr_func_e_abort: */
	logMsg( "hello from abort  tickGet=%d curr/targ= %d/%d\n", tickGet()
	       , timerTest_current_duration, timerTest_target_duration, 0,0,0 );
	break;
    }
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


int	timerTask_test(  int	target_duration
		       , int	duration_loops ) /* like retries not tries */
{							/* @-Public-@ */
	struct s_tmr_msg_fmt	msg;


    if (! timerTest_msgQ)
    {   
#if 0
	timerStart( 0, 40, 10 );
#endif
	timerTest_msgQ = msgQCreate( 2, sizeof(msg), MSG_Q_FIFO );
    }


    timerTest_current_duration = 0;
    timerTest_target_duration  = target_duration;
    timerTest_duration_loops   = duration_loops;

    timerSendF(  0				/* msg_type	*/
	       , tmr_e_add_func			/* cmd		*/
	       , timerTest_target_duration	/* tick_cnt	*/
	       , 0				/* mid		*/
	       , 0 /*timerTest_msgQ - expire! *//* return_q	*/
	       , timerTest_func			/* func_p	*/
	       , 0 );				/* func_arg	*/

    /* must wait for first test to complete b/c they reference the same */
    msgQReceive( timerTest_msgQ, (char *)&msg, sizeof(msg), WAIT_FOREVER );
    logMsg( "Test Complete: tickGet=%d curr/targ= %d/%d\n", tickGet()
	   , timerTest_current_duration, timerTest_target_duration, 0,0,0 );


    timerTest_current_duration = 0;
    timerTest_target_duration  = target_duration;
    timerTest_duration_loops   = duration_loops;

    timerSendF(  0				/* msg_type	*/
	       , tmr_e_add_func			/* cmd		*/
	       , timerTest_target_duration	/* tick_cnt	*/
	       , 0				/* mid		*/
	       , 0 /*timerTest_msgQ - expire! *//* return_q	*/
	       , timerTest_func			/* func_p	*/
	       , 1 );				/* func_arg	*/

    /* wait for test to complete */
    msgQReceive( timerTest_msgQ, (char *)&msg, sizeof(msg), WAIT_FOREVER );
    logMsg( "Test Complete: tickGet=%d curr/targ= %d/%d\n", tickGet()
	   , timerTest_current_duration, timerTest_target_duration, 0,0,0 );


    timerTest_current_duration = 0;
    timerTest_target_duration  = target_duration;
    timerTest_duration_loops   = duration_loops;

    timerSendF(  0				/* msg_type	*/
	       , tmr_e_add_func			/* cmd		*/
	       , timerTest_target_duration	/* tick_cnt	*/
	       , 0				/* mid		*/
	       , 0				/* return_q	*/
	       , timerTest_func			/* func_p	*/
	       , 1 );				/* func_arg	*/

    /* delay one plus because that is basically what the timerTask does (to
       sync up on a tick boundary to get the most accuate timig */
    taskDelay( 1 + timerTest_target_duration/2 );

    /* if the delay is 1 tick and this task has higher priority than the
       timerTask, the abort msg will be sent before the timer task wakes
       up. When the timer task wake up, it will end up processing the first
       timer msg and the abort msg at the same time - the first msg will be
       put into the list, but then taken off before the timer process phase */

    timerSend(  0		/* msg_type	*/
	      , tmr_e_abort	/* cmd		*/
	      , 0		/* tick_cnt -ignored w/ abort	*/
	      , 0		/* mid		*/
	      , 0 );		/* return_q	*/

    logMsg( "tickGet=%d curr/targ= %d/%d\n", tickGet()
	   , timerTest_current_duration, timerTest_target_duration, 0,0,0 );

    taskDelay( 2 ); /* let timerTask process abort */

    /* get to see if the expire routine send a msg before we could abort it */
    if (msgQReceive(timerTest_msgQ,(char *)&msg,sizeof(msg),NO_WAIT) != ERROR)
	logMsg( "expire routine occured: tickGet=%d curr/targ= %d/%d\n", tickGet()
	       , timerTest_current_duration, timerTest_target_duration, 0,0,0 );
    else
	logMsg( "after abort complete: tickGet=%d curr/targ= %d/%d\n", tickGet()
	       , timerTest_current_duration, timerTest_target_duration, 0,0,0 );

    if (timerTest_current_duration >= timerTest_target_duration)
    {    logMsg(  "current duration is now >= target duration ... exit\n"
		, 0,0,0,0,0,0 );
	 taskDelay( sysClkRateGet()*3 ); /* let logTask spit out msgs for
					    cleaner output - keep prompt out
					    of the way */
	 return (1);
    }

    timerSendF(  0				/* msg_type	*/
	       , tmr_e_add_func			/* cmd		*/
	       , timerTest_target_duration-timerTest_current_duration
						/* tick_cnt	*/
	       , 0				/* mid		*/
	       , 0				/* return_q	*/
	       , timerTest_func			/* func_p	*/
	       , 1 );				/* func_arg	*/

    /* wait for test to complete */
    msgQReceive( timerTest_msgQ, (char *)&msg, sizeof(msg), WAIT_FOREVER );
    logMsg( "Test Complete: tickGet=%d curr/targ= %d/%d\n", tickGet()
	   , timerTest_current_duration, timerTest_target_duration, 0,0,0 );


    taskDelay( sysClkRateGet()*3 ); /* let logTask spit out msgs for cleaner
				       output - keep prompt out of the way */

    return (1);
}   /* timerShow */

