/************************************************************************/
/* File: tod_device.c							*/
/* Time-Of-Day devices that are implemented				*/
/* with the FSCC PC4 (DS1386), FSCC PC3 (DS1286), MVME167 MK48T08	*/
/*									*/
/* Set GMT time (seconds since January 1 1970)				*/
/* Get GMT time (seconds since January 1 1970)				*/
/*									*/

/*
SCCS release:
@(#) tod_device.c 1.9 Delta: 94/01/11 14:29:52 Extraction: 94/01/11 14:43:04 @(#)
*/

/************************************************************************/
/* Copyright 1990 MBARI							*/
/************************************************************************/
/* $Header: /cvs/ols/vx_tools/src/tod_device.c,v 1.3 1996/12/11 20:55:37 vittone Exp $			*/
/* Summary  : Functions to read time from MK48T02 chip on SBE VPU-30 	*/
/*		under vxWorks						*/
/* Filename : rtc.c							*/
/* Author   : Bob Herlien (rah)						*/
/* Project  : New ROV							*/
/* $Revision: 1.3 $							*/
/* Created  : 02/28/91							*/
/************************************************************************/
/* Modification History:						*/
/* $Log: tod_device.c,v $
 * Revision 1.3  1996/12/11  20:55:37  vittone
 * Got rid of conditionals on old VXWORKS_5_0_2
 *
 * Revision 1.2  1995/10/05  21:26:57  vittone
 * adding MV162 supporT
 *
 * Revision 1.1.1.1  1995/03/17  22:40:43  vittone
 * Initial import
 *
 * Revision 1.2  91/09/23  09:38:59  09:38:59  hebo (Bob Herlien 408-647-3748)
 * Minor bug fixes, added -DMICROTIME
 * 
 * Revision 1.1  91/09/06  16:05:35  16:05:35  hebo (Bob Herlien 408-647-3748)
 * Initial revision
 * 
 * Revision 1.0  91/08/09  11:10:33  11:10:33  hebo (Bob Herlien 408-647-3748)
 * Initial revision
 * 	*/
/* 28feb91 rah, created							*/
/* 20dec93 das, changed include of timevw.h to time.h			*/
/************************************************************************/
/* The Unix Epoch (day 0) is Jan 1, 1970.  The real-time clock chip	*/
/* only gives 2 decimal digits of year.  Also, a Unix time_t value	*/
/* (signed long) can only go +/- 68 years from 1970.  Therefore, these	*/
/* routines assume that the year is between 1970 and 2070.  Note that	*/
/* the leap-year exception for centuries not divisible by 400 is not a	*/
/* factor in this period.						*/
/************************************************************************/

#include <vxWorks.h>			/* VxWorks types		    */

#include <time.h>			/* time defines			    */

#define INCstddefh			/* bypass second size_t declaration */


#ifndef MIPS_R3000
#include <stdioLib.h>			/* VxWorks standard I/O library	    */
#else
#include <fioLib.h>
#endif
#include <string.h>			/* VxWorks string types		    */
#include <sysLib.h>			/* VxWorks Driver defines	    */
#include <limits.h>			/* VxWorks integer limits	    */


#include "vx_tools.h"			/* hardware definitions */

#include "tod_time.h"			/* other time structure definitions */
#include "tod_prototypes.h"		/* tod function prototypes          */


/* Defines for MK48T02		Time-of-Day Clock Chip on MVME147 */
/* Defines for MK48T08(B)	Time-of-Day Clock Chip on MVME167 */
/* Defines for DS1386		Time-of-Day Clock Chip on FSCC */
/* Defines for DS1386		Time-of-Day Clock Chip on FRC */

/*
Hardware definitions moved to "vx_tools.h"
#if	defined (fscc)
#include "fscc.h"
#elif	defined (frc)
#include "frc.h"
#elif	defined (mv147)
#include "mv147.h"
#elif	defined (mv167)
#include "mv167.h"
#elif	defined (mv162)
#include "mv162.h"
#endif
*/

/* Device driver Time-of-Day macros FSCC, FRC, MV147, MV167 */

#ifdef DS1386
#include <tod_ds1386.h>			/* Time-of-Day DS Device driver */
#endif
#ifdef DS1286
#include <tod_ds1386.h>			/* Time-of-Day DS Device driver */
#endif
#ifdef MK48T02
#include <tod_mk48t08.h>		/* Time-of-Day SGS Device driver */
#endif
#ifdef MK48T08
#include <tod_mk48t08.h>		/* Time-of-Day SGS Device driver */
#endif

/************************************************************************/
/*									*/
/* Special Case, NO Time-Of-Day Calendar Device				*/
/*									*/

#ifndef TOD_DEVICE_ID			/* No driver was specified So, */
					/* use the System Clock */
#define TOD_DEVICE_ID "SysClk\0"

#ifndef TOD_SECS
#define TOD_SECS	1		/* tick increment period in seconds */
#endif

#ifndef TOD_NONE			/* Let the rest of the code know */
#define TOD_NONE
#endif

#include <usrLib.h>			/* VxWorks Driver defines */
#include <taskLib.h>			/* VxWorks Driver defines */
#endif /* tod_device_id */


/************************************************************************/
/*	External Functions						*/
/************************************************************************/
#ifndef INCtod_prototypesh
extern INT32 tod_julday( INT32 yr, INT32 mo, INT32 day );
extern INT32 tod_ep_days( INT32 year, INT32 yday );
#endif /*INCtod_prototypesh*/
/************************************************************************/
/* Prototypes								*/
/*									*/
#ifndef INCtod_prototypesh
void	tod_device_id(void);
					/* Touch Hardware Prototypes */
void	tod_device_on(void);
void	tod_device_off(void);
time_t	tod_device_read( struct tm *tp );
time_t	tod_read(void);
void	tod_device_write( const struct tm *tp );
void	tod_write(time_t timer);
#endif /*INCtod_prototypesh*/
/************************************************************************/
/* Global Data								*/
/*									*/

#ifdef TOD_NONE				/* System Clock NO TOD device */
					/* SysClk Globals */
LOCAL time_t tod_atick = 0;		/* seconds @ Jan 1, 1970 12:00 */
LOCAL int tod_task = 0;			/* FLAG onetime spawn */
LOCAL int tod_taskid = 0;		/* task identifier running */

					/* System Clock TOD functions */
LOCAL void tod_settick(time_t tick)	{	tod_atick = tick;	}
LOCAL time_t tod_gettick(void)		{	return (tod_atick);	}
LOCAL void tod_SysClk(void) { 
	int looptime;
	looptime = TOD_SECS * sysClkRateGet();
	for (;;) {
		tod_atick += TOD_SECS;
		taskDelay(looptime); 
	} /* forever */
} /* tod_sysclk */
#endif /* tod_none */

/************************************************************************/
/*									*/
/* Routine TOD_DEVICE_ID	Display Time-of-Day driver		*/
/*									*/
/* Arguments:								*/
/*									*/
/*		None							*/
/*									*/
void tod_device_id(void)
{

	printf("Time-Of-Day Device Driver: %s\n",TOD_DEVICE_ID);

} /* tod_device_id */

/************************************************************************/
/*									*/
/* Routine TOD_DEVICE_ON	Directly turn the Time-of-Day device ON	*/
/*									*/
/* Arguments:								*/
/*									*/
/*		None							*/
/*									*/
void tod_device_on(void)
{
#ifndef TOD_NONE
	TOD_DEVICE_ON;
#else
	char *tod_Clk = "tod_Clk";

	if (!tod_task)
	{
		taskSpawn(tod_Clk,100,0,800,(FUNCPTR) tod_SysClk,
			0,0,0,0,0,0,0,0,0,0);
		tod_taskid = taskNameToId(tod_Clk);
		tod_task = 1;
		/*printf("taskid = %x\n",tod_taskid);*/
	}
#endif
} /* tod_device_on */

/************************************************************************/
/*									*/
/* Routine TOD_DEVICE_OFF	Directly turn the Time-of-Day device OFF*/
/*									*/
/* Arguments:								*/
/*									*/
/*		None							*/
/*									*/
void tod_device_off(void)
{

#ifndef TOD_NONE
	TOD_DEVICE_OFF;
#else
	if (tod_task)
	{
		taskDelete(tod_taskid);
		tod_task = 0;
	}
#endif

} /* tod_device_off */

/************************************************************************/
/* Function    : tod_device_read					*/
/* Purpose     : Read the RTC chip, put the results in struct tm *tp	*/
/* Inputs      : Pointer to struct tm to put current time into		*/
/* Outputs     : Time in time_t format					*/
/************************************************************************/
	time_t
tod_device_read( struct tm *tp )
{
#ifndef TOD_NONE
	struct tm temp_tm;
	time_t rtn;

	if ( tp == (struct tm *)NULL )	/* If no tm, use temp one	*/
		tp = &temp_tm;

	TOD_READ_DEVICE(tp);

	if ( tp->tm_year < TOD_YEAR0 )	/* RTC_YR 00 is actual year 2000 */
		tp->tm_year += 100;

	tp->tm_yday = tod_julday( (tp->tm_year % 100), tp->tm_mon, tp->tm_mday ) - 1;

	tp->tm_isdst = 0;

	rtn = tod_ep_days(tp->tm_year - TOD_YEAR0, tp->tm_yday) * TOD_SECS_PER_DAY;
	rtn += (tp->tm_hour * 3600L) + (tp->tm_min * 60) + tp->tm_sec;

	return( rtn );
#else

	if (!tod_task) tod_device_on();		/* if off turn it on */
	return(tod_gettick());
#endif

} /* tod_device_read() */

/************************************************************************/
/* Function    : tod_device_write					*/
/* Purpose     : Write the RTC chip from the values of struct tm *tp	*/
/* Inputs      : Pointer to struct tm that contains current time	*/
/* Outputs     : None							*/
/************************************************************************/
	void
tod_device_write( const struct tm *tp )
{
#ifndef TOD_NONE

	TOD_WRITE_DEVICE(tp);

#else
	time_t rtn;

	rtn = tod_ep_days(tp->tm_year - TOD_YEAR0, tp->tm_yday) * TOD_SECS_PER_DAY;
	rtn += (tp->tm_hour * 3600L) + (tp->tm_min * 60) + tp->tm_sec;

	if (!tod_task) tod_device_on();		/* if off turn it on */
	tod_settick(rtn);
#endif

} /* tod_device_write() */

/************************************************************************/
/*									*/
/* Routine TOD_READ		Read a "time" in GMT Seconds since	*/
/* Jan 1, 1970 directly to the device					*/
/*									*/
/* Arguments:								*/
/*									*/
/*		none							*/
/*									*/
/* Return:								*/
/*									*/
/*		rtn		time_t	(Seconds since Jan 1, 1970)	*/
/*									*/
time_t tod_read(void)
{

	return(tod_device_read(NULL));

} /* tod_read */

/************************************************************************/
/*									*/
/* Routine TOD_WRITE		Write a "time" in GMT Seconds since	*/
/* Jan 1, 1970 directly to the device					*/
/*									*/
/* Arguments:								*/
/*									*/
/*	(in)	timer		time_t	(Seconds since Jan 1, 1970)	*/
/*									*/
void tod_write(time_t timer)
{
#ifndef TOD_NONE
	INT32	remainder, days, years, months;
	struct tm tmp;
	struct tm *tp;

	tp = &tmp;

	days = timer / TOD_SECS_PER_DAY;	/* Get days portion of *timer */
	remainder = timer % TOD_SECS_PER_DAY;	/* rem is time of that day */

	years = days / 365;			/* 1st approx of year has no leap */
	while ( tod_ep_days(years,0) > days )	/* Loop to refine approx */
	years --;

	tp->tm_wday = (days + TOD_DAY0) % 7;	/* Get day of week */
	days -= tod_ep_days(years, 0);		/* Subtract years to get yday */
	years += TOD_YEAR0;			/* Normalize year to 1900 */

	for ( months = 0; (days >= tod_julday(years,months+1,0)) && (months < 11); months++ )
	{;}					/* Loop to find month */

	tp->tm_yday = days;			/* Store unix julian day */
	tp->tm_year = years;			/* Store year */
	tp->tm_mon = months;			/* Store month */
	tp->tm_mday = days - tod_julday(years,months,0) + 1;
						/* Store day of month */
	tp->tm_hour = remainder / 3600;		/* Store hour */
	remainder %= 3600;			/* Get remaining seconds */
	tp->tm_min = remainder / 60;		/* Store minutes */
	tp->tm_sec = remainder % 60;		/* Store remaining seconds */
	tp->tm_isdst = 0; /* daylight */	/* DST flag */

	tod_device_write(tp);
#else

	if (!tod_task) tod_device_on();		/* if off turn it on */
	tod_settick(timer);
#endif

} /* tod_write */

/* file: tod_device.c */
