/************************************************************************/
/* File: tod_func.c							*/
/*									*/
/* This is a modified version of the Bob Herlien's Unix like routines	*/
/* that operate with month and dayofmonth strings.			*/
/* For Specific Time-Of-Day devices that are implemented like the	*/
/* FSCC PC4 (DS1386), FSCC PC3 (DS1286), MVME167 MK48T08		*/
/*									*/

/*
SCCS release:
@(#) tod_func.c 1.5 Delta: 94/01/11 14:30:07 Extraction: 94/01/11 14:38:42 @(#)
*/

/************************************************************************/
/* Copyright 1990 MBARI							*/
/************************************************************************/
/* $Header: /cvs/ols/vx_tools/src/tod_func.c,v 1.2 1996/12/11 20:55:37 vittone Exp $		*/
/* Summary  : Unix-compatible Time Routines for SBE VPU-30 under vxWorks*/
/* Filename : usrTime.c							*/
/* Author   : Bob Herlien (rah)						*/
/* Project  : New ROV							*/
/* $Revision: 1.2 $							*/
/* Created  : 02/28/91							*/
/************************************************************************/
/* Modification History:						*/
/* $Log: tod_func.c,v $
 * Revision 1.2  1996/12/11  20:55:37  vittone
 * Got rid of conditionals on old VXWORKS_5_0_2
 *
 * Revision 1.1.1.1  1995/03/17  22:40:43  vittone
 * Initial import
 *
 * Revision 1.2  91/09/23  09:39:23  09:39:23  hebo (Bob Herlien 408-647-3748)
 * Minor bug fixes, added -DMICROTIME
 * 
 * Revision 1.1  91/09/10  11:15:42  11:15:42  hebo (Bob Herlien 408-647-3748)
 * Moved RTC sync from task level to interrupt level, minor bug fixes, rah
 * 
 * Revision 1.0  91/08/09  11:10:33  11:10:33  hebo (Bob Herlien 408-647-3748)
 * Initial revision
 * 	*/
/* 28feb91 rah, created							*/
/* 20dec93 das, changed include to timevw.h to time.h			*/

/************************************************************************/
/* The Unix Epoch (day 0) is Jan 1, 1970.  The real-time clock chip	*/
/* only gives 2 decimal digits of year.  Also, a Unix time_t value	*/
/* (signed long) can only go +/- 68 years from 1970.  Therefore, these	*/
/* routines assume that the year is between 1970 and 2070.  Note that	*/
/* the leap-year exception for centuries not divisible by 400 is not a	*/
/* factor in this period.						*/
/************************************************************************/

/************************************************************************/
/* Include files							*/
/*									*/

#include <vxWorks.h>			/* VxWorks types		*/

#include <time.h>			/* VxWorks time defines		*/

#include <tod_time.h>			/* time structure definitions 	*/

/************************************************************************/
/* Prototypes								*/
/*									*/

INT32	tod_daysinmonth(INT32 year, INT32 month);
char	*tod_asciiday(INT32 day);
char	*tod_asciimonth(INT32 mo);

INT32	tod_julday( INT32 yr, INT32 mo, INT32 day );
INT32	tod_ep_days( int year, int yday );

/************************************************************************/
/* Global Data								*/
/*									*/

static char	*tod_months[12] =	/* Month names */
{"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};

static char	*tod_days_of_week[7] =	/* Weekday names */
{"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};

static UINT16	tod_jdays[12] =		/* Days since Jan. 1 in non-leap yr */
{   0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334};

/************************************************************************/
/* 									*/
/* ROUTINE TOD_DAYSINMONTH	Determine the number of days in a month	*/
/*									*/
/* Arguments:								*/
/*									*/
/*	(in)	year 	INT32	the current year			*/
/*	(in)	month	INT32	the given month (0-11)			*/
/*									*/
/* Return:								*/
/*		-1	Invalid						*/
/*		INT32	Number of days in the month			*/
/*									*/
INT32 tod_daysinmonth(INT32 year, INT32 month)
{
	int febleap;
	int nodays;		/* maximum number of days in month */

	if ((month < 0) || (month > 11))
		return(-1);

	febleap = ((( year % 4) == 0) && (month == 1)) ? 1 : 0;

	if (month == 11)	/* determine the days in the month */
		nodays = 365 - (int)tod_jdays[month];
	else
		nodays = (int)tod_jdays[month+1] - (int)tod_jdays[month] + febleap;

	return (nodays);

} /* tod_daysofmonth */

/************************************************************************/
/* 									*/
/* ROUTINE TOD_ASCIIDAY		String of DAY(0 - 6 = "Sun"-"Sat")	*/
/*									*/
/* Arguments:								*/
/*									*/
/*	(in)	day 	INT32	Day of the week				*/
/*									*/
/* Return:								*/
/*		* char	Dayofweek string				*/
/*									*/
char * tod_asciiday(INT32 day)
{
	return(tod_days_of_week[day % 7]);

} /* tod_asciiday */

/************************************************************************/
/* 									*/
/* ROUTINE TOD_ASCIIMONTH	String of Month(0 - 11 = "Jan"-"Dec")	*/
/*									*/
/* Arguments:								*/
/*									*/
/*	(in)	mo 	INT32	Month of the year			*/
/*									*/
/* Return:								*/
/*		* char	Month string					*/
/*									*/
char * tod_asciimonth(INT32 mo)
{
	return(tod_months[mo % 12]);

} /* tod_asciimonth */

/************************************************************************/
/* Function    : tod_julday						*/
/* Purpose     : Calculate Julian Day given year, month, day		*/
/* Inputs      : year (00 - 99), month (0 - 11), day (1 - 31)		*/
/* Outputs     : Julian day						*/
/* Comment     : Returns Julian day in range 1:366.  Unix wants 0:365	*/
/************************************************************************/
INT32 tod_julday( INT32 yr, INT32 mo, INT32 day )
{
    INT32		leap;

    leap = (((yr % 4) == 0) && (mo > 1)) ? 1 : 0;
    return( (int)tod_jdays[mo] + day + leap );

} /* tod_julday() */

/************************************************************************/
/* Function    : tod_ep_days						*/
/* Purpose     : Calculate number days since Unix epoch			*/
/* Inputs      : years since 1970, Unix-type julian day (0:365)		*/
/* Outputs     : Days since 1/1/1970					*/
/* Comment     : The (year + 1)/4 term accounts for leap years, the	*/
/*		 first of which was 1972 & should be added starting '73	*/
/************************************************************************/
INT32 tod_ep_days( int year, int yday )
{
    return( (365 * year) + ((year + 1)/4) + yday );

} /* tod_ep_days() */

/* file: tod_func.c */
