/************************************************************************/
/* File: tod_time.c							*/
/*									*/
/* This is a modified version of the Bob Herlien's Unix like routines	*/
/* written to use specific Time-Of-Day devices that are implemented	*/
/* with the FSCC PC4 (DS1386), FSCC PC3 (DS1286), MVME167 MK48T08	*/
/*									*/

/*
SCCS release:
@(#) tod_time.c 1.8 Delta: 94/01/11 14:30:42 Extraction: 94/01/11 14:51:47 @(#)
*/

/************************************************************************/
/* Copyright 1990 MBARI							*/
/************************************************************************/
/* $Header: /cvs/ols/vx_tools/src/tod_time.c,v 1.3 1996/12/11 20:55:38 vittone Exp $		*/
/* Summary  : Unix-compatible Time Routines for SBE VPU-30 under vxWorks*/
/* Filename : usrTime.c							*/
/* Author   : Bob Herlien (rah)						*/
/* Project  : New ROV							*/
/* $Revision: 1.3 $							*/
/* Created  : 02/28/91							*/
/************************************************************************/
/* Modification History:						*/
/* $Log: tod_time.c,v $
 * Revision 1.3  1996/12/11  20:55:38  vittone
 * Got rid of conditionals on old VXWORKS_5_0_2
 *
 * Revision 1.2  1996/03/07  16:47:01  vittone
 *  Fixing a formattting problme in asctime.
 * 	CVS: EntER lOG.  lINES BEGINNING WITH `cvs: ' ARE REMOVED AUTOMATICALLY
 *
 * Revision 1.1.1.1  1995/03/17  22:40:43  vittone
 * Initial import
 *
 * Revision 1.2  91/09/23  09:39:23  09:39:23  hebo (Bob Herlien 408-647-3748)
 * Minor bug fixes, added -DMICROTIME
 * 
 * Revision 1.1  91/09/10  11:15:42  11:15:42  hebo (Bob Herlien 408-647-3748)
 * Moved RTC sync from task level to interrupt level, minor bug fixes, rah
 * 
 * Revision 1.0  91/08/09  11:10:33  11:10:33  hebo (Bob Herlien 408-647-3748)
 * Initial revision
 * 	*/
/* 28feb91 rah, created							*/
/* 20dec93 das, changed include of timevw.h to time.h			*/

/************************************************************************/
/* The Unix Epoch (day 0) is Jan 1, 1970.  The real-time clock chip	*/
/* only gives 2 decimal digits of year.  Also, a Unix time_t value	*/
/* (signed long) can only go +/- 68 years from 1970.  Therefore, these	*/
/* routines assume that the year is between 1970 and 2070.  Note that	*/
/* the leap-year exception for centuries not divisible by 400 is not a	*/
/* factor in this period.						*/
/************************************************************************/
#include <vxWorks.h>			/* VxWorks types		    */


#include <time.h>			/* time defines			    */

#define INCstddefh                      /* bypass second size_t declaration */
#ifndef MIPS_R3000
#include <stdioLib.h>			/* VxWorks standard I/O library	    */
#else
extern int           sprintf (char *s, const char *format, ...);
#include <fioLib.h>
#endif
#include <string.h>                     /* VxWorks string types             */
#include <sysLib.h>			/* VxWorks Driver defines	    */
#include <limits.h>			/* VxWorks integer limits	    */

#include "tod_time.h"			/* time definitions and structures */
#include "tod_user.h"			/* user definitions and structures */
#include "tod_prototypes.h"		/* tod_ function prototypes        */


/************************************************************************/
/* Global Data								*/
/*									*/

long	timezone = 0;			/* Initialized for GMT */
long	altzone = 0;			/* Initialized for GMT */
int	daylight = 0;			/* 1 if DST, 0 if not */
char	*tzname[2] = { "GMT", "GMT" };	/* Timezones: GMT (CST,DST) */

static struct tm tmbuf;			/* buffer in localtime() and gmtime() */

static struct timeval gm_time;		/* Time, kept as GMT time */

static char asctm_buf[TOD_ASCTM_SIZE];	/* Static ASCII buffer for asctime()*/

static struct tzrules altzoneBegin;	/* Start Rules for daylight */
static struct tzrules altzoneEnd;	/* End Rules for daylight */


/************************************************************************/
/* External routines internally used					*/
/*									*/
#ifndef INCtod_prototypesh
extern INT32	tod_daysinmonth(INT32 year, INT32 month);
extern char	*tod_asciiday(INT32 day);
extern char	*tod_asciimonth(INT32 mo);

extern INT32	tod_julday( INT32 yr, INT32 mo, INT32 day );
extern INT32	tod_ep_days( int year, int yday );

extern time_t   tod_read(void);
extern void     tod_device_write(const struct tm *tp);
#endif /* INCtod_prototypesh */

/************************************************************************/
/* Prototypes								*/
/*									*/

LOCAL int tod_RdWrNVRAM(int WrRd);
LOCAL int tod_tzrules_conditional(int year, int month, int day, int dayofweek);
LOCAL void tod_tzset(int timer);
LOCAL INT32 tod_sectime(INT32 yymmdd, INT32 hhmmss);
LOCAL void tod_setusage(void);

/************************************************************************/
/* Function    : time							*/
/* Purpose     : Unix-compatible get time function			*/
/* Inputs      : Ptr to place to put time (in seconds from 1/1/70)	*/
/* Outputs     : Time in seconds from 1/1/70				*/
/************************************************************************/
time_t time( time_t *tp )
{
    gm_time.tv_sec = tod_read();

    if ( tp != (time_t *)NULL )
	*tp = gm_time.tv_sec;

    return( gm_time.tv_sec );

} /* time() */

/************************************************************************/
/* Function    : get_tm							*/
/* Purpose     : Convert time_t to struct tm				*/
/* Inputs      : time_t value to convert, ptr to struct tm to store result*/
/* Outputs     : Ptr to filled in struct tm				*/
/************************************************************************/
struct tm * get_tm( const time_t timer, struct tm *tmp )
{
    INT32	rem, days, year, mo;

    days = timer / TOD_SECS_PER_DAY;	/* Get days portion of *timer	*/
    rem = timer % TOD_SECS_PER_DAY;		/* rem is time of that day	*/

    year = days/365;			/* 1st approx of year has no leap*/
    while ( tod_ep_days(year,0) > days )	/* Loop to refine approx	*/
	year--;

    tmp->tm_wday = (days + TOD_DAY0) % 7;/* Get day of week		*/
    days -= tod_ep_days(year, 0);		/* Subtract years to get yday	*/
    year += TOD_YEAR0;			/* Normalize year to 1900	*/

    for ( mo = 0; (days >= tod_julday(year,mo+1,0)) && (mo < 11); mo++ )
	{;}				/* Loop to find month		*/

    tmp->tm_yday = days;		/* Store unix julian day	*/
    tmp->tm_year = year;		/* Store year			*/
    tmp->tm_mon = mo;			/* Store month			*/
    tmp->tm_mday = days - tod_julday(year,mo,0) + 1;
					/* Store day of month		*/
    tmp->tm_hour = rem / 3600;		/* Store hour			*/
    rem %= 3600;			/* Get remaining seconds	*/
    tmp->tm_min = rem / 60;		/* Store minutes		*/
    tmp->tm_sec = rem % 60;		/* Store remaining seconds	*/
    tmp->tm_isdst = daylight;		/* DST flag			*/
    return( tmp );			/* Return ptr to struct tm	*/

} /* get_tm() */

/************************************************************************/
/* Function    : gmtime							*/
/* Purpose     : Unix-compatible function to convert time_t to struct tm*/
/* Inputs      : Ptr to time_t value to convert				*/
/* Outputs     : Ptr to filled in struct tm				*/
/************************************************************************/
struct tm *gmtime( const time_t *timer )
{

    return( get_tm(*timer, &tmbuf) );

} /* gmtime() */

/************************************************************************/
/* Function    : localtime						*/
/* Purpose     : Unix-compatible function to convert time_t to struct tm*/
/* Inputs      : Ptr to time_t value to convert				*/
/* Outputs     : Ptr to filled in struct tm				*/
/* Comment     : localtime() corrects for time zone			*/
/************************************************************************/
struct tm *localtime( const time_t *timer )
{

    tod_tzset(*timer);

    if (daylight)
	return( get_tm(*timer + altzone, &tmbuf) );
    else
	return( get_tm(*timer + timezone, &tmbuf) );

} /* localtime () */

/************************************************************************/
/* Function    : stime							*/
/* Purpose     : Unix-compatible set time function			*/
/* Inputs      : Ptr to time in seconds from 1/1/70			*/
/* Outputs     : 0							*/
/************************************************************************/
int stime( long *tp )
{
    gm_time.tv_sec = *tp;
    gm_time.tv_usec = 0;
    tod_device_write( gmtime((time_t *)tp) );
    return( 0 );

} /* stime() */

/************************************************************************/
/* Function    : asctime						*/
/* Purpose     : Unix-compatible function to return time in ASCII string*/
/* Inputs      : Ptr to tm struct containing time to convert		*/
/* Outputs     : Ptr to 26 UCHARacter ASCII string containing time	*/
/************************************************************************/
char *asctime( const struct tm *tmp )
{
    sprintf( asctm_buf, "%s %s %2d %2.2d:%2.2d:%2.2d %d\n%c",
	     tod_asciiday(tmp->tm_wday), tod_asciimonth(tmp->tm_mon),
	     tmp->tm_mday, tmp->tm_hour, tmp->tm_min, tmp->tm_sec,
	     tmp->tm_year + 1900, '\0' );

    return( asctm_buf );

} /* asctime() */

/************************************************************************/
/* Function    : ctime							*/
/* Purpose     : Unix-compatible function to convert time to ASCII	*/
/* Inputs      : Ptr to time in seconds from 1/1/70			*/
/* Outputs     : Ptr to 26 UCHARacter ASCII string containing time	*/
/************************************************************************/
char	 *ctime( const time_t *timer )
{
    return( asctime(localtime(timer)) );

} /* ctime() */

/************************************************************************/
/* Function    : gettimeofday						*/
/* Purpose     : Unix-compatible function to get time in usecs		*/
/* Inputs      : Ptr to struct timeval, ptr to struct timezone		*/
/* Outputs     : 0							*/
/************************************************************************/
int gettimeofday( struct timeval *tp, struct timezone *tzp )
{
    if ( tp != (struct timeval *)0 )
    {
	tp->tv_sec = gm_time.tv_sec;

#ifdef MICROTIME
	tp->tv_usec = gm_time.tv_usec - microtime() + tick;
	if ( tp->tv_usec >= TOD_USECS_PER_SEC )
	{
	    tp->tv_usec -= TOD_USECS_PER_SEC;
	    tp->tv_sec++;
	}

#else
	tp->tv_usec = gm_time.tv_usec;
#endif
    }

    if ( tzp != (struct timezone *)0 )
    {
	tzp->tz_minuteswest = -timezone / 60;
	tzp->tz_dsttime = daylight;
    }
    return( 0 );

} /* gettimeofday() */

/************************************************************************/
/* Function    : settimeofday						*/
/* Purpose     : Almost Unix-compatible function to set GMT time	*/
/* Inputs      : Ptr to struct timeval to set, ptr to struct timezone to set*/
/* Outputs     : 0 if success, -1 if failure				*/
/* Comments    : Sets timezone offset and daylight savings flag		*/
/************************************************************************/
int settimeofday( struct timeval *tp, struct timezone *tzp )
{
    if ( tzp != (struct timezone *)0 )
    {
	timezone = 60 * -tzp->tz_minuteswest;
	daylight = tzp->tz_dsttime;
    }

    stime( (long *)(&(tp->tv_sec)) );
    gm_time.tv_usec = tp->tv_usec;
#ifdef MICROTIME
    gm_time.tv_usec += microtime() - tick;
#endif
    return( 0 );

} /* settimeofday() */

/************************************************************************/
/*									*/
/* ROUTINE MKTIME		Try to reconstruct a Local time value	*/
/*									*/
/* Arguments:								*/
/*									*/
/*	(in/out)	*tp	struct tm	Date/Time to convert	*/
/*									*/
/* Return:								*/
/*									*/
/*	0 >= Seconds past Jan 1, 1970					*/
/*	-1 = Invalid (year, month, day, hour, minute, second)		*/
/*									*/
time_t mktime( struct tm *tp )
{
	int nodays;		/* maximum number of days in month */
	time_t rtn;		/* return time in seconds since 1/1/70 */

	if ((tp->tm_year < 0) || (tp->tm_year > 100))
	return (-1);

	if ((tp->tm_mon < 0) || (tp->tm_mon > 11))
	return (-1);

	if ((tp->tm_hour < 0) || (tp->tm_hour > 59))
	return (-1);

	if ((tp->tm_min < 0) || (tp->tm_min > 59))
	return (-1);

	if ((tp->tm_sec < 0) || (tp->tm_sec > 59))
	return (-1);
	
	nodays = tod_daysinmonth(tp->tm_year, tp->tm_mon);

	if ((tp->tm_mday < 1) || (tp->tm_mday > nodays))
	return (-1);

	/* Valid! Compute DayOfYear, time in seconds past Jan 1, 1970, and DayOfWeek */

	tp->tm_yday = tod_julday(tp->tm_year, tp->tm_mon, tp->tm_mday) -1; /* set day of year */

	rtn = tod_ep_days(tp->tm_year - TOD_YEAR0, tp->tm_yday) * TOD_SECS_PER_DAY;
	rtn += (tp->tm_hour * 3600L) + (tp->tm_min * 60) + tp->tm_sec;

	tp->tm_wday = (rtn / TOD_SECS_PER_DAY + TOD_DAY0) % 7;	/* day of the week */

	return(rtn);
	
} /* mktime */

/************************************************************************/
/* Function    : date							*/
/* Purpose     : User interface function to show time and date		*/
/* Inputs      : None							*/
/* Outputs     : 0							*/
/* Comment     : Displays time and date to stdout			*/
/************************************************************************/
INT32 date( void )
{
    char	*s;
    time_t	timer;

    time( &timer );
    s = asctime( gmtime(&timer) );
    s[strlen(s)-1] = '\0';
    printf( "%s GMT\n", s );

    s = asctime( localtime(&timer) );
    s[strlen(s)-1] = '\0';
    printf( "%s %s\n", s, (daylight == 0 ? tzname[0] : tzname[1]) );

    return( 0 );

} /* date() */

/************************************************************************/
/************************************************************************/
/************************************************************************/
/*			    USER ROUTINES				*/ 
/************************************************************************/
/************************************************************************/
/************************************************************************/

/************************************************************************/
/*									*/
/* User Set/Get Time-Of-Day and timezone information			*/
/*									*/

/************************************************************************/
/*									*/
/* ROUTINE TOD_INITIAL		Initialize Tod device, Zones and Rules	*/
/*									*/
/* Arguments:								*/
/*									*/
/*		none							*/
/* Return:								*/
/*									*/
/*	 0 = Success							*/
/*									*/
int tod_initial(void)
{

	char *GMTtz ="GMT";		/* default GMT */
	char *DSTtz ="CDT";		/* assuming Central USA */

	timezone = 0;
	altzone = 0;
	daylight = 0;

	tzname[0] = GMTtz;
	tzname[1] = DSTtz;

	altzoneBegin.month = 0;
	altzoneBegin.day = 0;
	altzoneBegin.hour = 0;
	altzoneBegin.minute = 0;
	altzoneBegin.dayofweek = 0;

	altzoneEnd.month = 0;
	altzoneEnd.day = 0;
	altzoneEnd.hour = 0;
	altzoneEnd.minute = 0;
	altzoneEnd.dayofweek = 0;

	tod_RdWrNVRAM(TOD_RDNVRAM);	/* read NVRAM and fill in Zone data */

#ifdef TODDEBUG
	printf("Data Zone:%s %8d AltZone:%s %8d\nBegin Mo:%2d Day:%2d Hr:%2d Min:%2d DayOfWeek:%1d\nEnd   Mo:%2d Day:%2d Hr:%2d Min:%2d DayOfWeek:%1d\n",
		tzname[0], timezone, tzname[1], altzone,
		altzoneBegin.month,
		altzoneBegin.day,
		altzoneBegin.hour,
		altzoneBegin.minute,
		altzoneBegin.dayofweek,

		altzoneEnd.month,
		altzoneEnd.day,
		altzoneEnd.hour,
		altzoneEnd.minute,
		altzoneEnd.dayofweek
		);
#endif

	tod_device_on();		/* TOD device is running */

	return(0);

} /* tod_initial */

/************************************************************************/
/*									*/
/* ROUTINE TOD_RDWRNVRAM	Read/Write TOD DATA			*/
/*									*/
/* Arguments:								*/
/*									*/
/*	(in)	WrRd	int	0 = Write NVRAM; 1 = Read NVRAM		*/
/*									*/
/* Return:								*/
/*									*/
/*	 0 = Success							*/
/*									*/
LOCAL int tod_RdWrNVRAM(int WrRd)
{
#ifndef TOD_NONE

	char buf[80];
	char tz_name[10];
	char atz_name[10];

#ifdef TODDEBUG
	int siz = 50;
#endif
	int size = 256;

	if (TOD_MAXNVRAM > size)
	{

	switch (WrRd)
	{
	case TOD_WRNVRAM:

	sprintf(buf,"%s %6d %s %6d %2d %2d %2d %2d %1d %2d %2d %2d %2d %1d%c",
		tzname[0], timezone, tzname[1], altzone,
		altzoneBegin.month,
		altzoneBegin.day,
		altzoneBegin.hour,
		altzoneBegin.minute,
		altzoneBegin.dayofweek,

		altzoneEnd.month,
		altzoneEnd.day,
		altzoneEnd.hour,
		altzoneEnd.minute,
		altzoneEnd.dayofweek,
		'\0'
		);
#ifdef TODDEBUG
		siz = strlen(buf)+1;
		printf("WrBuffer size = %d\n",siz);
#endif
		sysNvRamSet(buf, TOD_NVRAM_SIZE, TOD_NVRAM_OFFSET);

	break;

	case TOD_RDNVRAM:

		sysNvRamGet(buf, TOD_NVRAM_SIZE, TOD_NVRAM_OFFSET);

		buf[TOD_NVRAM_SIZE+1] = '\0';

	sscanf(buf,"%s %6d %s %6d %2d %2d %2d %2d %1d %2d %2d %2d %2d %1d",
		&tz_name[0], &timezone, &atz_name[0], &altzone,
		&altzoneBegin.month,
		&altzoneBegin.day,
		&altzoneBegin.hour,
		&altzoneBegin.minute,
		&altzoneBegin.dayofweek,

		&altzoneEnd.month,
		&altzoneEnd.day,
		&altzoneEnd.hour,
		&altzoneEnd.minute,
		&altzoneEnd.dayofweek
		);

		tz_name[3] = '\0';
		atz_name[3] = '\0';

		strcpy(tzname[0],tz_name);
		strcpy(tzname[1],atz_name);

		break;
	} /* switch */

	} /* if nvram > 256 bytes */

#endif /* TOD_NONE */

	return (0);

} /* tod_RdWrNVRAM */

/************************************************************************/
/*									*/
/* ROUTINE TOD_SET_TZRULES	Assign timezone rules and names		*/
/*									*/
/* Arguments:								*/
/*									*/
/*	(in)	timezone_hhmm	INT32 timezone (+E/-W) Hour Minute	*/
/*	(in)	altzone_hhmm	INT32 alttimezone (+E/-W) Hour Minute	*/
/*	(in)	altzoneBegin_mmddhhmmd	INT32 Month Day Hour Minute DoW	*/
/*	(in)	altzoneEnd_mmddhhmmd	INT32 Month Day Hour Minute DoW	*/
/*	(in)	*timezone_name	char	"nnn" 3 character description	*/
/*	(in)	*altzone_name	char	"nnn" 3 character description	*/
/*									*/
/* Return:								*/
/*		 0 Success						*/
/*		-1 Invalid						*/
/*									*/
/* Timezone Values and Rules: (decimal)					*/
/*									*/
/* tod_set_tzrules(	INT32 timezone_hhmm,				*/
/*			INT32 altzone_hhmm,				*/
/*			INT32 altzoneBegin_mmddhhmmd,			*/
/*			INT32 altzoneEnd_mmddhhmmd,			*/
/*			char timezone_name[],				*/
/*			char altzone_name[]);				*/
/*									*/
/* Zone Format:  +/- hhmm (+East, -West GMT)				*/
/* 			hh = hours	(0-23)				*/
/*			mm = minutes	(00-59)				*/
/*									*/
/* Conditional Format:	altzoneBegin/altzoneEnd mmddhhmmd		*/
/*			mm = Month	(1 - 12)			*/
/*			dd = Day	(01 - 31) / 00,01 Conditional	*/
/*			hh = Hour	(00 - 23)			*/
/*			mm = Minute	(00 - 59)			*/
/*			 d = DayOfWeek	0=Exact Date / (1 - 7)Condition	*/
/*					1=Sunday, 7=Saturday		*/
/*									*/
/* Format:								*/
/* mmddhhmmd	Month, Day, Hour, Minutes, 0	Exact Day (1-31)	*/
/* mmddhhmmd	Month,  00, Hour, Minutes, 1 	Condition First Sunday	*/
/* mmddhhmmd	Month,  01, Hour, Minutes, 1	Condition Last Sunday	*/
/* mmddhhmmd	Month,  00, Hour, Minutes, 0 	IGNORE rules (GMT only)	*/
/*									*/
/* Example: tod_set_tzrules(-600,-500,40002001,100102001,"CST","CDT");	*/
/*	timezone = GMT - 6hr, 0min;	tzname   = CST			*/
/*	altzone  = GMT - 5hr, 0min;	altzname = CDT			*/
/*									*/
/* Example: tod_set_tzrules(0,0,0,0,"GMT","  ");			*/
/*	timezone = GMT time;		tzname   = GMT			*/
/*	altzone  = none;		altzname = none			*/
/*									*/
int tod_set_tzrules(INT32 tz_hhmm, INT32 atz_hhmm,
			INT32 atzBegin_mmddhhmmd,
			INT32 atzEnd_mmddhhmmd,
			char *tz_name, char *atz_name)
{
	INT32 month;
	INT32 day;
	INT32 hour;
	INT32 minute;
	INT32 dayofweek;
	INT32 remainder;

	hour = tz_hhmm / 100;
	minute = tz_hhmm % 100;
	timezone = (hour * 60 * 60) + minute * 60;

	hour = atz_hhmm / 100;
	minute = atz_hhmm % 100;
	altzone = (hour * 60 * 60) + minute * 60;
	
	
	month = atzBegin_mmddhhmmd / 10000000;
	if (month > 0) month --;
	remainder = atzBegin_mmddhhmmd % 10000000;
	day = remainder / 100000;
	remainder = remainder % 100000;
	hour = remainder / 1000;
	remainder = remainder % 1000;
	minute = remainder / 10;
	dayofweek = remainder % 10;

	/* check if valid */
	if (((month < 0) || (month > 11)) || 
		(day > 31) || (hour > 23) ||
		(minute > 59) || (dayofweek > 7))
		return(-1);
	
	altzoneBegin.month = month;
	altzoneBegin.day = day;
	altzoneBegin.hour = hour;
	altzoneBegin.minute = minute;
	altzoneBegin.dayofweek = dayofweek;

#ifdef TODDEBUG
	printf("Mo:%d; Day:%d; Hr:%d; Min:%d; WDay:%d\n",
		altzoneBegin.month,altzoneBegin.day,altzoneBegin.hour,
		altzoneBegin.minute,altzoneBegin.dayofweek);
#endif

	month = atzEnd_mmddhhmmd / 10000000;
	if (month > 0) month --;
	remainder = atzEnd_mmddhhmmd % 10000000;
	day = remainder / 100000;
	remainder = remainder % 100000;
	hour = remainder / 1000;
	remainder = remainder % 1000;
	minute = remainder / 10;
	dayofweek = remainder % 10;

	/* check if valid */
	if (((month < 0) || (month > 11)) || 
		(day > 31) || (hour > 23) ||
		(minute > 59) || (dayofweek > 7))
		return(-1);
	
	altzoneEnd.month = month;
	altzoneEnd.day = day;
	altzoneEnd.hour = hour;
	altzoneEnd.minute = minute;
	altzoneEnd.dayofweek = dayofweek;

#ifdef TODDEBUG
	printf("Mo:%d; Day:%d; Hr:%d; Min:%d; WDay:%d\n",
		altzoneEnd.month,altzoneEnd.day,altzoneEnd.hour,
		altzoneEnd.minute,altzoneEnd.dayofweek);
#endif
	if ((strlen(tz_name) <= 3) && (strlen(atz_name) <= 3))
	{
		tzname[0] = tz_name;
		tzname[1] = atz_name;
	}
	else return(-1);
#ifdef TODDEBUG
		printf("TZ:%s ATZ:%s\n",tzname[0],tzname[1]);
#endif

	tod_RdWrNVRAM(TOD_WRNVRAM);	/* write the Zone data to NVRAM */
	daylight = 0;

	return(0);

} /* tod_set_tzrules */

/************************************************************************/
/*									*/
/* ROUTINE TOD_SHOW_TZRULES	Show TimeZone Data and Display		*/
/*									*/
/* Arguments:								*/
/*									*/
/*		none							*/
/* Return:								*/
/*									*/
/*	 0 = Success							*/
/*									*/
int tod_show_tzrules(void)
{
	printf("Timezone Set Rules: (decimal)\n\n");
	printf("tod_set_tzrules( INT32 timezone_hhmm, INT32 altzone_hhmm,\n");
	printf("		 INT32 altzoneBegin_mmddhhmmd, INT32 altzoneEnd_mmddhhmmd,\n"); 
	printf("		 char *timezone_name, char *altzone_name);\n\n");
	printf("Zone Format:  +/- hhmm (+East, -West GMT)\n");
	printf("		hh = hours	(0-23)\n");
	printf("		mm = minutes	(00-59)\n\n");
	printf("Conditional Format:	altzoneBegin/altzoneEnd mmddhhmmd\n");
	printf("		mm = Month	(1 - 12)\n");
	printf("		dd = Day	(01 - 31) / 00,01 Conditional\n");
	printf("		hh = Hour	(00 - 23)\n");
	printf("		mm = Minute	(00 - 59)\n");
	printf("		 d = DayOfWeek	0 Exact Date / (1 - 7) Conditional\n\n");
	printf("Example: tod_set_tzrules(-600, -500, 40002001, 100102001, \"CST\", \"CDT\");\n");
	printf("		timezone = GMT - 6hr, 0min;	tzname   = CST\n");
	printf("		altzone  = GMT - 5hr, 0min;	altzname = CDT\n");
	printf("		Altzone begins the first Sunday in April at 2:00\n");
	printf("		Altzone ends   the last  Sunday in October at 2:00\n");

	return(0);

 } /* tod_show_tzrules */

/************************************************************************/
/*									*/
/* ROUTINE TOD_GET_TZRULES	Get TimeZone Data and Display		*/
/*									*/
/* Arguments:								*/
/*									*/
/*		none							*/
/* Return:								*/
/*									*/
/*	 0 = Success							*/
/*									*/
int tod_get_tzrules(void)
{
	printf("Timezone Data:\n\n");
 
	printf("Timezone = %d Seconds;	Altzone = %d Seconds\n", timezone, altzone);
	printf("Timezone   %s 		Altzone   %s\n", tzname[0], tzname[1]);

	if (!altzoneBegin.dayofweek)
		printf("DST RULE Begins on %s %d %02d:%02d:00\n",
			tod_asciimonth(altzoneBegin.month),
			altzoneBegin.day, altzoneBegin.hour, altzoneBegin.minute);
	else
	{
		if (!altzoneBegin.day)
			printf("DST RULE Begins on %s FIRST %s %02d:%02d:00\n",
				tod_asciimonth(altzoneBegin.month),
				tod_asciiday(altzoneBegin.dayofweek-1),
				altzoneBegin.hour, altzoneBegin.minute);
		else
			printf("DST RULE Begins on %s LAST  %s %02d:%02d:00\n",
				tod_asciimonth(altzoneBegin.month),
				tod_asciiday(altzoneBegin.dayofweek-1),
				altzoneBegin.hour, altzoneBegin.minute);
	}			

	if (!altzoneEnd.dayofweek)
		printf("DST RULE Ends on   %s %d %02d:%02d:00\n",
			tod_asciimonth(altzoneEnd.month),
			altzoneEnd.day, altzoneEnd.hour, altzoneEnd.minute);
	else
	{
		if (!altzoneEnd.day)
			printf("DST RULE Ends on   %s FIRST %s %02d:%02d:00\n",
				tod_asciimonth(altzoneEnd.month),
				tod_asciiday(altzoneEnd.dayofweek-1),
				altzoneEnd.hour, altzoneEnd.minute);
		else
			printf("DST RULE Ends on   %s LAST  %s %02d:%02d:00\n", 
				tod_asciimonth(altzoneEnd.month), 
				tod_asciiday(altzoneEnd.dayofweek-1),
				altzoneEnd.hour, altzoneEnd.minute);
	}

	return(0);

} /* tod_get_tzrules */

/************************************************************************/
/*									*/
/* ROUTINE TOD_TZRULES_CONDITIONAL Determine the new day (First/Last)	*/
/*									*/
/* The Day of the Month to check agnist for "daylight"			*/
/*									*/
/* Arguments:								*/
/*									*/
/*	(in)	year	int	Current year				*/
/*	(in)	month	int	Current month				*/
/*	(in)	day	int	0 = First Day, 1 - Last Day of month	*/
/*	(in)	dayofweek	int	1(Sun) - 7(Sat)			*/
/*									*/
/* Return:								*/
/*									*/
/*	 newday int							*/
/*									*/
LOCAL int tod_tzrules_conditional(int year, int month, int day, int dayofweek)
{
	int test_yday;
	int dayofmonth;
	int daysinmonth;
	int ndays;
	int newday;

	if (!day)
	{
		/* Knowing Year, compute first dayofweek condition */

		test_yday = tod_julday(year,month,0);
		ndays = tod_ep_days(year -TOD_YEAR0, test_yday);
		dayofmonth = (ndays + TOD_DAY0) % 7;

		/* find first (dayofweek) day of month */
		for (newday = 1, ndays = dayofmonth;
			(newday <= 7) && ((ndays % 7) != (dayofweek - 1));
			 newday++, ndays++) {;}
#ifdef TODDEBUG
		printf("Condition First: %d; NewDay:%d\n",dayofmonth,newday);
#endif
	}
	else
	{
		/* Knowing Year, compute last condition */
		daysinmonth = tod_daysinmonth(year,month);
		test_yday = tod_julday(year,month,daysinmonth-1);
		ndays = tod_ep_days(year -TOD_YEAR0, test_yday);
		dayofmonth = (ndays + TOD_DAY0) % 7;

		/* find last (dayofweek) day of month */
		for (newday = daysinmonth - 6, ndays = dayofmonth + 1;
			(newday <= daysinmonth) && 
			((ndays % 7) != (dayofweek - 1));
			 newday++, ndays++) {;}

#ifdef TODDEBUG
		printf("Condition Last: %d; NewDay:%d\n",dayofmonth,newday);
#endif

	}

	return (newday);

} /* tod_tzrules_conditional */

/************************************************************************/
/*									*/
/* ROUTINE TOD_TZSET	Determine altTimezone (NOT UNIX like)		*/
/*									*/
/* Check current time aganist the tzrules struct to determine the	*/
/* the setting of "daylight"						*/
/*									*/
/* Arguments:		Check current time aganist the tzrules struct	*/
/*									*/
/*		int timer						*/
/* Return:								*/
/*									*/
/*		0							*/
/*									*/
LOCAL void tod_tzset(int timer)
{
	int dststart = 0;
	int dststop = 1;

	INT32 year;
	INT32 month;
	INT32 hour;
	INT32 minute;
	INT32 second;

	INT32 days;
	INT32 mday;		/* day of the month */

	/*INT32 test_yday;*/	/* Conditional	day of the year */
	/*INT32 dayofmonth;*/	/*		First/Last dayofweek */
	INT32 new_day;		/*		day from dayofweek */

	INT32 remainder;
	int hysteresis;		/* Prevent toggling between zone changes */


	if (!(altzoneBegin.dayofweek + altzoneBegin.day))
		return;

	hysteresis = (timezone - altzone) / 3600;

	/* Determine the current Year */

	if (daylight)
		timer += altzone;
	else
		timer += timezone;

	days = timer / TOD_SECS_PER_DAY;
	remainder = timer % TOD_SECS_PER_DAY;
	hour = remainder / 3600;
	remainder = remainder % 3600;
	minute = remainder / 60;
	second = remainder % 60;

	year = days / 365;
	while ( tod_ep_days(year,0) > days)
		year --;

	days -= tod_ep_days(year, 0);
	year += TOD_YEAR0;

	for (month = 0; (days >= tod_julday(year,month+1,0)) && (month<11); month++)
	{;}

	mday = days - tod_julday(year,month,0) + 1;

#ifdef TODDEBUG
	printf("Current Yr: %d; Mo:%d; Day:%d; Hr:%d; Min:%d; Sec:%d\n",
		year,month,mday,hour,minute,second);
#endif
	if (altzoneBegin.dayofweek != 0)
	{
		new_day = tod_tzrules_conditional(year,altzoneBegin.month,
			altzoneBegin.day, altzoneBegin.dayofweek);

	}
	else new_day = altzoneBegin.day;

	if (month > altzoneBegin.month) dststart = 1;
	else if (month == altzoneBegin.month)
	{
		if (mday > new_day) dststart = 1;
		else if (mday == new_day)
		{
#ifdef TODDEBUG
		printf("BEGIN Dailight:%d Hour:%d altzoneBegin.hour:%d +/-TZ:%d\n",
			daylight,hour,altzoneBegin.hour,hysteresis);
#endif
			/* special case hysteresis */
			if ((daylight) && (hour >= (altzoneBegin.hour+hysteresis)))
				dststart = 1;
			else if (hour > altzoneBegin.hour) dststart = 1;
			else if (hour == altzoneBegin.hour)
			{
				if (minute > altzoneBegin.minute) dststart = 1;
				else if (minute == altzoneBegin.minute)
				{
					if (second > 0) dststart = 1;
				}
			}
		}	
	}


#ifdef TODDEBUG
	printf("BEGIN CMP Yr:%d; Mo:%d; Day:%d; Hr:%d; Min:%d; DSTSTART:%d\n",
	year,altzoneBegin.month,mday,altzoneBegin.hour,altzoneBegin.minute,dststart);
#endif

	if (altzoneEnd.dayofweek != 0)
	{
	new_day = tod_tzrules_conditional(year,altzoneEnd.month,
		altzoneEnd.day, altzoneEnd.dayofweek);
	}
	else new_day = altzoneEnd.day;

	if (month > altzoneEnd.month) dststop = 0;
	else if (month == altzoneEnd.month)
	{
		if (mday > new_day) dststop = 0;
		else if (mday == new_day)
		{
			/* special case hysteresis */
#ifdef TODDEBUG
		printf("END Dailight:%d Hour:%d altzoneEnd.hour:%d +/-TZ:%d\n",
		daylight,hour,altzoneEnd.hour,hysteresis);
#endif
			if ((!daylight) && (hour >= (altzoneEnd.hour+hysteresis)))
				dststop = 0;
			else if (hour > altzoneEnd.hour) dststop = 0;
			else if (hour == altzoneEnd.hour)
			{
				if (minute >= altzoneEnd.minute) dststop = 0;
			}
		}
	}


#ifdef TODDEBUG
	printf("END   CMP Yr:%d; Mo:%d; Day:%d; Hr:%d; Min:%d; DSTSTOP:%d\n",
	year,altzoneEnd.month,mday,altzoneEnd.hour,altzoneEnd.minute,dststop);
#endif
	daylight = dststart & dststop;

#ifdef TODDEBUG
	printf("DAYLIGHT:%d\n",daylight);
#endif

} /* tod_tzset */

/************************************************************************/

/************************************************************************/
/*									*/
/* ROUTINE TOD_SECTIME		Determine user interface Date/Time	*/
/*									*/
/* Converts user time to year month day hours minutes seconds		*/
/*									*/
/* Arguments:								*/
/*		(decimal)						*/
/*	(in)	yymmdd	INT32	Year(1970-2070) Month(1-12) Day(1-31)	*/
/*	(in)	hhmmss	INT32	Hours(0-23) Minutes(0-59) Seconds(0-59)	*/
/*									*/
/* Return:								*/
/*									*/
/*	 0 = Seconds							*/
/*	-1 = Invalid							*/
/*									*/
LOCAL INT32 tod_sectime(INT32 yymmdd, INT32 hhmmss)
{
    struct tm	*tp;
    struct tm	tmp;
    time_t	curtime;
    INT32	rem;

    tp = &tmp;

    time( &curtime );			/* Get current time		*/
    tp = gmtime( &curtime );		/* Fill in *tp			*/

    if ( yymmdd )			/* If date is present,		*/
    {					/*   set date			*/
	tp->tm_year = yymmdd / 10000;	/* Get year field		*/
	if ( tp->tm_year < TOD_YEAR0 )	/* If < 70, we're in 21st century*/
	    tp->tm_year += 100;
	rem = yymmdd % 10000;
	tp->tm_mon = (rem / 100) - 1;	/* Get month			*/
	tp->tm_mday = rem % 100;	/* Get day			*/
	if ( (tp->tm_mon > 11) || (tp->tm_mday > 31) )
	    return(-1);			/* Check for valid date		*/
    }
    else if ( hhmmss == 0 )
	return(-1);			/* If no parms, print usage	*/

    tp->tm_hour = hhmmss / 10000;	/* Get hour			*/
    rem = hhmmss % 10000;
    tp->tm_min = rem / 100;		/* Get minute			*/
    tp->tm_sec = rem % 100;		/* Get second			*/
    if ( (tp->tm_hour > 23) || (tp->tm_min > 59) || (tp->tm_sec > 59) )	
	return(-1);			/* Check for valid time		*/

    tp->tm_yday = tod_julday(tp->tm_year, tp->tm_mon, tp->tm_mday) - 1;

    curtime = tod_ep_days(tp->tm_year - TOD_YEAR0, tp->tm_yday) * TOD_SECS_PER_DAY;
    curtime += (tp->tm_hour * 3600L) + (tp->tm_min * 60) + tp->tm_sec;

    return(curtime);

} /* tod_sectime */

/************************************************************************/
/*									*/
/* ROUTINE TOD_SETUSAGE		Display Format expected			*/
/*									*/
/* Arguments:								*/
/*		none							*/
/* Return:								*/
/*									*/
/*	 	none							*/
LOCAL void tod_setusage(void)
{
	printf("Usage: (decimal)\n");
	printf("		tod_setlocal (yymmdd, hhmmss)\n");
	printf("		tod_setgmt   (yymmdd, hhmmss)\n");
	printf("		yy = year\n");
	printf("		mm = month\n");
	printf("		dd = day\n");
	printf("		hh = hour\n");
	printf("		mm = minute\n");
	printf("		ss = second\n");

} /* tod_setusage */

/************************************************************************/
/*									*/
/* ROUTINE TOD_SETLOCAL		Set the TOD device using local time	*/
/*									*/
/* This routine converts local time to GMT time and sets the device in	*/
/* GMT time.								*/
/*									*/
/* Arguments:								*/
/*		(decimal)						*/
/*	(in)	yymmdd	INT32	Year(1970-2070) Month(1-12) Day(1-31)	*/
/*	(in)	hhmmss	INT32	Hours(0-23) Minutes(0-59) Seconds(0-59)	*/
/*									*/
/* Return:								*/
/*									*/
/*	 0 = Success							*/
/*	-1 = Invalid							*/
/*									*/
int tod_setlocal(INT32 yymmdd, INT32 hhmmss)
{
	time_t current;

	current = tod_sectime(yymmdd, hhmmss);

	tod_tzset(current);	/* set daylight if appropriate */

	if (current == -1)
	{
		tod_setusage();
		return (current);
	}

	if (daylight)
		current -= altzone;
	else
		current -= timezone;

	stime ((long *)&current);

	date();

	return (0);

} /* tod_setlocal */

/************************************************************************/
/*									*/
/* ROUTINE TOD_SETGMT		Set the TOD device using GMT time	*/
/*									*/
/* This routine sets the device in GMT time.				*/
/*									*/
/* Arguments:								*/
/*		(decimal)						*/
/*	(in)	yymmdd	INT32	Year(1970-2070) Month(1-12) Day(1-31)	*/
/*	(in)	hhmmss	INT32	Hours(0-23) Minutes(0-59) Seconds(0-59)	*/
/*									*/
/* Return:								*/
/*									*/
/*	 0 = Success							*/
/*	-1 = Invalid							*/
/*									*/
int tod_setgmt(INT32 yymmdd, INT32 hhmmss)
{
	time_t current;

	current = tod_sectime(yymmdd, hhmmss);

	if (current == -1)
	{
		tod_setusage();
		return (current);
	}
	stime ((long *)&current);

	date();

	return (0);

} /* tod_setgmt */

/* file: tod_time.c */
