/************************************************************************/
/*									*/
/*	Computing Department						*/
/*	P.O. BOX 500							*/
/*	Fermilab, Batavia, Il 60510, U.S.A.				*/
/*									*/
/* File: vx_tools.c							*/
/*									*/
/* This file provides supplemental generic routines for a global use	*/
/* with VxWorks OS.							*/
/*									*/
/* History:								*/
/* Jun 01, 1993 swk-added auto stack checking hooks for enable/disable	*/
/* Jun 01, 1993 swk-added a sloppy socket cleanup that invokes reboot	*/
/* Jan 07, 1994 das-added VXWORKS conditionals on header files		*/
/*									*/

/*
SCCS release:
@(#) vx_tools.c 1.5 Delta: 94/01/10 14:57:17 Extraction: 94/01/11 09:41:35 @(#)
*/

/* common include files */
#include <vxWorks.h>
#ifdef MIPS_R3000
#include <fioLib.h>
extern 	STATUS       shutdown (int s, int how);	/* in lieu of sockLib.h */
#else
#include <stdioLib.h>
#endif
#include <types.h>			/* FD_SETSIZE def */


/* stack checking specific include files */
#include <taskLib.h>			/* task defs */
#include <taskHookLib.h>		/* task hook defs */
#include <logLib.h>
#include <sigLib.h>


/* cleanup and reboot specific include files */
#include <socket.h>			/* socket defs */
#ifndef MIPS_R3000
#include <sockLib.h>			/* socket shutdown() defs */
#endif
#include <sysLib.h>			/* reboot() defs */
#include <rebootLib.h>			/* reboot */

/* vxtool definitions */
#include "vx_tools.h"

/************************************************************************/
/* Section: Stack Checking						*/
/*									*/
/* This tool provides a mechanism to dynamiclly turn on/off stack	*/
/* checking NOTE - considerable overhead.				*/
/*									*/
/* The enable adds a hook so that the hook function is invoked when	*/
/* a task is created, deleted or switched. The hook function does the	*/
/* actual stack checking. The disable removes a hook funcion form a	*/
/* hook list.								*/
/*									*/
/* AUTO STACK CHECKING ROUTINES in General				*/
/*									*/
/*	AutoStackSwiCheck()	Hook on task switch for stack check	*/
/*	AutoStackDelCheck()	Hook on task delete for stack check	*/
/*	AutoStackCreCheck()	Hook on task create for stack check	*/
/*									*/
/* DEFINITIONS								*/
/*									*/
/*	AUTOSTK_SWITCH 1						*/
/*	AUTOSTK_DELETE 2						*/
/*	AUTOSTK_CREATE 4						*/
/* 	AUTOSTACKUNDERFLOW	if defined, underflow will be checked	*/
/*									*/
/*									*/
/* These routines do stack checking. Stack space is initialized with	*/
/* a 0xee pattern used for stack tracing. The int 0xeeeeeeee was 	*/
/* chosen instead of the byte 0xee the VxWorks Manual describes		*/
/* because, it is thought that this pattern MAY be less likely to	*/
/* occur within real usage (maybe).					*/
/*									*/
/* This routine first, checks the interupt stack to verify that the 	*/
/* end = 0xeeeeeeee. Second, The pStackLimit (end) is tested for the	*/
/* int 0xeeeeeeee pattern. Third the task stack pointer is tested	*/
/* to verify that SP is within the pStackBase and pStackLimit.		*/
/*									*/
/* Routines:								*/
/*									*/
/*	VOID	AutoStackSwiCheck (WIND_TCB *pOldTcb,WIND_TCB *pNewTcb)	*/
/*	VOID	AutoStackDelCheck (WIND_TCB *pNewTcb)			*/
/*	VOID	AutoStackCreCheck (WIND_TCB *pNewTcb)			*/
/*									*/
/*	STATUS	autoStackCheck (UINT8 chk)				*/
/*									*/
/*		chk = 0	ALL OFF	(Normal)				*/
/*		chk = 1	SWITCH  (Default)	mask 0-7		*/
/*		chk = 2 DELETE						*/
/*		chk = 4 CREATE						*/
/*									*/
/*	void	autoStackShow (void)					*/
/*									*/

/************************************************************************/
/*									*/
/* PROTOTYPES								*/
/*									*/
VOID AutoStackSwiCheck (WIND_TCB *pOldTcb, WIND_TCB *pNewTcb);
VOID AutoStackDelCheck (WIND_TCB *pNewTcb);
VOID AutoStackCreCheck (WIND_TCB *pNewTcb);

STATUS autoStackCheck (UINT8 chk);

void autoStackShow (void);

/************************************************************************/
/*									*/
/* GLOBAL DEFINITIONS & VARIABLES					*/
/*									*/

#ifndef AUTOSTK_NORMAL
#define AUTOSTK_NORMAL 0	/* NORMAL */
#endif

#ifndef AUTOSTK_SWITCH
#define AUTOSTK_SWITCH 1	/* Mask bit check on task switch */
#endif

#ifndef AUTOSTK_DELETE
#define	AUTOSTK_DELETE 2	/* Mask bit check on task delete */
#endif

#ifndef AUTOSTK_CREATE
#define	AUTOSTK_CREATE 4	/* Mask bit check on task create */
#endif

IMPORT char *vxIntStackBase;	/* Intr Stack base and end */
IMPORT char *vxIntStackEnd;

				/* Bits  7 6 5 4 3 2 1 0  */     
LOCAL UINT8 enableflag = 0;	/* Flag |0|0|0|0|0|S|D|C| */
				/* S = Switch */
				/* D = Delete */
				/* C = Create */

/************************************************************************/
/*									*/
/* ROUTINE AUTOSTACKSWICHECK	Hook for switch stack checking routine	*/
/*									*/
/* This routine does stack checking on the switching of a task.		*/
/*									*/
/* Arguments:								*/
/*	(in)	pNewTcb	WIND_TCB	New Task control block		*/
/*	(in)	pOldTcb	WIND_TCB	Old Task control block		*/
/*									*/
/* Return:								*/
/*	 none	Error messages are logged with "logMsg()"		*/
/*									*/
VOID AutoStackSwiCheck (WIND_TCB *pOldTcb, WIND_TCB *pNewTcb)
{

	/* check Intr stack */
	if ( *(int *) vxIntStackEnd != 0xeeeeeeee )
		logMsg("Task Switch(%s): INTR Error Stack Overflow\n",
			(int)pOldTcb->name,0,0,0,0,0);

	if ( ( *(int *) pOldTcb->pStackLimit != 0xeeeeeeee) ||
		((int) pOldTcb->regs.spReg < (int) pOldTcb->pStackLimit) )
		logMsg("Task Switch(%s): Error Stack Overflow\n",
			(int)pOldTcb->name,0,0,0,0,0);

#ifdef AUTOSTACKUNDERFLOW
	if ( (int) pOldTcb->regs.spReg > (int) pOldTcb->pStackBase )
		logMsg("Task Switch(%s): Error Stack Underflow\n",
			pOldTcb->name);
#endif
	return;

} /* autostackswicheck */

/************************************************************************/
/*									*/
/* ROUTINE AUTOSTACKDELCHECK	Hook for delete stack checking routine	*/
/*									*/
/* This routine does stack checking on the deletion of a task.		*/
/*									*/
/* Arguments:								*/
/*	(in)	pNewTcb	WIND_TCB	Task control block		*/
/*									*/
/* Return:								*/
/*	 none	Error messages are logged with "logMsg()"		*/
/*									*/
VOID AutoStackDelCheck (WIND_TCB *pNewTcb)
{
	/* check Intr stack */
	if ( *(int *) vxIntStackEnd != 0xeeeeeeee )
		logMsg("Task Delete(%s): INTR Error Stack Overflow\n",
			(int)pNewTcb->name,0,0,0,0,0);

	if ( ( *(int *) pNewTcb->pStackLimit != 0xeeeeeeee) ||
		((int) pNewTcb->regs.spReg < (int) pNewTcb->pStackLimit) )
		logMsg("Task Delete(%s): Error Stack Overflow\n",
			(int)pNewTcb->name,0,0,0,0,0);

#ifdef AUTOSTACKUNDERFLOW
	if ( (int) pNewTcb->regs.spReg > (int) pNewTcb->pStackBase )
		logMsg("Task Delete(%s): Error Stack Underflow\n",
			pNewTcb->name);
#endif
	return;

} /* autostackdelcheck */

/************************************************************************/
/*									*/
/* ROUTINE AUTOSTACKCRECHECK	Hook for create stack checking routine	*/
/*									*/
/* This routine does stack checking on the creation of a task.		*/
/*									*/
/* Arguments:								*/
/*	(in)	pNewTcb	WIND_TCB	Task control block		*/
/*									*/
/* Return:								*/
/*	 none	Error messages are logged with "logMsg()"		*/
/*									*/
VOID AutoStackCreCheck (WIND_TCB *pNewTcb)
{
	/* check Intr stack */
	if ( *(int *) vxIntStackEnd != 0xeeeeeeee )
		logMsg("Task Create(%s): INTR Error Stack Overflow\n",
			(int)pNewTcb->name,0,0,0,0,0);

	if ( ( *(int *) pNewTcb->pStackLimit != 0xeeeeeeee) ||
		((int) pNewTcb->regs.spReg < (int) pNewTcb->pStackLimit) )
		logMsg("Task Create(%s): Error Stack Overflow\n",
			(int)pNewTcb->name,0,0,0,0,0);

#ifdef AUTOSTACKUNDERFLOW
	if ( (int) pNewTcb->regs.spReg > (int) pNewTcb->pStackBase )
		logMsg("Task Create(%s): Error Stack Underflow\n",
			pNewTcb->name);
#endif
	return;

} /* autostackcrecheck */

/************************************************************************/
/*									*/
/* ROUTINE AUTOSTACKCHECK	Hook Add/Delete for stack checking	*/
/*									*/
/* This routine enables/disables auto stack checking with mask.		*/
/*									*/
/* You Don't want to add the hook more than once.			*/
/*									*/
/* Arguments:								*/
/*	(in)	chk	UINT8 	0: Delete all	(Normal)		*/
/*				1: onSwitch	(Default)		*/
/*				2: onDelete				*/
/*				4: onCreate				*/
/*									*/
/* Return:								*/
/*	 STATUS = OK							*/
/*	 STATUS = ERROR							*/
/*									*/
STATUS autoStackCheck (UINT8 chk)
{
	STATUS err;

	/* Is it invalid ? */
	if ((int)chk > (AUTOSTK_SWITCH | AUTOSTK_DELETE | AUTOSTK_CREATE) )
		return (ERROR);

	if ( !chk)				/* All OFF (Normal) */
	{
		err = taskSwitchHookDelete( (FUNCPTR ) AutoStackSwiCheck);
		err = taskDeleteHookDelete( (FUNCPTR ) AutoStackDelCheck);
		err = taskCreateHookDelete( (FUNCPTR ) AutoStackCreCheck);

		enableflag = 0x00;

		return (OK);	/* who cares? */
	} /* normal */


	if ((int)chk & AUTOSTK_SWITCH)			/* Switch */
	{
		if ( !((int)enableflag & AUTOSTK_SWITCH))	/* Add */
		{
		if ( taskSwitchHookAdd( (FUNCPTR ) AutoStackSwiCheck) != OK)
			return (ERROR);
		enableflag |= AUTOSTK_SWITCH;
		}
	}
	else
	{
		if ((int)enableflag & AUTOSTK_SWITCH)	/* Remove */
		{
		if ( taskSwitchHookDelete( (FUNCPTR ) AutoStackSwiCheck) != OK)
			return (ERROR);
		enableflag &= ~AUTOSTK_SWITCH;
		}
	} /* switch */

	if ((int)chk & AUTOSTK_DELETE)			/* Delete */
	{
		if ( !((int)enableflag & AUTOSTK_DELETE))	/* Add */
		{
		if ( taskDeleteHookAdd( (FUNCPTR ) AutoStackDelCheck) != OK)
			return (ERROR);
		enableflag |= AUTOSTK_DELETE;
		}
	}
	else
	{
		if ((int)enableflag & AUTOSTK_DELETE)	/* Remove */
		{
		if ( taskDeleteHookDelete( (FUNCPTR ) AutoStackDelCheck) != OK)
			return (ERROR);
		enableflag &= ~AUTOSTK_DELETE;
		}
	} /* delete */

	if ((int)chk & AUTOSTK_CREATE)			/* Create */
	{
		if ( !((int)enableflag & AUTOSTK_CREATE))	/* Add */
		{
		if ( taskCreateHookAdd( (FUNCPTR ) AutoStackCreCheck) != OK)
			return (ERROR);
		enableflag |= AUTOSTK_CREATE;
		}
	}
	else
	{
		if ((int)enableflag & AUTOSTK_CREATE)	/* Remove */
		{
		if ( taskCreateHookDelete( (FUNCPTR ) AutoStackCreCheck) != OK)
			return (ERROR);
		enableflag &= ~AUTOSTK_CREATE;
		}
	} /* create */

	return (OK);

} /* autostackcheck */

/************************************************************************/
/*									*/
/* ROUTINE AUTOSTACKSHOW	Which Hooks are Active for stack check	*/
/*									*/
/* This routine shows what's enabled.					*/
/*									*/
/* Arguments:								*/
/*	(none)								*/
/*									*/
/* Return:								*/
/*	(none)								*/
/*									*/
void autoStackShow (void)
{

	if ( !enableflag)
		printf("No task auto stack check hooks added.\n");
	else
		printf("Stack Check Hook(s):\n");
	if ((int)enableflag & AUTOSTK_SWITCH)
		printf("  Task Switch (AutoStackSwiCheck)\n");
	if ((int)enableflag & AUTOSTK_DELETE)
		printf("  Task Delete (AutoStackDelCheck)\n");
	if ((int)enableflag & AUTOSTK_CREATE)
		printf("  Task Create (AutoStackCreCheck)\n");

	return;

} /* autostackshow */
/************************************************************************/

/************************************************************************/

/************************************************************************/
/* Section: fermi cleanup sockets					*/
/*									*/
/* This tool provides a mechanism to close remaining open sockets	*/
/* prior to reboot.							*/
/*									*/
/* Currently, the user is responsible for executing "fermi_reboot()"	*/
/* instead of the "reboot()" or ^X.					*/
/*									*/
/* This routine does NOT address socket closings on the following	*/
/* conditions:								*/
/*									*/
/*	1) executing reboot() or ^X					*/
/*	2) power cycle of power glitch					*/
/*	3) front panel reset						*/
/*	4) module hung							*/
/*									*/
/* The problem in 20 words or so:					*/
/*									*/
/* The rebootHookAdd function is useless here because the rebootHookAdd	*/
/* executes the added routines on a first come basis. The first hookadd	*/
/* is a vxworks routine in netlib that shuts down the network (doesn't	*/
/* cleanup) so, what is need, is a routine to cleanup sockets prior to	*/
/* reboot().								*/
/*									*/
/* cleanup() scans through and closes any open sockets, and then	*/
/* calls reboot() internally. 						*/
/*									*/

/************************************************************************/
/*									*/
/* GLOBAL DEFINITIONS							*/
/*									*/

/* VxWorks FD_SETSIZE userdefined else VxWorks predefined as 256	*/

#ifdef FD_SETSIZE 
#define FERMI_MAXSOCKET FD_SETSIZE
#else
#define FERMI_MAXSOCKET 64
#endif

/************************************************************************/
/*									*/
/* PROTOTYPES								*/
/*									*/
LOCAL void closeOpenSockets(int mode);
STATUS cleanup(int starttype);

/************************************************************************/
/*									*/
/* ROUTINE CLOSEOPENSOCKETS	Scan through fd and close valid sockets	*/
/*									*/
/* This routine cleans up sockets.					*/
/*									*/
/* Arguments:								*/
/*	(in)	mode	int	 Use: 0=Rcv, 1=Snd, 2=Rcv/Snd		*/
/*									*/
/* Return:								*/
/*	 none								*/
/*									*/
LOCAL void closeOpenSockets(int mode)
{
	int	no_sockets = 0;	/* for test, count valid sockets */
	int	socket_no;

	/* scan through and shutdown all valid sockets */

	for (socket_no = FERMI_MAXSOCKET; socket_no >= 0; socket_no--)
	{
		if (shutdown(socket_no, mode) == OK)
			no_sockets++; /* count of valid closed sockets */

	} /* for */

	return;

} /* closeopensockets */

/************************************************************************/
/*									*/
/* ROUTINE CLEANUP		Close sockets and reboot		*/
/*									*/
/* This routine cleans up sockets and then executes the reboot.		*/
/*									*/
/* Arguments:								*/
/*	(in)	starttype	int	Use:	BOOT_NORMAL (0)		*/
/*						BOOT_NO_AUTOBOOT (1)	*/
/*						BOOT_CLEAR (2)		*/
/*						BOOT_QUICK_AUTOBOOT (4)	*/
/*									*/
/* Return:								*/
/*									*/
/*	OK = Success							*/
/*	ERROR = Invalid							*/
/*									*/
STATUS cleanup(int starttype)
{

	/* verify that the starttype values are valid */

	switch (starttype)
	{
	case  BOOT_NORMAL:
	case  BOOT_NO_AUTOBOOT:
	case  BOOT_CLEAR:
	case  BOOT_QUICK_AUTOBOOT:
		break;
	default:
		return(ERROR);	/* invalid starttype, IGNORE */
	} /* switch */

	closeOpenSockets(2);	/* Close the sockets, try 2=Rcv/Snd */

	reboot(starttype);

	return(OK);

} /* cleanup */
/************************************************************************/

/************************************************************************/

/************************************************************************/
/* next	tool								*/
/* .									*/
/* .									*/
/* .									*/
/*									*/

/* file: vx_tools.c */
